/******************************************************************************
 * 
 * Capacitated m-Ring Star
 * 
 * Solución de m-anillos estrella con capacidades
 * 
 ******************************************************************************/

#include <iostream>
#include <fstream>
#include <iomanip>
#include "CMRS.h"

using namespace std;


CMRS::CMRS(int n_rings, int ring_capacity)
{
    this->n_rings = n_rings;
    this->ring_capacity = ring_capacity;
    customers_labels = set<string>();
    steiners_labels = set<string>();
    ring_connections = set<WeightedArc>();
    star_connections = set<WeightedArc>();
}

void CMRS::depot(const string& depot_label) throw(CMRSPException)
{
    this->depot_label = depot_label;
}

void CMRS::add_customer(const string& customer_label) throw(CMRSPException)
{
    customers_labels.insert(customer_label);
}

void CMRS::add_steiner(const string& steiner_label) throw(CMRSPException)
{
    steiners_labels.insert(steiner_label);
}

void CMRS::add_ring_connection(const string& source_label, const string& target_label, double weight) throw(CMRSPException)
{    
    Arc arc = pair<string, string>();
    arc.first = source_label;
    arc.second = target_label;
    WeightedArc w_arc = pair<Arc, double>();
    w_arc.first = arc;
    w_arc.second = weight;
    ring_connections.insert(w_arc);
}

void CMRS::add_star_connection(const string& source_label, const string& target_label, double weight) throw(CMRSPException)
{
    Arc arc = pair<string, string>();
    arc.first = source_label;
    arc.second = target_label;
    WeightedArc w_arc = pair<Arc, double>();
    w_arc.first = arc;
    w_arc.second = weight;
    star_connections.insert(w_arc);
}

string CMRS::get_depot() const
{
    return depot_label;
}

set<string> CMRS::get_customers() const
{
    return customers_labels;
}

set<string> CMRS::get_steiners() const
{
    return steiners_labels;
}

set<CMRS::WeightedArc> CMRS::get_ring_connections() const
{
    return ring_connections;
}

set<CMRS::WeightedArc> CMRS::get_star_connections() const
{
    return star_connections;
}

bool CMRS::validate() const
{
    // verificar la estructura sea valida:
    // exactamente m anillos, ninguno con mas de q clientes conectados,
    // y estructura anillo-estrella
    return true;
}

void CMRS::save(const string& filename) throw(CMRSPException)
{
    ofstream ofs(filename.c_str(), ofstream::out);
    if (!ofs)
        throw CMRSPException("Error guardando archivo: "+filename);
    
    ofs << ring_connections.size() << endl;
    ofs << star_connections.size() << endl;
    
    set<WeightedArc>::iterator it;
    for (it = ring_connections.begin(); it != ring_connections.end(); it++)
    {
        Arc arc = it->first;
        ofs << arc.first << "\t" << arc.second << endl;
    }
    
    for (it = star_connections.begin(); it != star_connections.end(); it++)
    {
        Arc arc = it->first;
        ofs << arc.first << "\t" << arc.second << endl;
    }
    
    ofs.close();
}

void CMRS::save_dot(const string& filename) throw(CMRSPException)
{
    ofstream ofs(filename.c_str(), ofstream::out);
    if (!ofs)
        throw CMRSPException("Error guardando archivo: "+filename);
    
    ofs << "digraph CMRS {" << endl;
    
    set<WeightedArc>::iterator it;
    for (it = ring_connections.begin(); it != ring_connections.end(); it++)
    {
        Arc arc = it->first;
        double weight = it->second;
        //ofs << arc.first << " -> " << arc.second << "[label=" << setprecision(2) << ",color=blue];" << endl;
        ofs << arc.first << " -> " << arc.second << "[color=blue];" << endl;
    }
    
    for (it = star_connections.begin(); it != star_connections.end(); it++)
    {
        Arc arc = it->first;
        double weight = it->second;
        //ofs << arc.first << " -> " << arc.second << "[label=" << setprecision(2) << weight << ",color=green];" << endl;
        ofs << arc.first << " -> " << arc.second << "[color=green];" << endl;
    }
    
    ofs << "}" << endl;
    ofs.close();
}

void CMRS::print()
{
    cout << "----------------" << endl
            << "Ring connections:" << endl;
    set<WeightedArc>::iterator it;
    for (it = ring_connections.begin(); it != ring_connections.end(); it++)
    {
        Arc arc = it->first;
        double weight = it->second;
        cout << arc.first << " -> " << arc.second << " (" << weight << ")" << endl;
    }
    
    cout << "Star connections:" << endl;
    for (it = star_connections.begin(); it != star_connections.end(); it++)
    {
        Arc arc = it->first;
        double weight = it->second;
        cout << arc.first << " -> " << arc.second << " (" << weight << ")" << endl;
    }
    cout << "----------------" << endl;
}
