
/******************************************************************************
 * 
 * Capacitated m-Ring Star Problem
 * 
 * Instancia del problema de m-anillos estrella con capacidades
 * 
 ******************************************************************************/

#include <iostream>
#include <fstream>
#include <cmath>
#include <map>
#include <ctime>
#include "CMRSP.h"

#define __DEBUG
#define __MSG_ERR_LEN 4096

using namespace std;

CPXENVptr env;
CPXLPptr cmrsp_model;
int n_rings;
int ring_capacity;
int n_customers;
int n_steiners;
string depot0_label;
string depot1_label;
vector<string> customers_labels;
vector<string> steiners_labels;
int _V, _V1, _V2;
vector<vector<double> > connection_cost, routing_cost;
const static float machine_eps = 0.000001;
double optim_time;
int cuts1_applied = 0;
int cuts4_applied = 0;
ExecParam param;

typedef struct var_index
{
    int D0;
    int D1;
    int CUSTOMER;
    int STEINER;
} var_index;

var_index index_x, index_y, index_z, index_f, index_p, index_connection_cost, index_routing_cost;


CMRSP::CMRSP(const ExecParam& parameters) throw(CMRSPException)
{
    /***************************************************************************
    * Inicialización del entorno CPLEX
    ***************************************************************************/
    env = NULL;
    int status = 0;
    
    env = CPXopenCPLEX (&status);
    
    if (env == NULL)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        throw CMRSPException("Error incializando el entorno de CPLEX"+string(errmsg));
    }
    
   /***************************************************************************
    * Setea los parametros de optimización de CPLEX
    ***************************************************************************/
    param = parameters;
    
    //status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
    
    CPXsetintparam(env, CPX_PARAM_MIPINTERVAL, param.logging_interval);
    
    if (param.initial_heuristic) CPXsetintparam (env, CPX_PARAM_ADVIND, 2); // Considera la solucion inicial del usuario
    
    if (!param.clique_cuts) CPXsetintparam (env, CPX_PARAM_CLIQUES, -1);
    if (!param.cover_cuts) CPXsetintparam (env, CPX_PARAM_COVERS, -1);
    if (!param.disjuntive_cuts) CPXsetintparam (env, CPX_PARAM_DISJCUTS, -1);
    if (!param.flowcover_cuts) CPXsetintparam (env, CPX_PARAM_FLOWCOVERS, -1);
    if (!param.flowpath_cuts) CPXsetintparam (env, CPX_PARAM_FLOWPATHS, -1);
    if (!param.fractional_cuts) CPXsetintparam (env, CPX_PARAM_FRACCUTS, -1);
    if (!param.gubcover_cuts) CPXsetintparam (env, CPX_PARAM_GUBCOVERS, -1);
    if (!param.mir_cuts) CPXsetintparam (env, CPX_PARAM_MIRCUTS, -1);
    
    // 0 [CPX_NODESEL_DFS] Depth-first search 
    // 1 [CPX_NODESEL_BESTBOUND] Best-bound search 
    // 2 [CPX_NODESEL_BESTEST] Best-estimate search 
    // 3 [CPX_NODESEL_BESTEST_ALT] Alternative best-estimate search
    CPXsetintparam (env, CPX_PARAM_NODESEL, param.node_sel);

    // 0 [CPX_MIPEMPHASIS_BALANCED] Balance optimality and feasibility 
    // 1 [CPX_MIPEMPHASIS_FEASIBILITY] Emphasize feasibility over optimality 
    // 2 [CPX_MIPEMPHASIS_OPTIMALITY] Emphasize optimality over feasibility 
    // 3 [CPX_MIPEMPHASIS_BESTBOUND] Emphasize moving best bound 
    // 4 [CPX_MIPEMPHASIS_HIDDENFEAS] Emphasize finding hidden feasible solutions 
    // CPXsetintparam (env, CPX_PARAM_MIPEMPHASIS, CPX_MIPEMPHASIS_BESTBOUND);
    CPXsetintparam (env, CPX_PARAM_MIPEMPHASIS, param.mip_emphasis);


    if (param.cutting_plane_1 || param.cutting_plane_4)
    {
        CPXsetintparam (env, CPX_PARAM_PRELINEAR, CPX_OFF);
        CPXsetintparam (env, CPX_PARAM_MIPCBREDLP, CPX_OFF);
        
        status = CPXsetcutcallbackfunc (env, mycutcallback, NULL);
        if (status)
        {
            char errmsg[__MSG_ERR_LEN];
            CPXgeterrorstring (env, status, errmsg);
            throw CMRSPException("Error seteando la funcion callback para la aplicación de cortes. "+string(errmsg));
        }
    }
    
   /***************************************************************************
    * Construcción de instancia del problema
    ***************************************************************************/
    cmrsp_model = NULL;
    cmrsp_model = CPXcreateprob (env, &status, "cmrsp_model");
    if (cmrsp_model == NULL)
    {
        throw CMRSPException("Error creando instancia del problema");
    }
    
}

CMRSP::~CMRSP()
{
    /***************************************************************************
    * Liberación de espacio de entorno CPLEX
    ***************************************************************************/
    int status;
    if (cmrsp_model != NULL)
        status = CPXfreeprob (env, &cmrsp_model);
    
    status = CPXcloseCPLEX (&env);

    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cerr << "Error cerrando entorno de CPLEX" << endl;
        cerr << errmsg << endl;
    }
    
}


void CMRSP::optimize() throw(CMRSPException)
{
    /**************************************************************************
     * Construye la instancia CPLEX del problema
     **************************************************************************/
    #ifdef __DEBUG
    cout << "Construyendo instancia del problema " << param.filename << endl;
    #endif

    create_instance(param.filename);
    
    /**************************************************************************
     * Busca una solución con una heurística inicial y la agrega al problema
     **************************************************************************/
    CPXFILEptr logfile = create_logfile(param.filename);
    
    time_t t_start = time(NULL);
    if (param.initial_heuristic)
    { 
        #ifdef __DEBUG
        cout << param.filename << ": heuristica inicial" << endl;
        #endif
        initial_heuristic();
    }
    
    /**************************************************************************
     * Optimización
     **************************************************************************/
    #ifdef __DEBUG
    cout << param.filename << ": optimizando" << endl;
    #endif
    mip_optimize();
    
    time_t t_end = time(NULL);
    optim_time = difftime(t_end, t_start);
    
    CPXfclose (logfile);
    
    /**************************************************************************
     * Obtiene el resultado de la optimización
     **************************************************************************/
    #ifdef __DEBUG
    cout << param.filename << ": obteniendo solucion" << endl;
    #endif
    mip_solution();
    
}


CMRS CMRSP::solution() throw(CMRSPException)
{
    /**************************************************************************
     * Obtiene la solucion encontrada por CPLEX
     **************************************************************************/
    int cur_numrows = CPXgetnumrows(env, cmrsp_model);
    int cur_numcols = CPXgetnumcols(env, cmrsp_model);
    
    double*  x = (double *) malloc (cur_numcols * sizeof(double));
    
    int status = CPXgetmipx(env, cmrsp_model, x, 0, cur_numcols-1);
    if (status)
    {
        cerr << "Error obteniendo la solucion" << endl;
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cerr << errmsg << endl;
        CPXcloseCPLEX(&env);
        free(x);
        throw CMRSPException(errmsg);
    }
    
    CMRS cmrs = solution(x);
    free(x);
    
    return cmrs;
}
   
CMRS CMRSP::solution(double* x)
{
    /**************************************************************************
     * Construye un CMRS optimo con la solución devuelta por CPLEX
     **************************************************************************/
    CMRS cmrs(n_rings, ring_capacity);
    
    // agrega el depósito
    cmrs.depot(depot0_label+"_"+depot1_label);

    // agrega clientes
    for (int i = 0; i < customers_labels.size(); i++)
    {
        cmrs.add_customer(customers_labels[i]);
    }

    // agrega nodos de steiner
    for (int i = 0; i < steiners_labels.size(); i++)
    {
        cmrs.add_steiner(steiners_labels[i]);
    }
    
    // agrega arcos de conexiones
    string source_label, target_label;
    for (int i = 0; i < _V; i++)
    {
        for (int j = 0; j < _V; j++)
        {
            if (x[var_x(i, j)] > machine_eps)
            {
                source_label = label_x(0, i);
                target_label = label_x(1, j);
                if ((source_label.compare(depot0_label) == 0) || (source_label.compare(depot1_label) == 0))
                    source_label = cmrs.get_depot();
                if ((target_label.compare(depot0_label) == 0) || (target_label.compare(depot1_label) == 0))
                    target_label = cmrs.get_depot();
                cmrs.add_ring_connection(source_label, target_label);
            }
        }
    }
    
    // agrega arcos de ruteo
    for (int i = 0; i < n_customers; i++)
    {
        for (int j = 0; j < n_customers + n_steiners; j++)
        {            
            if (x[var_y(i, j)] > machine_eps)
            {
                source_label = label_y(0, i);
                target_label = label_y(1, j);
                cmrs.add_star_connection(source_label, target_label);
            }
        }
    }
    
    return cmrs;
}


int var_x(int i, int j)
{
    return i*_V + j;
}


int var_y(int i, int j)
{
    return _V2 + i*_V1 + j;
}


int var_z(int i)
{
    return _V2 + n_customers*_V1 + i;
}


int var_f(int i)
{
    return _V2 + n_customers*_V1 + _V1 + i;
}

int var_p(int i)
{
    return _V2 + n_customers*_V1 + _V1 + _V1 + i;
}


string label_x(int index, int i)
{
    if (i >= 2 + n_customers)
        return steiners_labels[i - 2 - n_customers];
    else if (i >= 2)
        return customers_labels[i - 2];
    else if (i == 1)
        return depot1_label;
    else if (i == 0)
        return depot0_label;
}


string label_y(int index, int i)
{
    if (i < n_customers)
        return customers_labels[i];
    else if ((index == 1) && (i >= n_customers))
        return steiners_labels[i - n_customers];
}


void CMRSP::create_instance(const string& filename) throw(CMRSPException)
{
    /***************************************************************************
     * Lee los parametros de la instancia:
     * cantidad de anillos, capacidad de anillo, matrices de costos de conexión
     * y matrices de costo de ruteo
     ***************************************************************************/
    load_parameters(filename);
    
    /***************************************************************************
     * Construye la estructura CPLEX de la instancia
     ***************************************************************************/
    build_instance();
}


void CMRSP::load_parameters(const string& filename) throw(CMRSPException)
{
    /***************************************************************************
     * 
     * Lee los parametros de la instancia:
     * cantidad de anillos, capacidad de anillo, matrices de costos de conexión
     * y de costos de ruteo
     * 
     ***************************************************************************/
    int i, j;

    ifstream ifs(filename.c_str());
    if (!ifs)
        throw CMRSPException("Error abriendo archivo: "+filename);
        
    ifs >> n_rings;         // cantidad de anillos
    ifs >> ring_capacity;   // capacidad de cada anillo
    ifs >> n_customers;     // cantidad de nodos cliente
    ifs >> n_steiners;      // cantidad de nodos de transicion
    ifs >> depot0_label;    // nombre del nodo de deposito (salida/inicial)
    ifs >> depot1_label;    // nombre del nodo de deposito (llegada/final)

    _V = n_customers + n_steiners + 2;
    _V1 = n_customers + n_steiners;
    _V2 = _V*_V;
    
    customers_labels = vector<string>(n_customers);   // nombres de los nodos cliente
    for (i = 0; i < n_customers; i++)
        ifs >> customers_labels[i];
    
    steiners_labels = vector<string>(n_steiners);    // nombres de los nodos de transicion
    for (i = 0; i < n_steiners; i++)
        ifs >> steiners_labels[i];
    
    connection_cost = vector<vector<double> >(_V);    // matriz de costos de conexion
    for (i = 0; i < _V; i++)
    {
        connection_cost[i] = vector<double>(_V);
        for (j = 0; j < _V; j++)
        {
            ifs >> connection_cost[i][j];
        }
    }

    routing_cost = vector<vector<double> >(n_customers);        // matriz de costos de ruteo
    for (i = 0; i < n_customers; i++)
    {
        routing_cost[i] = vector<double>(_V1);
        for (j = 0; j < _V1; j++)
        {
            ifs >> routing_cost[i][j];
        }
    }
        
    ifs.close();
}


void CMRSP::build_instance() throw(CMRSPException)
{
    /***************************************************************************
     *
     *  Construye la estructura CPLEX de la formulación para la instancia
     *
     ***************************************************************************/
    vector<string> nodo = vector<string>(_V);
    int numvars_x = _V2;
    int numvars_y = n_customers*_V1;
    int numvars_z = _V1;
    int numvars_f = _V1;
    int numvars_p = _V1;
    int numvars = numvars_x + numvars_y + numvars_z + numvars_f + numvars_p;
    int index, cindex, row, row1, row2, col, col1, col2;
    int numrows_x = _V;
    int numcols_x = _V;
    int numrows_y = n_customers;
    int numcols_y = _V1;
    int numcnstrs_2 = 1;
    int numcnstrs_3 = 1;
    int numcnstrs_4 = _V;
    int numcnstrs_5 = n_customers;
    int numcnstrs_6 = n_customers;
    int numcnstrs_7 = n_steiners;
    int numcnstrs_8 = _V1;
    int numcnstrs_9 = n_customers*_V1;
    int numcnstrs_10 = _V1;
    int numcnstrs_11 = _V1;
    int numcnstrs_12 = _V1*_V1;
    int numcnstrs_13 = n_customers;
    int numcnstrs_14 = n_steiners;
    int numcnstrs_15 = _V1;
    int numcnstrs_16 = n_customers*_V1;
    int numcnstrs_17 = n_steiners*_V1;
    // constraints para rompimiento de simetria    
    int numcnstrs_18 = _V1;      // p
    int numcnstrs_19 = _V1*(_V1 - 1);
    int numcnstrs_20 = _V1;
    
    int numcnstrs = numcnstrs_2 + numcnstrs_3 + numcnstrs_4 +    // Cantidad total de restricciones
                    numcnstrs_5 + numcnstrs_6 + numcnstrs_7 + 
                    numcnstrs_8 + numcnstrs_9 + numcnstrs_10 + 
                    numcnstrs_11 + numcnstrs_12 + numcnstrs_13 + 
                    numcnstrs_14 + numcnstrs_15 + numcnstrs_16 + numcnstrs_17 +
                    numcnstrs_18 + numcnstrs_19 + numcnstrs_20;
    int numnz_2 = _V1;                              // Cantidad de elementos no cero en restricciones de la familia 2
    int numnz_3 = _V1;                              // Cantidad de elementos no cero en restricciones de la familia 3
    int numnz_4 = _V;                               // Cantidad de elementos no cero en restricciones de la familia 4
    int numnz_5 = n_customers;                      // Cantidad de elementos no cero en restricciones de la familia 5
    int numnz_6 = (2*_V1 + 1)*n_customers;          // Cantidad de elementos no cero en restricciones de la familia 6
    int numnz_7 = (_V1 + 1)*n_steiners;             // Cantidad de elementos no cero en restricciones de la familia 7
    int numnz_8 = 2*(_V1 + 1)*_V1;                  // Cantidad de elementos no cero en restricciones de la familia 8
    int numnz_9 = n_customers*_V1*_V;               // Cantidad de elementos no cero en restricciones de la familia 9
    int numnz_10 = _V1;                             // Cantidad de elementos no cero en restricciones de la familia 10
    int numnz_11 = _V1;                             // Cantidad de elementos no cero en restricciones de la familia 11
    int numnz_12 = 3*_V1*_V1;                       // Cantidad de elementos no cero en restricciones de la familia 12
    int numnz_13 = n_customers*(n_customers + 1);   // Cantidad de elementos no cero en restricciones de la familia 13
    int numnz_14 = n_steiners*(n_customers + 1);    // Cantidad de elementos no cero en restricciones de la familia 14
    int numnz_15 = _V1;                             // Cantidad de elementos no cero en restricciones de la familia 15
    int numnz_16 = (3 + n_customers)*n_customers*_V1;   // Cantidad de elementos no cero en restricciones de la familia 16
    int numnz_17 = (3 + n_customers)*n_steiners*_V1;    // Cantidad de elementos no cero en restricciones de la familia 17
    int numnz_18 = 2*_V1;
    int numnz_19 = 3*_V1*(_V1 - 1);
    int numnz_20 = 2*_V1;
    int numnz = numnz_2 + numnz_3  + numnz_4 + numnz_5 + // Cantidad total de elementos no cero
                numnz_6 + numnz_7 + numnz_8 + numnz_9 + 
                numnz_10 + numnz_11 + numnz_12 + numnz_13 + 
                numnz_14 + numnz_15 + numnz_16 + numnz_17 +
                numnz_18 + numnz_19 + numnz_20;
    double obj[numvars];
    char ctype[numvars];
    double lb[numvars];
    double ub[numvars];
    int rmatbeg[numcnstrs];
    int rmatind[numnz];
    double rmatval[numnz];
    double rhs[numcnstrs];
    char sense[numcnstrs];
    int u, u1, v, v1, v2, w;
    int name_size = 20;
    char* colname[numvars];
    int status;
    int i, j;

    /* Inicializacion de indices. 
     * Usados para obtener los coeficientes de las variables en la matriz de restricciones
     */    
    index_x.D0 = 0;
    index_x.D1 = 1;
    index_x.CUSTOMER = 2;
    index_x.STEINER = 2 + n_customers;
    index_y.CUSTOMER = 0;
    index_y.STEINER = n_customers;
    index_z.CUSTOMER = 0;
    index_z.STEINER = n_customers;
    index_f.CUSTOMER = 0;
    index_f.STEINER = n_customers;
    index_p.CUSTOMER = 0;
    index_p.STEINER = n_customers;
    index_connection_cost.D0 = 0;
    index_connection_cost.D1 = 1;
    index_connection_cost.CUSTOMER = 2;
    index_connection_cost.STEINER = 2 + n_customers;
    index_routing_cost.CUSTOMER = 0;
    index_routing_cost.STEINER = n_customers;
    
    /* Etiqueta los nodos del grafo
     */
    nodo[0] = depot0_label;
    nodo[1] = depot1_label;
    for (int i = 0; i < n_customers; i++) nodo[2+i] = customers_labels[i];
    for (int i = 0; i < n_steiners; i++) nodo[2+n_customers+i] = steiners_labels[i];
    
    /* (1) Objetivo: 
        \min \sum_{v \in V} \sum_{v' \in V} c_{v, v'} x_{v, v'} + \sum_{u \in U} \sum_{v' \in V'} d_{u, v'} y_{u, v'} 
    */
    index = 0;
    for (i = 0; i < numrows_x; i++)     // x[v1,v2]
    {
        for (j = 0; j < numcols_x; j++)
        {
            obj[index] = connection_cost[i][j];
            lb[index] = 0.0;
            ub[index] = 1.0;
            
            string varname = "x_" + nodo[i] + "_" + nodo[j];
            colname[index] = new char[name_size];
            strlcpy(colname[index], varname.c_str(), name_size);
            
            index++;
        }
    }
    
    for (i = 0; i < numrows_y; i++)             // y[v1,v2]
    {
        for (j = 0; j < numcols_y; j++)
        {
            obj[index] = routing_cost[i][j];
            lb[index] = 0.0;
            ub[index] = 1.0;
            
            string varname = "y_" + nodo[i+2] + "_" + nodo[j+2];
            colname[index] = new char[name_size];
            strlcpy(colname[index], varname.c_str(), name_size);
            
            index++;
        }
    }
    
    for (i = 0; i < _V1; i++)         // z[v]
    {
        obj[index] = 0.0;
        lb[index] = 0.0;
        ub[index] = _V1 - 1;
        
        string varname = "z_" + nodo[i+2];
        colname[index] = new char[name_size];
        strlcpy(colname[index], varname.c_str(), name_size);
        
        index++;
    }

    for (i = 0; i < _V1; i++)         // f[v]
    {
        obj[index] = 0.0;
        lb[index] = 0.0;
        ub[index] = ring_capacity;
        
        string varname = "f_" + nodo[i+2];
        colname[index] = new char[name_size];
        strlcpy(colname[index], varname.c_str(), name_size);
        
        index++;        
    }
    
    for (i = 0; i < _V1; i++)     // p[v]
    {
        obj[index] = 0.0;
        lb[index] = 0.0;
        ub[index] = _V1;
        
        string varname = "p_" + nodo[i+2];
        colname[index] = new char[name_size];
        strlcpy(colname[index], varname.c_str(), name_size);
        
        index++;
    }
    
    status = CPXnewcols(env, cmrsp_model, numvars, obj, lb, ub, NULL, colname);
    if (status) throw CMRSPException("Error construyendo la instancia del problema (obj)");

    for (i = 0; i < numvars_x + numvars_y; i++)   // vector de variables binarias
        ctype[i] = CPX_BINARY;
    
    for (i = numvars_x + numvars_y; i < numvars_x + numvars_y + numvars_z + numvars_f; i++) // vector de variables enteras
        if (param.continuous_vars) ctype[i] = CPX_CONTINUOUS;
        else ctype[i] = CPX_INTEGER;
    
    for (i = numvars_x + numvars_y + numvars_z + numvars_f; i < numvars_x + numvars_y + numvars_z + numvars_f + numvars_p; i++) 
        ctype[i] = CPX_INTEGER;
    
    CPXcopyctype (env, cmrsp_model, ctype);
    
    /* Inicializacion de indices
     */
    index = 0;  // indice de columna
    cindex = 0; // indice de constraint
    rmatbeg[cindex] = 0;
    
    /* (2)
        \sum_{v \in V'} x_{d_0, v} = m
    */
    col = index_x.CUSTOMER;
    for (i = 0; i < n_customers + n_steiners; i++)
    {
        rmatind[index] = var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        col++;
        index++;
    }
    sense[cindex] = 'E';
    rhs[cindex] = n_rings;
    cindex++;
    
    /* (3)
        \sum_{v \in V'} x_{v,d_1} = m
    */
    rmatbeg[cindex] = index;
    col = index_x.CUSTOMER;
    for (i = 0; i < n_customers + n_steiners; i++)
    {
        rmatind[index] = var_x(col, index_x.D1);
        rmatval[index] = 1.0;
        col++;
        index++;
    }
    sense[cindex] = 'E';
    rhs[cindex] = n_rings;
    cindex++;
    
    /* (4)
        x_{v, v} = 0 \forall v \in V
    */
    col = index_x.D0;
    for (v = 0; v < _V; v++)
    {
        rmatbeg[cindex] = index;
        
        rmatind[index] = var_x(col, col);
        rmatval[index] = 1.0;
        index++;
        
        sense[cindex] = 'E';
        rhs[cindex] = 0.0;
        cindex++;
        
        col++;
    }
    
    /* (5)
        y_{u, u} = 0 \forall u \in U
    */
    col = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        rmatbeg[cindex] = index;
        
        rmatind[index] = var_y(col, col);;
        rmatval[index] = 1.0;
        index++;
        
        sense[cindex] = 'E';
        rhs[cindex] = 0.0;
        cindex++;
        
        col++;
    }
    
    /* (6)
        \sum_{v' \in V' \cup \{d_0\}} x_{v',u} + \sum_{v' \in V'} y_{u,v'} = 1 \forall u \in U
    */  
    col = index_x.CUSTOMER;
    col1 = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        rmatbeg[cindex] = index;
        
        rmatind[index] = var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        index++;
        row = index_x.CUSTOMER;
        row1 = index_y.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatind[index] = var_x(row, col);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = var_y(col1, row1);
            rmatval[index] = 1.0;
            index++;
            
            row++;
            row1++;
        }
        sense[cindex] = 'E';
        rhs[cindex] = 1.0;
        col++;
        col1++;
        cindex++;
    }

    /* (7)
        \sum_{v' \in V' \cup \{d_0\}} x_{v',w} \leq 1 \forall w \in W
    */
    col = index_x.STEINER;
    for (w = 0; w < n_steiners; w++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        index++;
        
        row = index_x.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatind[index] = var_x(row, col);
            rmatval[index] = 1.0;
            index++;
            row++;
        }
        sense[cindex] = 'L';
        rhs[cindex] = 1.0;
        cindex++;
        col++;       
    }

    /* (8)
        \sum_{v'' \in V' \cup \{d_0\}} x_{v'',v'} =  \sum_{v'' \in V' \cup \{d_1\}} x_{v',v''} \forall v' \in V'
    */
    col = index_x.CUSTOMER;
    for (v1 = 0; v1 < _V1; v1++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        index++;
        row = index_x.CUSTOMER;
        for (v2 = 0; v2 < _V1; v2++)
        {
            rmatind[index] = var_x(row, col);
            rmatval[index] = 1.0;
            index++;
            row++;
        }
        
        row = col;
        col = index_x.CUSTOMER;
        rmatind[index] = var_x(row, index_x.D1);
        rmatval[index] = -1.0;
        index++;
        for (v2 = 0; v2 < _V1; v2++)
        {
            rmatind[index] = var_x(row, col);
            rmatval[index] = -1.0;
            index++;
            col++;         
        }
        col = row;
        col++;
        sense[cindex] = 'E';
        rhs[cindex] = 0.0;
        cindex++;
    }
    
    /* (9)
        y_{u, v'} \leq \sum_{v \in V' \cup \{d_0\}} x_{v, v'} \forall u \in U \forall v' \in V'
    */
    row = index_x.CUSTOMER;
    row1 = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        col = index_x.CUSTOMER;
        col1 = index_y.CUSTOMER;
        for (v1 = 0; v1 < _V1; v1++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = var_y(row1, col1);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = var_x(index_x.D0, col);
            rmatval[index] = -1.0;
            index++;
            
            row2 = index_x.CUSTOMER;
            for (v = 0; v < _V1; v++)
            {
                rmatind[index] = var_x(row2, col);
                rmatval[index] = -1.0;
                index++;
                row2++;
            }
            sense[cindex] = 'L';
            rhs[cindex] = 0.0;
            cindex++;
            col++;
            col1++;
        }
        row++;
        row1++;
    }
    
    /* (10)
        z_{v'} \geq 1 \forall v' \in V'
    */
    col = index_z.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_z(col);
        rmatval[index] = 1.0;
        index++;
        col++;
    
        sense[cindex] = 'G';
        rhs[cindex] = 1.0;
        cindex++;
    }
    
    /* (11)
        z_{v'} \leq \vert V' \vert - 1 \forall v' \in V'
    */
   col = index_z.CUSTOMER;
   for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_z(col);
        rmatval[index] = 1.0;
        index++;
        col++;
    
        sense[cindex] = 'L';
        rhs[cindex] = _V1 - 1;
        cindex++;
    }
    
    /* (12)
        z_{v'} \geq (z_{v} + 1) - (\vert V' \vert - 1)(1 - x_{v, v'}) \forall v, v' \in V'
    */
    row = index_x.CUSTOMER;
    row1 = index_z.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        col = index_x.CUSTOMER;
        col1 = index_z.CUSTOMER;
        for (v1 = 0; v1 < _V1; v1++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = var_z(col1);
            rmatval[index] = 1.0;
            index++;
        
            rmatind[index] = var_z(row1);
            rmatval[index] = -1.0;
            index++;
    
            rmatind[index] = var_x(row, col);
            rmatval[index] = 1 - _V1;
            index++;
    
            sense[cindex] = 'G';
            rhs[cindex] = 2 - _V1;
            cindex++;
            col++;
            col1++;
        }
        row++;
        row1++;
    }
    
    /* (13)
        f_{u'} \geq 1 + \sum_{u \in U} y_{u, u'} \forall u' \in U
    */
    col = index_y.CUSTOMER;
    col1 = index_f.CUSTOMER;
    for (u1 = 0; u1 < n_customers; u1++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_f(col1);
        rmatval[index] = 1.0;
        index++;
        
        row = index_y.CUSTOMER;
        for (u = 0; u < n_customers; u++)
        {
            rmatind[index] = var_y(row, col);
            rmatval[index] = -1.0;
            index++;
            row++;
        }
        sense[cindex] = 'G';
        rhs[cindex] = 1.0;
        cindex++;
        col++;
        col1++;
    }
    
    /* (14)
        f_{w'} \geq \sum_{u \in U} y_{u, w'} \forall w' \in W
    */
    col = index_y.STEINER;
    col1 = index_f.STEINER;
    for (w = 0; w < n_steiners; w++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_f(col1);
        rmatval[index] = 1.0;
        index++;
        row = index_y.CUSTOMER;
        for (u = 0; u < n_customers; u++)
        {
            rmatind[index] = var_y(row, col);
            rmatval[index] = -1.0;
            index++;
            row++;
        }
        sense[cindex] = 'G';
        rhs[cindex] = 0.0;
        cindex++; 
        col++;
        col1++;
    }
    
    /* (15)
        f_{v'} \leq Q \forall v' \in V'
    */
    col = index_f.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_f(col);
        rmatval[index] = 1.0;
        index++;
        
        sense[cindex] = 'L';
        rhs[cindex] = ring_capacity;
        cindex++;
        col++;
    }

    /* (16)
        f_{u} \geq (f_{v'} + 1 + \sum_{u' \in U} y_{u', u}) - \vert U \vert(1 - x_{v', u}) \forall u \in U \forall v' \in V'
    */
    col = index_x.CUSTOMER;
    col1 = index_f.CUSTOMER;
    col2 = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        row = index_x.CUSTOMER;
        row1 = index_f.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = var_f(col1);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = var_f(row1);
            rmatval[index] = -1.0;
            index++;
            
            rmatind[index] = var_x(row, col);
            rmatval[index] = -n_customers;
            index++;
            
            row2 = index_y.CUSTOMER;
            for (u1 = 0; u1 < n_customers; u1++)
            {
                rmatind[index] = var_y(row2, col2);
                rmatval[index] = -1.0;
                index++;
                row2++;
            }
            
            sense[cindex] = 'G';
            rhs[cindex] = 1 - n_customers;
            cindex++;
            row++;
            row1++;
        }
        
        col++;
        col1++;
        col2++;
    }
    
    /* (17)
        f_{w} \geq (f_{v'} + \sum_{u' \in U} y_{u', w}) - \vert U \vert(1 - x_{v', w}) \forall w \in W \forall v' \in V'
    */
    col = index_x.STEINER;
    col1 = index_f.STEINER;
    col2 = index_y.STEINER;
    for (w = 0; w < n_steiners; w++)
    {
        row = index_x.CUSTOMER;
        row1 = index_f.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = var_f(col1);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = var_f(row1);
            rmatval[index] = -1.0;
            index++;
            
            rmatind[index] = var_x(row, col);
            rmatval[index] = -n_customers;
            index++;
            
            row2 = index_y.CUSTOMER;
            for (u1 = 0; u1 < n_customers; u1++)
            {
                rmatind[index] = var_y(row2, col2);
                rmatval[index] = -1.0;
                index++;
                row2++;
            }

            sense[cindex] = 'G';
            rhs[cindex] = -n_customers;
            cindex++;
            row++;
            row1++;
            row2++;
        }
        col++;
        col1++;
        col2++;
    }
    

    /* (18)
        p_{v_{j}} \geq j*x_{d0, v_{j}} \forall v_j \in V'
    */
cout << "18..." << endl;
    row = index_p.CUSTOMER;
    col1 = index_x.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_p(row);
        rmatval[index] = 1.0;
        index++;
        
        rmatind[index] = var_x(index_x.D0, col1);
        rmatval[index] = -v - 1;
        index++;
        
        sense[cindex] = 'G';
        rhs[cindex] = 0.0;
        cindex++;
        row++;
        col1++;
    }
    
    /* (19) 
        p_{j} \geq p_{i} + (-M)(1 - x_{v_{i}, v_{j}}) \forall v_{i}, v_{j} \in V' con M = |U|+|W|
    */

cout << "19..." << endl;
    for (v = 0; v < _V1; v++)
    {
        row = index_p.CUSTOMER+v;
        row1 = index_x.CUSTOMER+v;
        for (v1 = 0; v1 < _V1; v1++)
        {            
            if (v1 == v) continue;
            
            col = index_p.CUSTOMER+v1;
            col1 = index_x.CUSTOMER+v1;
            
            rmatbeg[cindex] = index;
            rmatind[index] = var_p(row);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = var_p(col);
            rmatval[index] = -1.0;
            index++;

            rmatind[index] = var_x(row1, col1);
            rmatval[index] = _V1;
            index++;
            
            sense[cindex] = 'G';
            rhs[cindex] = -_V1;
            cindex++;
        }
    }
    
    /* (20) 
        p_{j} \leq j*x_{v_{j}, d1} + M(1 - x_{v_{j}, d1}) \forall v_{j} \in V' con M = |U|+|W|
    */
cout << "20..." << endl;
    row = index_p.CUSTOMER;
    row1 = index_x.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = var_p(row);
        rmatval[index] = 1.0;
        index++;
        
        rmatind[index] = var_x(row1, index_x.D1);
        rmatval[index] = _V1;
        index++;

        sense[cindex] = 'L';
        rhs[cindex] = _V1 + v + 1;
        cindex++;
        row++;
        row1++;
    }
    
    

    /***************************************************************************
     * Agrega las constraints en el modelo
     ***************************************************************************/
cout << "agregando restricciones..." << endl;
    status = CPXaddrows(env, cmrsp_model, 0, cindex, index, rhs, sense, rmatbeg, rmatind, rmatval, NULL, NULL /*rowname*/);
    
cout << "restricciones agregadas..." << endl;
    if (status) throw CMRSPException("Error construyendo la instancia del problema (constraints)");
    
    /***************************************************************************
     * Agrega los conjuntos SOS1
     ***************************************************************************/
    if (param.sos)
    {
        int numsos = n_customers + n_steiners;
        int numsosnz_1 = n_customers*(2*_V1 + 1);
        int numsosnz_2 = n_steiners*_V1;
        int numsosnz = numsosnz_1 + numsosnz_2;
        char sostype[numsos];
        int sosbeg[numsos];
        int sosind[numsosnz];
        double soswt[numsosnz];
        
        for (i = 0; i < numsos; i++)
        {
            sostype[numsos] = CPX_TYPE_SOS1;
        }
        
        /*
            SOS1 que se deduce de (6)
            { x{d0, u}, x{v'[1],u}, ..., x{v'[n_customers+n_steiners],u}, y{u,v'[1]}, ..., y{u,v'[n_customers+n_steiners]} } para cada cliente u es un SOS1
        */
        index = 0;
        cindex = 0;
        for (u = 0; u < n_customers; u++)
        {
            sosbeg[cindex] = index;
            cindex++;
            col = index_x.CUSTOMER;
            row = index_x.CUSTOMER;
            row1 = index_y.CUSTOMER;
            col1 = index_y.CUSTOMER;
            for (v = 0; v < _V1; v++)
            {
                sosind[index] = var_x(row, col);
                soswt[index] = 0.1*v;
                row++;
                index++;
                
                sosind[index] = var_y(row1, col1);
                soswt[index] = 0.1*_V1 + 0.1*v;
                col1++;
                index++;
            }
            
            sosind[index] = var_x(index_x.D0, col);
            soswt[index] = 0.1*_V1*2 + 0.1;
            index++;
        }
        
        /*
            SOS1 que se deduce de (7)
            { x{v'[1],w}, ..., x{v'[n_customers+n_steiners],w} } para cada nodo de steiner w
        */
        for (w = 0; w < n_steiners; w++)
        {
            sosbeg[cindex] = index;
            cindex++;
            row = index_x.CUSTOMER;
            col = index_x.STEINER;
            for (v = 0; v < n_customers + n_steiners; v++)
            {
                sosind[index] = var_x(row, col);
                soswt[index] = 0.1*v;
                row++;
                index++;
            }
        }
        
        status = CPXcopysos(env, cmrsp_model, numsos, numsosnz, sostype, sosbeg, sosind, soswt, NULL);
        
        if (status) 
        {
            char errmsg[__MSG_ERR_LEN];
            CPXgeterrorstring (env, status, errmsg);
            throw CMRSPException("Error construyendo la instancia del problema (SOS1) "+string(errmsg));
        }
    }
    
    /***************************************************************************
     * Temporal: guarda el modelo en un archivo con formato lp para verificar 
     * si fue creado correctamente
     ***************************************************************************/
    
    status = CPXwriteprob(env, cmrsp_model, "test.lp", NULL);
    if (status) 
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        throw CMRSPException("Error grabando el lp del problema"+string(errmsg));
    }

}


CPXFILEptr CMRSP::create_logfile(const string& filename) throw(CMRSPException)
{
    CPXFILEptr logfile = CPXfopen ((filename+".cplex.log").c_str(), "a");
    int status = CPXsetlogfile (env, logfile);
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        CPXcloseCPLEX (&env);
        throw CMRSPException("Error creando archivo de log");
    }
    return logfile;
}

void CMRSP::mip_optimize() throw(CMRSPException)
{
    #ifdef __DEBUG
    cout << param.filename << ": CPXmipopt inicio" << endl;
    #endif

    int status = CPXmipopt (env, cmrsp_model);
    
    #ifdef __DEBUG
    cout << param.filename << ": CPXmipopt fin" << endl;
    #endif
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        #ifdef __DEBUG
        cout << param.filename << ": " << string(errmsg) << endl;
        #endif
        CPXcloseCPLEX (&env);
        throw CMRSPException("Error optimizando "+string(errmsg));
    }
}

void CMRSP::mip_solution() throw(CMRSPException)
{
    double objval;
    int solstat = CPXgetstat(env, cmrsp_model);
    int status = CPXgetmipobjval(env, cmrsp_model, &objval);
    
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        CPXcloseCPLEX (&env);
        throw CMRSPException("Error obteniendo el valor de la función objetivo"+string(errmsg));
    }
    
    cout << endl << "solstat = " << solstat << ", objval = " << objval << endl;    
}


void CMRSP::initial_heuristic() throw(CMRSPException)
{
    if (n_rings == 1)
        initial_heuristic1();
    else
        initial_heuristic3();
}


void CMRSP::initial_heuristic1() throw(CMRSPException)
{
    /**************************************************************************
     * Heurística inicial (versión 1):
     * 
     * Se construye anillo por anillo (hasta no terminar la construcción de
     * un anillo, no se comienza la construcción de otro).
     * La construcción de los anillos se raliza desde el depóstio.
     * 
     * La cantidad diferencia en la cantidad de clientes entre dos anillos
     * no es más que uno. 
     * Es decir, algunos anillos tienen floor(n_customers/n_rings) clientes
     * y n_customers%n_rings anillos tienen floor(n_customers/n_rings)+1.
     * En los primeros (n_customers%n_rings) anillos se asigna un cliente 
     * más que en los ultimos.
     * 
     * Se elige primero el de menor costo con deposito, se
     * lo conecta al deposito, luego se elige el de menor costo con el ultimo que
     * se agregó, registrando que clientes ya fueron visitados para no
     * volver a visitarlos. Cuando se alcanza la cantidad de clientes
     * asignados al anillo, el ultimo cliente agregado se conecta al deposito.
     * Se continua asi con el siguiente anillo. 
     * (Se asume n_customers >= n_rings)
     * 
     **************************************************************************/
    int n_ring_customers = (int) floor(n_customers/n_rings);
    int n_ring_customers_rem = n_customers % n_rings;
    int customers_left = n_customers;
    vector<bool> customer_processed = vector<bool>(n_customers);
    vector<vector<int> > rings;
    double solution_cost = 0.0;
    int numnz = n_customers + n_rings;
    
    for (int i = 0; i < n_customers; i++) customer_processed[i] = false;
    
    
    while (customers_left > 0)
    {
        int customers_ring = 0;
        vector<int> ring;
        
        int last_node = index_connection_cost.D0;
        int customers_in_ring = n_ring_customers;
        if (rings.size() < n_ring_customers_rem) customers_in_ring++; // agrego 1 del resto (porque la cantidad de clientes puede no ser múltiplo de m)
        while (customers_ring < customers_in_ring && customers_left > 0)
        {
            int best_node;
            double min_cost = DBL_MAX;
            
            for (int next_node = index_connection_cost.CUSTOMER; next_node < index_connection_cost.CUSTOMER + n_customers; next_node++)
            {
                if (!customer_processed[next_node - index_connection_cost.CUSTOMER])
                {
                    double cost = connection_cost[last_node][next_node];
                    if (cost < min_cost)
                    {
                        best_node = next_node;
                        min_cost = cost;
                    }
                }                
            }
            
            customer_processed[best_node - index_connection_cost.CUSTOMER] = true;
            ring.push_back(best_node);
            last_node = best_node;
            customers_ring++;
            customers_left--;
            solution_cost += min_cost;           
        }
        solution_cost += connection_cost[last_node][index_connection_cost.D1];
        rings.push_back(ring);
    }
    
    for (int r = 0; r < rings.size(); r++)
    {
        for (int cust = 0; cust < rings[r].size(); cust++)
        {
            cout << customers_labels[rings[r][cust]-index_connection_cost.CUSTOMER] << endl;
        }
    }
    
    
    /**************************************************************************
     * Aplicando la solución como solución inicial en la instancia CPLEX
     **************************************************************************/
    // parametro de CPLEX para que considere la solucion inicial
    CPXsetintparam (env, CPX_PARAM_ADVIND, 2);
    
    int indices[numnz];
    double value[numnz];
    int index = 0;
    
    for (int r = 0; r < rings.size(); r++)
    {
        // arcos anillo
        int last_node = index_connection_cost.D0;
        for (int cust = 0; cust < rings[r].size(); cust++)
        {
            int next_node = rings[r][cust];
            indices[index] = var_x(last_node, next_node);
            value[index] = 1.0;
            last_node = next_node;
            index++;
        }
        indices[index] = var_x(last_node, index_connection_cost.D1);
        value[index] = 1.0;
        index++;
    }
    cout << "index = " << index << ", numnz = " << numnz << endl;
    
    int status = CPXcopymipstart(env, cmrsp_model, numnz, indices, value);
    
    if (status)
    {
        cerr << "Error seteando la solución inicial" << endl;
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cerr << errmsg << endl;
        CPXcloseCPLEX(&env);
        throw CMRSPException(errmsg);
    }

}


void CMRSP::initial_heuristic2() throw(CMRSPException)
{
    /**************************************************************************
     * Heurística inicial (versión 2):
     * 
     * Se construyen los anillos de manera simultanea. 
     * La construcción de los anillos se raliza desde el depóstio.
     * 
     * Se buscan los n_rings clientes mas cercanos al deposito. 
     * Se realizan n_rings conexiones, desde el deposito con cada uno
     * de estos clientes. Es decir, cada uno de estos clientes se asgina
     * a un anillo distinto. Estos clientes se marcan como ultimos visitados
     * de cada anillo. 
     * 
     * Mientras haya clientes sin visitar. Se busca el cliente sin visitar
     * que tiene menor costo con los ultimos visitados (de los distintos anillos 
     * que no estan cerrados o terminados). Se conecta el cliente a ese 
     * ultimo cliente visitado. Una vez que un anillo alcanza los ring_capacity
     * clientes, se toma el ultimo cliente y se lo conecta al deposito. 
     * 
     **************************************************************************/
    set<int> not_visited = set<int>();
    vector<vector<int> > rings = vector<vector<int> >(n_rings);
    vector<double> mins = vector<double>(n_rings);
    vector<int> last_nodes = vector<int>(n_rings);
    int customers_left = n_customers;
    int last_node, next_node;
    
    for (int cust = index_connection_cost.CUSTOMER; cust < index_connection_cost.CUSTOMER + n_customers; cust++)
        not_visited.insert(cust);
    
    for (int i = 0; i < n_rings; i++) { mins[i] = DBL_MAX; last_nodes[i] = -1; }    
    
    last_node = index_connection_cost.D0;
    for (int next_node = index_connection_cost.CUSTOMER; next_node < index_connection_cost.CUSTOMER + n_customers; next_node++)
    {
        for (int j = 0; j < n_rings; j++)
        {
            double cost = connection_cost[last_node][next_node];
            if (cost < mins[j])
            {
                for (int k = n_rings-1; k > j; k--)
                {
                    mins[k] = mins[k-1];
                    last_nodes[k] = last_nodes[k-1];
                }
                mins[j] = cost;
                last_nodes[j] = next_node;
                break;
            }
        }
    }
    
    for (int i = 0; i < n_rings; i++)
    {
        rings[i].push_back(last_nodes[i]);
        not_visited.erase(last_nodes[i]);
    }
    
    int best_ring, best_node;
    double min_cost;
    while (!not_visited.empty())
    {
        min_cost = DBL_MAX;
        set<int>::const_iterator it;
        for (it = not_visited.begin(); it != not_visited.end(); it++)
        {
            next_node = *it;
            for (int i = 0; i < n_rings; i++)
            {
                if (rings[i].size() < ring_capacity)
                {
                    double cost = connection_cost[last_nodes[i]][next_node];
                    if (cost < min_cost)
                    {
                        best_ring = i;
                        best_node = next_node;
                        min_cost = cost;
                    }
                }
            }
        }
        rings[best_ring].push_back(best_node);
        not_visited.erase(best_node);
        last_nodes[best_ring] = best_node;
    }
    
    cout << "Resultado de heuristíca inicial" << endl;
    double solution_cost = 0.0;
    for (int i = 0; i < n_rings; i++)
    {
        last_node = index_connection_cost.D0;
        cout << "anillo " << i << endl;
        for (int cust = 0; cust < rings[i].size(); cust++)
        {
            solution_cost += connection_cost[last_node][rings[i][cust]];
            last_node = rings[i][cust];
            cout << customers_labels[rings[i][cust]-index_connection_cost.CUSTOMER] << endl;
        }
        solution_cost += connection_cost[last_node][index_connection_cost.D1];
    }
    cout << "costo = " << solution_cost << endl;
    
    /**************************************************************************
     * Aplicando la solución como solución inicial en la instancia CPLEX
     **************************************************************************/    
    int numnz = n_customers + n_rings;
    int indices[numnz];
    double value[numnz];
    int index = 0;
    
    for (int r = 0; r < rings.size(); r++)
    {
        last_node = index_connection_cost.D0;
        for (int cust = 0; cust < rings[r].size(); cust++)
        {
            next_node = rings[r][cust];
            indices[index] = var_x(last_node, next_node);
            value[index] = 1.0;
            last_node = next_node;
            index++;
        }
        indices[index] = var_x(last_node, index_connection_cost.D1);
        value[index] = 1.0;
        index++;
    }
    cout << "index = " << index << ", numnz = " << numnz << endl;
    
    int status = CPXcopymipstart(env, cmrsp_model, numnz, indices, value);
    
    if (status)
    {
        cerr << "Error seteando la solución inicial" << endl;
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cerr << errmsg << endl;
        CPXcloseCPLEX(&env);
        throw CMRSPException(errmsg);
    }

}


void CMRSP::initial_heuristic3() throw(CMRSPException)
{
    /**************************************************************************
     * Heurística inicial (versión 3):
     * 
     * Busqueda de clientes iniciales
     * Construccion simultanea de anillos
     * 
     **************************************************************************/    
    
    if (n_rings > 1)
    {
        /**********************************************************************
         * Busqueda heurística de n_rings clientes iniciales alejados
         **********************************************************************/
        double max_cost = DBL_MIN;
        vector<int> initials = vector<int>(n_rings);
        set<int> not_visited;
        int cust1, cust2;
        
        for (int cust = index_connection_cost.CUSTOMER; cust < index_connection_cost.CUSTOMER + n_customers; cust++)
            not_visited.insert(cust);

        for (int i = index_connection_cost.CUSTOMER; i < index_connection_cost.CUSTOMER + n_customers; i++)
        {
            for (int j = index_connection_cost.CUSTOMER; j < index_connection_cost.CUSTOMER + n_customers; j++)
            {
                if (i != j)
                {
                    double cost = connection_cost[i][j];
                    if (cost > max_cost)
                    {
                        cust1 = i;
                        cust2 = j;
                        max_cost = cost;
                    }
                }
            }
        }
        
        initials[0] = cust1;
        initials[1] = cust2;
        int n_initials = 2;
        not_visited.erase(cust1);
        not_visited.erase(cust2);
        while (n_initials < n_rings)
        {
            vector<pair<double, int> > costs_sum = vector<pair<double, int> >(n_customers - n_initials);
            vector<pair<double, int> > costs_var = vector<pair<double, int> >(n_customers - n_initials);
            vector<pair<double, int> > costs_var_orig = vector<pair<double, int> >(n_customers - n_initials);
            map<int, int> customer_index = map<int, int>();
            
            int k = 0;
            for (int candidate_customer = index_connection_cost.CUSTOMER; candidate_customer < index_connection_cost.CUSTOMER + n_customers; candidate_customer++)
            {
                if (find(initials.begin(), initials.end(), candidate_customer) == initials.end())
                {
                    /* Costos entre los clientes iniciales que ya agregué y el candidato */
                    double cost_sum = 0.0, cost_var = 0.0, cost_avg;
                    
                    for (int j = 0; j < n_initials; j++)
                    {
                        cost_sum += connection_cost[initials[j]][candidate_customer];
                    }
                    cost_avg = cost_sum / n_initials;
                    
                    for (int j = 0; j < n_initials; j++)
                    {
                        cost_var += (connection_cost[initials[j]][candidate_customer] - cost_avg)*(connection_cost[initials[j]][candidate_customer] - cost_avg);
                    }
                    
                    customer_index[candidate_customer] = k;
                    costs_sum[k] = pair<double, int> (cost_sum, candidate_customer);
                    costs_var[k] = pair<double, int> (cost_var, candidate_customer);
                    costs_var_orig[k] = pair<double, int> (cost_var, candidate_customer);
                    k++;
                }
            }
            
            sort(costs_sum.begin(), costs_sum.end());
            sort(costs_var.begin(), costs_var.end());
            
            // busco que en promedio/suma sea grande y que en varianza o desvios sea suficientemente pequeño 
            // (>=cost_var[n_initials/2] mayor o igual que la mediana de los desvios)
            double max_var = costs_var[(n_customers - n_initials)/2].first;

            int cust;
            for (int i = 0; i < n_customers - n_initials; i++)
            {
                cust = costs_sum[n_customers - n_initials - 1 - i].second;
                
                if (costs_var_orig[cust].first < max_var) break;
            }
            
            initials[n_initials] = cust;
            not_visited.erase(cust);
            n_initials++;
        }
        
        cout << "Iniciales:" << endl;
        for (int i = 0; i < initials.size(); i++)
        {
            cout << customers_labels[initials[i] - index_connection_cost.CUSTOMER] << endl;
        }

        cout << "---------" << endl;

        /**********************************************************************
         * Construcción de anillos a partir de los clientes iniciales
         **********************************************************************/
        vector<vector<int> > rings = vector<vector<int> >(n_rings);
        vector<int> heads = vector<int>(n_rings);
        vector<int> tails = vector<int>(n_rings);
        vector<vector<pair<int, int> > > attachs = vector<vector<pair<int, int> > >(n_rings);
        vector<int> attachs_exists = vector<int>(n_customers);
        
        // inicialmente los anillos contienen los clientes iniciales
        for (int i = 0; i < n_rings; i++)
        {
            rings[i].push_back(initials[i]);
            heads[i] = initials[i];
            tails[i] = initials[i];
            attachs[i] = vector<pair<int, int> >();
        }
        
        // inicialmente ningun customer tiene otros adjuntos
        for (int i = 0; i < n_customers; i++) attachs_exists[i] = false;
        
        while (!not_visited.empty())
        {
            // se busca el cliente no visitado y anillo sin completar su capacidad lo mas cercano posible
            set<int>::const_iterator new_cust_it;
            int best_cust, best_ring;
            double cost, min_cost = DBL_MAX;
            bool head_connection;
            
            for (new_cust_it = not_visited.begin(); new_cust_it != not_visited.end(); new_cust_it++)
            {
                int new_cust = *new_cust_it;
                
                for (int ring = 0; ring < n_rings; ring++)
                {
                    if (rings[ring].size() < ring_capacity)
                    {
                        cost = connection_cost[new_cust][tails[ring]];
                        if (cost < min_cost)
                        {
                            best_cust = new_cust;
                            best_ring = ring;
                            head_connection = false;
                            min_cost = cost;
                        }
                        
                        cost = connection_cost[heads[ring]][new_cust];
                        if (cost < min_cost)
                        {
                            best_cust = new_cust;
                            best_ring = ring;
                            head_connection = true;
                            min_cost = cost;
                        }
                    }
                }
            }
            
            if (head_connection)
            {
                if (rings[best_ring].size() > 1)
                {
                    // conexion actual: ... -> cust_last2 -> cust_last1
                    // best_cust es el cliente con menor costo de conexion a cust_last1
                    int cust_last1 = rings[best_ring][rings[best_ring].size() - 1];
                    int cust_last2 = rings[best_ring][rings[best_ring].size() - 2];
                    
                    // se analizan las siguientes alternativas:
                    // 1) ... -> cust_last2 -> cust_last1 -> best_cust (anillo)
                    // 2) ... -> cust_last2 -> best_cust (anillo)  +   cust_last1 -> cust_last2 (arco estrella)
                    // 3) ... -> cust_last2 -> best_cust (anillo)  +   cust_last1 -> best_cust (arco estrella)
                    // para las alternativas 2) y 3) no pueden haber clientes adjuntos a cust_last1 (arcos estrella conectados a cust_last1)
                    int best_cust_route = best_cust - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_last1_route = cust_last1 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_last2_route = cust_last2 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;

                    double cost1 = connection_cost[cust_last2][cust_last1] + connection_cost[cust_last1][best_cust];
                    double cost2 = connection_cost[cust_last2][best_cust] + routing_cost[cust_last1_route][cust_last2_route];
                    double cost3 = connection_cost[cust_last2][best_cust] + routing_cost[cust_last1_route][best_cust_route];
                    
                    // elige una mejor alternativa de las 3
                    if ( attachs_exists[cust_last1_route - index_routing_cost.CUSTOMER] || ( (cost1 <= cost2) && (cost1 <= cost3) ) )
                    {
                        rings[best_ring].push_back(best_cust);
                    }
                    else if ((cost2 <= cost1) && (cost2 <= cost3))
                    {
                        rings[best_ring].pop_back();
                        rings[best_ring].push_back(best_cust);
                        attachs[best_ring].push_back(pair<int, int>(cust_last1_route, cust_last2_route));
                        attachs_exists[cust_last2_route - index_routing_cost.CUSTOMER] = true;
                    }
                    else // (cost3 <= cost1) && (cost3 <= cost2)
                    {
                        rings[best_ring].pop_back();
                        rings[best_ring].push_back(best_cust);
                        attachs[best_ring].push_back(pair<int, int>(cust_last1_route, best_cust_route));
                        attachs_exists[best_cust_route - index_routing_cost.CUSTOMER] = true;
                    }
                }
                else
                {
                    rings[best_ring].push_back(best_cust);
                }
                heads[best_ring] = best_cust;
            }
            else // !head_connection
            {
                if (rings[best_ring].size() > 1)
                {
                    // conexion actual: cust_fist1 -> cust_first2 -> ...
                    // best_cust es el cliente con menor costo de conexion a cust_first1
                    int cust_first1 = rings[best_ring][0];
                    int cust_first2 = rings[best_ring][1];
                    
                    // se analizan las siguientes alternativas:
                    // 1) best_cust -> cust_first1 -> cust_first2 -> ... (anillo)
                    // 2) best_cust -> cust_first2 -> ... (anillo)  +   cust_first1 -> cust_first2 (arco estrella)
                    // 3) best_cust -> cust_first2 -> ... (anillo)  +   cust_first1 -> best_cust (arco estrella)
                    // para las alternativas 2) y 3) no pueden haber clientes adjuntos a cust_first1 (arcos estrella conectados a cust_first1)
                    int best_cust_route = best_cust - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_first1_route = cust_first1 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_first2_route = cust_first2 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;

                    double cost1 = connection_cost[best_cust][cust_first1] + connection_cost[cust_first1][cust_first2];
                    double cost2 = connection_cost[best_cust][cust_first2] + routing_cost[cust_first1_route][cust_first2_route];
                    double cost3 = connection_cost[best_cust][cust_first2] + routing_cost[cust_first1_route][best_cust_route];

                    // elige una mejor alternativa de las 3
                    if (attachs_exists[cust_first1_route - index_routing_cost.CUSTOMER] || ((cost1 <= cost2) && (cost1 <= cost3)))
                    {
                        rings[best_ring].insert(rings[best_ring].begin(), best_cust);
                    }
                    else if ((cost2 <= cost1) && (cost2 <= cost3))
                    {
                        rings[best_ring][0] = best_cust;
                        attachs[best_ring].push_back(pair<int, int>(cust_first1_route, cust_first2_route));
                        attachs_exists[cust_first1_route - index_routing_cost.CUSTOMER] = true;
                    }
                    else // (cost3 <= cost1) && (cost3 <= cost2)
                    {
                        rings[best_ring][0] = best_cust;
                        attachs[best_ring].push_back(pair<int, int>(cust_first1_route, best_cust_route));
                        attachs_exists[best_cust_route - index_routing_cost.CUSTOMER] = true;
                    }
                }
                else
                {
                    rings[best_ring].insert(rings[best_ring].begin(), best_cust);
                }
                tails[best_ring] = best_cust;
            }
            
            not_visited.erase(best_cust);
        }
        
        
        /**********************************************************************
         * Solucion generada por la heurística
         **********************************************************************/
        for (int ring = 0; ring < n_rings; ring++)
        {
            cout << "anillo " << ring << endl;
            for (int cust = 0; cust < rings[ring].size(); cust++)
            {
                cout << "\t" << customers_labels[rings[ring][cust] - index_connection_cost.CUSTOMER] << endl;
            }
            
            cout << "adjuntos: " << attachs[ring].size() << endl;
            for (int cust = 0; cust < attachs[ring].size(); cust++)
            {
/*                cout << index_routing_cost.CUSTOMER << " " << attachs[ring][cust].first << " -> " 
                        << attachs[ring][cust].second << endl;*/
                cout << attachs[ring][cust].first - index_routing_cost.CUSTOMER << " " << customers_labels[attachs[ring][cust].first - index_routing_cost.CUSTOMER] << " -> " 
                        << attachs[ring][cust].second - index_routing_cost.CUSTOMER << " " << customers_labels[attachs[ring][cust].second - index_routing_cost.CUSTOMER] << endl;
            }
        }
        
        cout << "fin" << endl;
        /**********************************************************************
         * Aplicando la solución como solución inicial en la instancia CPLEX
         **********************************************************************/    
        int numnz = n_customers + n_rings;
        int indices[numnz];
        double value[numnz];
        int index = 0;
        int last_node, next_node;
        
        for (int r = 0; r < n_rings; r++)
        {
            // arcos anillo
            last_node = index_connection_cost.D0;
            for (int cust = 0; cust < rings[r].size(); cust++)
            {
                next_node = rings[r][cust];
                indices[index] = var_x(last_node, next_node);
                value[index] = 1.0;
                last_node = next_node;
                index++;
            }
            indices[index] = var_x(last_node, index_connection_cost.D1);
            value[index] = 1.0;
            index++;
        
            // arcos estrella
            for (int cust = 0; cust < attachs[r].size(); cust++)
            {
                int source = attachs[r][cust].first;
                int target = attachs[r][cust].second;
                indices[index] = var_y(source, target);
                value[index] = 1.0;
                index++;
            }
        }
        cout << "index = " << index << ", numnz = " << numnz << endl;
        
        int status = CPXcopymipstart(env, cmrsp_model, numnz, indices, value);
        
        if (status)
        {
            cerr << "Error seteando la solución inicial" << endl;
            char errmsg[__MSG_ERR_LEN];
            CPXgeterrorstring (env, status, errmsg);
            cerr << errmsg << endl;
            CPXcloseCPLEX(&env);
            throw CMRSPException(errmsg);
        }
    
    }
    
    
}

void CMRSP::save_log() throw(CMRSPException)
{
    ofstream ofs;
    ofs.open((param.filename+".log").c_str(), ios_base::app);
    if (!ofs) throw CMRSPException("Error grabando log.");
    
    ofs << param.filename 
        << " " << n_rings                      
        << " " << ring_capacity
        << " " << n_customers
        << " " << n_steiners 
        << " " << param.initial_heuristic
        << " " << param.node_sel
        << " " << param.mip_emphasis
        << " " << param.clique_cuts 
        << " " << param.cover_cuts 
        << " " << param.disjuntive_cuts 
        << " " << param.flowcover_cuts 
        << " " << param.flowpath_cuts 
        << " " << param.fractional_cuts 
        << " " << param.gubcover_cuts 
        << " " << param.mir_cuts 
        << " " << param.cutting_plane_1 
        << " " << param.cutting_plane_2 
        << " " << param.cutting_plane_3 
        << " " << param.cutting_plane_4 
        << " " << optim_time 
        << endl;    
    
    ofs.close();
}

int CPXPUBLIC CMRSP::mycutcallback (CPXCENVptr env, void *cbdata, int wherefrom, void *cbhandle, int *useraction_p)
{    
    // cout << "ejecutando mycutcallback ..." << endl;
    int cur_numrows = CPXgetnumrows(env, cmrsp_model);
    int cur_numcols = CPXgetnumcols(env, cmrsp_model);
    *useraction_p = CPX_CALLBACK_DEFAULT;
    int status = 0;
    int node_count, nodes_left, mip_iterations;
    double best_integer;
    
    if (/*wherefrom != CPX_CALLBACK_MIP && wherefrom != CPX_CALLBACK_MIP_BRANCH && wherefrom != CPX_CALLBACK_MIP_INCUMBENT &&
        wherefrom != CPX_CALLBACK_MIP_NODE && wherefrom != CPX_CALLBACK_MIP_HEURISTIC && */wherefrom != CPX_CALLBACK_MIP_CUT)
    {
        cout << "no se aplican cortes (wherefrom invalido)" << endl; // no deberia suceder (?)
        return status;
    }
    
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_NODE_COUNT, &node_count);
    if (status)
    {
        cout << "Error recuperando la cantidad de nodos procesados" << endl;
        return status;
    }
    
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_NODES_LEFT, &nodes_left);
    if (status)
    {
        cout << "Error recuperando la cantidad de nodos por procesar" << endl;
        return status;
    }
        
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_MIP_ITERATIONS, &mip_iterations);
    if (status)
    {
        cout << "Error recuperando la cantidad de iteraciones" << endl;
        return status;
    }
    
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &best_integer);
    if (status)
    {
        cout << "Error recuperando la mejor solucion entera encontrada hasta el momento" << endl;
        return status;
    }        
    
    time_t now = time(NULL);
    
    if (node_count % 100 == 0) 
    {
        cout << ctime(&now) << ": procesados = " << node_count << ", activos = " << nodes_left << ", iter = " << mip_iterations << ", best int = " << best_integer;
        if (node_count <= 10) cout << ": buscando cortes..." << endl;
        else cout << endl;
    }
    
    if (node_count > 1000)
    {
        return status;
    }

    
    double* x = (double *) malloc (cur_numcols * sizeof(double));
    status = CPXgetcallbacknodex (env, cbdata, wherefrom, x, 0, cur_numcols-1);
    if (status)
    {
        cout << "Error recuperando la solucion del nodo actual" << endl;
        *useraction_p = CPX_CALLBACK_DEFAULT;
        return status;
    }
    
    // solution(x).print();
    // cout << "Buscando cortes... ";
    bool cuts_1 = false, cuts_4 = false;
    
    if (param.cutting_plane_1) cuts_1 = apply_cuts1(x, cbdata, wherefrom);
    if (param.cutting_plane_4) cuts_4 = apply_cuts4(x, cbdata, wherefrom);
        
    if (cuts_1 || cuts_4) 
    {
        // cout << "Se aplicaron cortes" << endl;
        *useraction_p = CPX_CALLBACK_SET;
    }
    
    free(x);
    
    return status;
}

bool CMRSP::apply_cuts1(double* x, void *cbdata, int wherefrom) throw(CMRSPException)
{    
    /**************************************************************************
     * Busqueda de las componentes conexas que contienen clientes
     * aislados del deposito
     **************************************************************************/
    int status;
    int solstat;
    double objval;
    bool cuts_generated = false;
    bool cuts_added = false;
    vector<vector<int> > components;
    vector<int> n_component_customers;
    int index_x_customer = index_x.CUSTOMER;
    int index_y_customer = index_y.CUSTOMER;
    set<int> not_visited_customers;
    vector<bool> visited = vector<bool>(n_customers + n_steiners);
    int cust, n_cust;
    int x_source, x_target, y_source, y_target;
    
    for (int i = 0; i < n_customers; i++)
        not_visited_customers.insert(i);
    
    for (int i = 0; i < n_customers+n_steiners; i++)
        visited[i] = false;
    
    while (!not_visited_customers.empty())
    {
        vector<int> component = vector<int>();
        vector<int> visit_stack = vector<int>();
        cust = *not_visited_customers.begin();
        
        component.push_back(cust); n_cust = 1;
        not_visited_customers.erase(cust);
        visited[cust] = true;
        visit_stack.push_back(cust);
        while (visit_stack.size() > 0)
        {
            bool depot = false;
            
            int source = visit_stack.back();
            visit_stack.pop_back();
            
            x_source = index_x_customer + source;
            y_source = index_y_customer + source;
            for (int target = 0; target < _V1; target++)
            {
                x_target = index_x_customer + target;
                y_target = index_y_customer + target;
                
                if (!visited[target])
                {
                    if ( (x[var_x(x_source, index_x.D1)] > machine_eps) || (x[var_x(index_x.D0, x_source)] > machine_eps) )
                        depot = true;
                    
                    /* verifica si esta conectado por medio de una conexion anillo o estrella */
                    if ( (x[var_x(x_source, x_target)] > machine_eps) || (x[var_x(x_target, x_source)] > machine_eps) ||
                         ( (source < n_customers) && (x[var_y(y_source, y_target)] > machine_eps) ) ||
                         ( (target < n_customers) && (x[var_y(y_target, y_source)] > machine_eps) )
                       )
                    {
                        component.push_back(target);
                        visit_stack.push_back(target);
                        if (target < n_customers)
                        {
                            not_visited_customers.erase(target);
                            n_cust++;
                        }
                        visited[target] = true;
                    }                    
                }
            }
            
            if (!depot) cuts_generated = true;
            // if (depot) cout << "el componente contiene al deposito" << endl; else cout << "el componente NO contiene al deposito" << endl;
        }
        
        if (component.size() > 1)
        {
            components.push_back(component);
            n_component_customers.push_back(n_cust);
        }
    }    
    
    /**************************************************************************
     * Agregando restricciones a la formulación
     **************************************************************************/
    if (cuts_generated)
    {
        vector<int> in_component(n_customers + n_steiners);
        for (int i = 0; i < n_customers + n_steiners; i++)
            in_component[i] = -1;
        
        for (int k = 0; k < components.size(); k++)
        {
            for (int i = 0; i < components[k].size(); i++)
            {
                in_component[components[k][i]] = k;
            }
        }
        
        if (components.size() > 1)
        {
            int numnz = 2*_V*_V1 + _V;
            
            for (int k = 0; k < components.size(); k++)
            {
                if (components[k].size() > 1)
                {
                    int rmatind[numnz];
                    double rmatval[numnz];
                    int rmatbeg[1];
                    double rhs_value;
                    int index = 0;
                    double lhs_value = 0.0;
                    rmatbeg[0] = index;
                    for (int i = 0; i < components[k].size(); i++)
                    {
                        for (int j = 0; j < _V1; j++)
                        {
                            if (in_component[j] != k)
                            {
                                //cout << "x_"; if (components[k][i] < n_customers) cout << customers_labels[components[k][i]]; else cout << steiners_labels[components[k][i] - n_customers]; cout << "_"; if (j < n_customers) cout << customers_labels[j]; else cout << steiners_labels[j - n_customers];cout << " ";
                                lhs_value += x[var_x(index_x_customer + components[k][i], index_x_customer + j)];
                                rmatind[index] = var_x(index_x_customer + components[k][i], index_x_customer + j);
                                rmatval[index] = 1.0;
                                index++;
                                
                                if (components[k][i] < n_customers) // i es cliente
                                {
                                    //cout << "y_" << customers_labels[components[k][i]]; cout << "_"; if (j < n_customers) cout << customers_labels[j]; else cout << steiners_labels[j - n_customers];cout << " ";
                                    lhs_value += x[var_y(index_y_customer + components[k][i], index_y_customer + j)];
                                    rmatind[index] = var_y(index_y_customer + components[k][i], index_y_customer + j);
                                    rmatval[index] = 1.0;
                                    index++;
                                }
                            }
                        }
                        
                        //cout << "x_"; if (components[k][i] < n_customers) cout << customers_labels[components[k][i]]; else cout << steiners_labels[components[k][i] - n_customers]; cout << "_D1";cout << " ";
                        lhs_value += x[var_x(index_x_customer + components[k][i], index_x.D1)];
                        rmatind[index] = var_x(index_x_customer + components[k][i], index_x.D1);
                        rmatval[index] = 1.0;
                        index++;
                        
                        if (i == components[k].size() - 1)
                        {
                            // cout << "lhs = " << lhs_value << ", rhs = " << rhs << endl;
                            rhs_value = ceil( ((double)n_component_customers[k])/ring_capacity );
                            if (rhs_value - lhs_value > machine_eps) 
                            {
                                // cout << "la restriccion corta la solucion" << endl; 
                                cuts_added = true;
                                cuts1_applied++;
                            }
                            else 
                            {
                                // cout << "la restriccion no corta la solucion" << endl;
                                continue;
                            }
                            // cout << ">=" << rhs << endl;
                            
                            // status = CPXaddrows(env, cmrsp_model, 0, 1 /*cindex*/, index, rhs, sense, rmatbeg, rmatind, rmatval, NULL, NULL /*rowname*/);
                            status = CPXcutcallbackadd (env, cbdata, wherefrom, index, rhs_value, 'G', rmatind, rmatval);
                            
                            if (status)
                            {
                                char errmsg[__MSG_ERR_LEN];
                                CPXgeterrorstring (env, status, errmsg);
                                throw CMRSPException("Error agregando restricciones de 1er familia: "+string(errmsg));
                            }
                            
//                             cout << "COMPONENTE: " << endl;
//                             vector<int> component = components[k];
//                             for (int i = 0; i < component.size(); i++)
//                             {
//                                 if (component[i] < n_customers)
//                                     cout << customers_labels[component[i]] << " ";
//                                 else
//                                     cout << steiners_labels[component[i]-n_customers] << " ";
//                             }
//                             cout << endl << "----------" << endl;
                        }
                        
                    }                    
                    
                }
            }
            
        }
    }
        
    return cuts_added;
}


bool CMRSP::apply_cuts4(double* x, void *cbdata, int wherefrom) throw(CMRSPException)
{    
    /**************************************************************************
     * Busqueda de las componentes conexas que contienen clientes
     * aislados del deposito
     **************************************************************************/
    int status;
    int solstat;
    double objval;
    bool cuts_generated = false;
    bool cuts_added = false;
    vector<vector<int> > components;
    vector<int> n_component_customers;
    int index_x_customer = index_x.CUSTOMER;
    int index_y_customer = index_y.CUSTOMER;
    set<int> not_visited_customers;
    vector<bool> visited = vector<bool>(n_customers + n_steiners);
    int cust, n_cust;
    int x_source, x_target, y_source, y_target;
    
    for (int i = 0; i < n_customers; i++)
        not_visited_customers.insert(i);
    
    for (int i = 0; i < n_customers+n_steiners; i++)
        visited[i] = false;
    
    while (!not_visited_customers.empty())
    {
        vector<int> component = vector<int>();
        vector<int> visit_stack = vector<int>();
        cust = *not_visited_customers.begin();
        
        component.push_back(cust); n_cust = 1;
        not_visited_customers.erase(cust);
        visited[cust] = true;
        visit_stack.push_back(cust);
        while (visit_stack.size() > 0)
        {
            bool depot = false;
            
            int source = visit_stack.back();
            visit_stack.pop_back();
            
            x_source = index_x_customer + source;
            y_source = index_y_customer + source;
            for (int target = 0; target < _V1; target++)
            {
                x_target = index_x_customer + target;
                y_target = index_y_customer + target;
                
                if (!visited[target])
                {
                    if ( (x[var_x(x_source, index_x.D1)] > machine_eps) || (x[var_x(index_x.D0, x_source)] > machine_eps) )
                        depot = true;
                    
                    /* verifica si esta conectado por medio de una conexion anillo o estrella */
                    if ( (x[var_x(x_source, x_target)] > machine_eps) || (x[var_x(x_target, x_source)] > machine_eps) ||
                         ( (source < n_customers) && (x[var_y(y_source, y_target)] > machine_eps) ) ||
                         ( (target < n_customers) && (x[var_y(y_target, y_source)] > machine_eps) )
                       )
                    {
                        component.push_back(target);
                        visit_stack.push_back(target);
                        if (target < n_customers)
                        {
                            not_visited_customers.erase(target);
                            n_cust++;
                        }
                        visited[target] = true;
                    }                    
                }
            }
            
            if (!depot) cuts_generated = true;
            // if (depot) cout << "el componente contiene al deposito" << endl; else cout << "el componente NO contiene al deposito" << endl;
        }
        
        if (component.size() > 1)
        {
            components.push_back(component);
            n_component_customers.push_back(n_cust);
        }
    }    
    
    /**************************************************************************
     * Agregando restricciones a la formulación
     **************************************************************************/
    if (cuts_generated)
    {
        vector<int> in_component(n_customers + n_steiners);
        for (int i = 0; i < n_customers + n_steiners; i++)
            in_component[i] = -1;
        
        for (int k = 0; k < components.size(); k++)
        {
            for (int i = 0; i < components[k].size(); i++)
            {
                in_component[components[k][i]] = k;
            }
        }
        
        if (components.size() > 1)
        {
            int numnz = 2*_V*_V1 + _V;
            
            for (int k = 0; k < components.size(); k++)
            {
                if (components[k].size() > 1)
                {
                    int rmatind[numnz];
                    double rmatval[numnz];
                    int rmatbeg[1];
                    double rhs_value;
                    int index = 0;
                    double lhs_value = 0.0;
                    rmatbeg[0] = index;
                    for (int i = 0; i < components[k].size(); i++)
                    {
                        for (int j = 0; j < _V1; j++)
                        {
                            if (in_component[j] != k)
                            {
                                //cout << "x_"; if (components[k][i] < n_customers) cout << customers_labels[components[k][i]]; else cout << steiners_labels[components[k][i] - n_customers]; cout << "_"; if (j < n_customers) cout << customers_labels[j]; else cout << steiners_labels[j - n_customers];cout << " ";
                                lhs_value += x[var_x(index_x_customer + components[k][i], index_x_customer + j)];
                                rmatind[index] = var_x(index_x_customer + components[k][i], index_x_customer + j);
                                rmatval[index] = 1.0;
                                index++;                            
                            }
                        }
                        
                        //cout << "x_"; if (components[k][i] < n_customers) cout << customers_labels[components[k][i]]; else cout << steiners_labels[components[k][i] - n_customers]; cout << "_D1";cout << " ";
                        lhs_value += x[var_x(index_x_customer + components[k][i], index_x.D1)];
                        rmatind[index] = var_x(index_x_customer + components[k][i], index_x.D1);
                        rmatval[index] = 1.0;
                        index++;
                        
                        if (i == components[k].size() - 1)
                        {
                            // cout << "lhs = " << lhs_value << ", rhs = " << rhs << endl;
                            rhs_value = ceil( ((double)n_component_customers[k])/ring_capacity );
                            if (rhs_value - lhs_value > machine_eps) 
                            {
                                // cout << "la restriccion corta la solucion" << endl; 
                                cuts_added = true;
                                cuts4_applied++;
                            }
                            else 
                            {
                                // cout << "la restriccion no corta la solucion" << endl;
                                continue;
                            }
                            // cout << ">=" << rhs << endl;
                            
                            // status = CPXaddrows(env, cmrsp_model, 0, 1 /*cindex*/, index, rhs, sense, rmatbeg, rmatind, rmatval, NULL, NULL /*rowname*/);
                            status = CPXcutcallbackadd (env, cbdata, wherefrom, index, rhs_value, 'G', rmatind, rmatval);
                            
                            if (status)
                            {
                                char errmsg[__MSG_ERR_LEN];
                                CPXgeterrorstring (env, status, errmsg);
                                throw CMRSPException("Error agregando restricciones de 1er familia: "+string(errmsg));
                            }
                            
//                             cout << "COMPONENTE: " << endl;
//                             vector<int> component = components[k];
//                             for (int i = 0; i < component.size(); i++)
//                             {
//                                 if (component[i] < n_customers)
//                                     cout << customers_labels[component[i]] << " ";
//                                 else
//                                     cout << steiners_labels[component[i]-n_customers] << " ";
//                             }
//                             cout << endl << "----------" << endl;
                        }
                        
                    }                    
                    
                }
            }
            
        }
    }
        
    return cuts_added;
}


// bool CMRSP::apply_cuts2(double* x, void *cbdata, int wherefrom) throw(CMRSPException)
// {    
//     /**************************************************************************
//      * Busqueda de ciclos compuestos por clientes y steiners
//      **************************************************************************/
//     bool cuts_added = false;
//     vector<vector<int> > circuits;
//     int index_x_customer = index_x.CUSTOMER;
//     int index_y_customer = index_y.CUSTOMER;
//     set<int> not_visited;
//     vector<int> visited(_V1);
//     vector<int> parent(_V1);
//     
//     for (int i = 0; i < _V1; i++)
//     {
//         not_visited.insert(i);
//         visited[i] = 0;
//         parent[i] = -1;
//     }
//     
//     while (!not_visited.empty())
//     {
//         vector<int> visit_stack;
//         
//         int current_node = *not_visited.begin();
//         visited[current_node] = -1;
//         not_visited.erase(current_node);
//         
//         bool circuit_detected = false;
//         visit_stack.push_back(current_node);
//         while (visit_stack.size() > 0)
//         {
//             current_node = visit_stack.back();
//             visited[current_node] = -1;
//             not_visited.erase(current_node);
//             visit_stack.pop_back();
//             
//             int x_current_node = index_x_customer + current_node;
//             int y_current_node = index_y_customer + current_node;
//             bool adjacents_found = false;
//             for (int adjacent_node = 0; adjacent_node < _V1; adjacent_node++)
//             {
//                 if ((parent[current_node] != adjacent_node) && (visited[adjacent_node] != 1))
//                 {
//                     int x_adjacent_node = index_x_customer + adjacent_node;
//                     int y_adjacent_node = index_y_customer + adjacent_node;
//                     if ( x[var_x(x_current_node, x_adjacent_node)] > machine_eps ||
//                          x[var_x(x_adjacent_node, x_current_node)] > machine_eps ||
//                          ( (y_current_node < n_customers) && (x[var_y(y_current_node, y_adjacent_node)] > machine_eps) ) ||
//                          ( (y_adjacent_node < n_customers) && (x[var_y(y_adjacent_node, y_current_node)] > machine_eps) )
//                     )
//                     {
//                         adjacents_found = true;
//                         visit_stack.push_back(adjacent_node);
//                         parent[adjacent_node] = current_node;
//                         if ((parent[current_node] != adjacent_node) && (visited[adjacent_node] == -1))
//                         {
//                             if (adjacent_node < n_customers) cout << customers_labels[adjacent_node] << " ";
//                             else cout << steiners_labels[adjacent_node - n_customers] << " ";
//                             circuit_detected = true;
//                             break;
//                         }
//                     }
//                     if (circuit_detected) break;
//                 }
//             }
//             
//             if (circuit_detected) break;            
//         }
//         
//         for (int i = 0; i < _V1; i++) if (visited[i] == -1) visited[i] = 1;
//         
//         vector<int> circuit;
//         if (circuit_detected)
//         {
//             int first_node = visit_stack.back();
//             int current_node = first_node;
//             int adjacent_node = parent[current_node];
//             circuit.push_back(current_node);
//             while (adjacent_node != first_node)
//             {
//                 circuit.push_back(adjacent_node);
//                 adjacent_node = parent[adjacent_node];
//             }
//             
//             circuits.push_back(circuit);
//         }
//         
//         
// //         if (circuit_detected)
// //         {        
// //             cout << "CIRCUITO: " << endl;
// //             for (int i = 0; i < circuit.size(); i++)
// //             {
// //                 if (circuit[i] < n_customers)
// //                     cout << customers_labels[circuit[i]] << " ";
// //                 else
// //                     cout << steiners_labels[circuit[i] - n_customers] << " ";
// //             }
// //             cout << endl;
// //         }
//     }
//     
//     /**************************************************************************
//      * Agregando restricciones a la formulación
//      **************************************************************************/
//     cout << "Lista de candidatos: " << endl;
//     for (int c = 0; c < circuits.size(); c++)
//     {
//         vector<int> circuit = circuits[c];
//         for (int i = 0; i < circuit.size(); i++)
//         {
//             if (circuit[i] < n_customers)
//                 cout << customers_labels[circuit[i]] << " ";
//             else
//                 cout << steiners_labels[circuit[i] - n_customers] << " ";
//         }
//         cout << endl;
//     }
//     
//     if (circuits.size() > 0)
//     {        
//         int numnz = 8*_V*_V*circuits.size();
//         int rmatbeg[0];
//         int rmatind[numnz];
//         double rmatval[numnz];
//         char sense[0];
//         
//         for (int i = 0; i < circuits.size(); i++)
//         {
//             int index = 0;
//             double rhs_value = 0.0;
//             double lhs_value;
//             
//             cout << "evaluando circuito " << i << endl;
//             rmatbeg[0] = index;
//             lhs_value = 0.0;
//             for (int j = 0; j < circuits[i].size() - 1; j++)
//             {
//                 lhs_value += x[var_x(index_x_customer + circuits[i][j], index_x_customer + circuits[i][j+1])];
//                 rmatind[index] = var_x(index_x_customer + circuits[i][j], index_x_customer + circuits[i][j+1]);
//                 rmatval[index] = 1.0;
//                 index++;
//                 
//                 lhs_value += x[var_x(index_x_customer + circuits[i][j+1], index_x_customer + circuits[i][j])];
//                 rmatind[index] = var_x(index_x_customer + circuits[i][j+1], index_x_customer + circuits[i][j]);
//                 rmatval[index] = 1.0;
//                 index++;
//                 
//                 if (circuits[i][j] < n_customers)
//                 {
//                     lhs_value += x[var_y(index_y_customer + circuits[i][j], index_y_customer + circuits[i][j+1])];
//                     rmatind[index] = var_y(index_y_customer + circuits[i][j], index_y_customer + circuits[i][j+1]);
//                     rmatval[index] = 1.0;
//                     index++;
//                 }
//                 
//                 if (circuits[i][j+1] < n_customers)
//                 {
//                     lhs_value += x[var_y(index_y_customer + circuits[i][j+1], index_y_customer + circuits[i][j])];
//                     rmatind[index] = var_y(index_y_customer + circuits[i][j+1], index_y_customer + circuits[i][j]);
//                     rmatval[index] = 1.0;
//                     index++;
//                 }
//             }
//             
//             lhs_value += x[var_x(index_x_customer + circuits[i][circuits[i].size()-1], index_x_customer + circuits[i][0])];
//             rmatind[index] = var_x(index_x_customer + circuits[i][circuits[i].size()-1], index_x_customer + circuits[i][0]);
//             rmatval[index] = 1.0;
//             index++;
//             
//             lhs_value += x[var_x(index_x_customer + circuits[i][0], index_x_customer + circuits[i][circuits[i].size()-1])];
//             rmatind[index] = var_x(index_x_customer + circuits[i][0], index_x_customer + circuits[i][circuits[i].size()-1]);
//             rmatval[index] = 1.0;
//             index++;
//             
//             if (circuits[i][circuits[i].size()-1] < n_customers)
//             {
//                 lhs_value += x[var_y(index_y_customer + circuits[i][circuits[i].size()-1], index_y_customer + circuits[i][0])];
//                 rmatind[index] = var_y(index_y_customer + circuits[i][circuits[i].size()-1], index_y_customer + circuits[i][0]);
//                 rmatval[index] = 1.0;
//                 index++;
//             }
//             
//             if (circuits[i][0] < n_customers)
//             {
//                 lhs_value += x[var_y(index_y_customer + circuits[i][0], index_y_customer + circuits[i][circuits[i].size()-1])];
//                 rmatind[index] = var_y(index_y_customer + circuits[i][0], index_y_customer + circuits[i][circuits[i].size()-1]);
//                 rmatval[index] = 1.0;
//                 index++;
//             }
//             
//             rhs_value = circuits[i].size() - 1;
//             cout << "lhs_value = " << lhs_value << endl;
//             if (lhs_value - rhs_value > machine_eps) cout << "la restriccion corta la solucion" << endl; else cout << "la restriccion no corta la solucion" << endl;
//             if (lhs_value - rhs_value > machine_eps)
//             {
//                 cuts_added = true;
//                 cuts2_applied++;
//                 // status = CPXaddrows(env, cmrsp_model, 0, 1 /*cindex*/, index, rhs, sense, rmatbeg, rmatind, rmatval, NULL, NULL /*rowname*/);
//                 status = CPXcutcallbackadd (env, cbdata, wherefrom, index, rhs_value, 'L', rmatind, rmatval);
//                 
//                 if (status)
//                 {
//                     char errmsg[__MSG_ERR_LEN];
//                     CPXgeterrorstring (env, status, errmsg);
//                     throw CMRSPException("Error agregando restricciones de 2da familia: "+string(errmsg));
//                 }
//             }
//         }
// 
//     }
//     
//     return cuts_added;
// }


// bool CMRSP::apply_cuts3() throw(CMRSPException)
// {
//     /**************************************************************************
//      * Obtiene la solucion encontrada por CPLEX
//      **************************************************************************/
//     bool cuts_added = false;
//     int cur_numrows = CPXgetnumrows(env, cmrsp_model);
//     int cur_numcols = CPXgetnumcols(env, cmrsp_model);
//     
//     int      solstat;
//     double   objval;
//     double*  x = (double *) malloc (cur_numcols * sizeof(double));
//     
//     int status = CPXgetx(env, cmrsp_model, x, 0, cur_numcols-1);
//     if (status)
//     {
//         cerr << "Error obteniendo la solucion" << endl;
//         char errmsg[__MSG_ERR_LEN];
//         CPXgeterrorstring (env, status, errmsg);
//         cerr << errmsg << endl;
//         CPXcloseCPLEX(&env);
//         free(x);
//         throw CMRSPException(errmsg);
//     }
//     
//     /**************************************************************************
//      * Busqueda de estructuras de arcos anillo y arcos estrella de la forma
//      * (u, v), (u', v), (u, u') con u y u' nodos cliente 
//      * y v nodos cliente o steiner
//      **************************************************************************/
//     vector<vector<int> > candidates;
//     int index_x_customer = index_x.CUSTOMER;
//     int index_y_customer = index_y.CUSTOMER;
//     int x_customer_i, x_customer_j, y_customer_i, y_customer_j, y_node_k;
//     
//     for (int i = 0; i < n_customers; i++)
//     {
//         x_customer_i = index_x_customer + i;
//         y_customer_i = index_y_customer + i;
//         
//         for (int j = 0; j < i; j++)
//         {
//             x_customer_j = index_x_customer + j;
//             y_customer_j = index_y_customer + j;
//             
//             for (int k = 0; k < _V1; k++)
//             {
//                 if ((k == i) || (k == j)) continue;
//                 
//                 y_node_k = index_y_customer + k;
//                 
//                 if (    ( x[var_y(y_customer_i, y_node_k)] > machine_eps ) &&
//                         ( x[var_y(y_customer_j, y_node_k)] > machine_eps ) &&
//                         ( (x[var_x(x_customer_i, x_customer_j)] > machine_eps ) ||  
//                           (x[var_x(x_customer_j, x_customer_i)] > machine_eps )    )    )
//                 {
//                     vector<int> candidate = vector<int>(3);
//                     candidate[0] = i; candidate[1] = j; candidate[2] = k;
//                     candidates.push_back(candidate);
//                 }
// 
//             }
//             
//         }
//         
//     }
//     
//     /**************************************************************************
//      * Aplicacion de cortes
//      **************************************************************************/
//     cout << "Lista de candidatos: " << endl;
//     for (int i = 0; i < candidates.size(); i++)
//     {
//         for (int j = 0; j < 3; j++)
//         {
//             if (candidates[i][j] < n_customers) cout << customers_labels[candidates[i][j]] << " ";
//             else cout << steiners_labels[candidates[i][j] - n_customers] << " ";
//             
//         }
//         cout << endl;
//     }
//     
//     
//     int numnz = 2*(_V1 + 1) + 1;
//     int rmatbeg[1];
//     int rmatind[numnz];
//     double rmatval[numnz];
//     double rhs[1];
//     char sense[1];
//     int index = 0, cindex = 0;
//     
//     double lhs_value;
//     int x_node_k, y_node_j;
//     
//     for (int i = 0; i < candidates.size(); i++)
//     {        
//         // analizar los casos:
//         // Debe ser: 1/2 suma(v'=0..V1+{d0}) x[v'][u] + suma(v'=0..V1+{d1}) x[u][v'] <= 1 - y[u][v]
//         // (j == 0) (caso 1) u = candidates[i][0] v = candidates[i][2] 
//         // (j == 1) (caso 2) u = candidates[i][1] v = candidates[i][2]
//         // lhs = (1/2 suma(v'=0..V1+{d0}) x[v'][u] + suma(v'=0..V1+{d1}) x[u][v']) + y[u][v]
//         // rhs = 1
//         for (int j = 0; j <= 1; j++)
//         {
//             lhs_value = 0.0;
//             index = 0;
//             
//             x_customer_j = index_x_customer + candidates[i][j];
//             y_customer_j = index_y_customer + candidates[i][j];
//             y_node_j = candidates[i][2];
//             
//             for (int k = 0; k < _V1; k++)
//             {
//                 x_node_k = index_x_customer + k;
//                 
//                 if (x[var_x(x_node_k, x_customer_j)] > machine_eps)
//                 {
//                     lhs_value += x[var_x(x_node_k, x_customer_j)];
//                     rmatind[index] = var_x(x_node_k, x_customer_j);
//                     rmatval[index] = 0.5;
//                     index++;
//                 }
//                 
//                 if (x[var_x(x_node_k, x_customer_j)] > machine_eps)
//                 {
//                     lhs_value += x[var_x(x_customer_j, x_node_k)];
//                     rmatind[index] = var_x(x_customer_j, x_node_k);
//                     rmatval[index] = 0.5;
//                     index++;
//                 }
//                 
//             }
//             
//             if (x[var_x(index_x.D0, x_customer_j)] > machine_eps)
//             {
//                 lhs_value += x[var_x(index_x.D0, x_customer_j)];
//                 rmatind[index] = var_x(index_x.D0, x_customer_j);
//                 rmatval[index] = 0.5;
//                 index++;
//             }
//             
//             if (x[var_x(x_customer_j, index_x.D1)] > machine_eps)
//             {
//                 lhs_value += x[var_x(x_customer_j, index_x.D1)];
//                 rmatind[index] = var_x(x_customer_j, index_x.D1);
//                 rmatval[index] = 0.5;
//                 index++;
//             }
//             
//             cout << "entrada + salida " << customers_labels[j] << " = " << lhs_value << endl;
//             cout << "y " << customers_labels[j] << "_* = " << x[var_y(y_customer_j, y_node_j)] << endl;
//             lhs_value = lhs_value / 2.0;
//             lhs_value += x[var_y(y_customer_j, y_node_j)];
//             rmatind[index] = var_y(y_customer_j, y_node_j);
//             rmatval[index] = 1.0;
//             index++;
//             
//             cout << "lhs_value = " << lhs_value << endl;
//             if (lhs_value > 1.0) // corta la solución
//             {
//                 cout << "la restriccion corta la solucion" << endl;
//                 cuts_added = true;
//                 sense[0] = 'L';
//                 rhs[0] = 1.0;
//                 
//                 status = CPXaddrows(env, cmrsp_model, 0, 1 /*cindex*/, index, rhs, sense, rmatbeg,
//                                     rmatind, rmatval, NULL, NULL /*rowname*/);
//                 
//                 if (status)
//                 {
//                     free(x);
//                     char errmsg[__MSG_ERR_LEN];
//                     CPXgeterrorstring (env, status, errmsg);
//                     throw CMRSPException("Error agregando restricciones de 3er familia: "+string(errmsg));
//                 }                
//             }
//             else
//             {
//                 cout << "la restriccion no corta la solucion" << endl;
//             }
//             
//         }
//         
//     }
//     
// }


// bool component_cut(CPXCENVptr env, void *cbdata, int wherefrom, void *cbhandle, int *useraction_p)
// {
//     bool add_cuts = false;
//     int numcols = CPXgetnumcols(env, cmrsp_model);
//     double*  x = new double[numcols];
//     
//     bool cuts_generated = false;
//     vector<vector<int> > components;
//     vector<int> n_component_customers;
//     int index_x_customer = index_x.CUSTOMER;
//     int index_y_customer = index_y.CUSTOMER;
//     set<int> not_visited_customers;
//     vector<bool> visited = vector<bool>(n_customers + n_steiners);
//     int cust, n_cust;
//     int x_source, x_target, y_source, y_target;
//     
//     int status = CPXgetcallbacknodex (env, cbdata, wherefrom, x, 0, numcols-1);
//     if ( status ) 
//     {
//         cout << "Error obteniendo la solución de la relajación actual" << endl;
//         goto TERMINATE;
//     }
//     
//     
//     /**************************************************************************
//      * Busqueda de las componentes conexas que contienen clientes
//      * aisladas del deposito
//      **************************************************************************/
//     for (int i = 0; i < n_customers; i++)
//         not_visited_customers.insert(i);
//     
//     for (int i = 0; i < n_customers+n_steiners; i++)
//         visited[i] = false;
//     
//     while (!not_visited_customers.empty())
//     {
//         vector<int> component = vector<int>();
//         vector<int> visit_stack = vector<int>();
//         cust = *not_visited_customers.begin();
//         
//         component.push_back(cust); n_cust = 1;
//         not_visited_customers.erase(cust);
//         visited[cust] = true;
//         visit_stack.push_back(cust);
//         while (visit_stack.size() > 0)
//         {
//             int source = visit_stack.back();
//             visit_stack.pop_back();
//             
//             x_source = index_x_customer + source;
//             y_source = index_y_customer + source;
//             for (int target = 0; target < _V1; target++)
//             {
//                 x_target = index_x_customer + target;
//                 y_target = index_y_customer + target;
//                 
//                 if (!visited[target])
//                 {
//                     if ( (x[var_x(x_source, index_x.D1)] > machine_eps) || (x[var_x(index_x.D0, x_source)] > machine_eps) )
//                     {
//                         cuts_generated = true;
//                     }
//                     
//                     /* verifica si esta conectado por medio de una conexion anillo o estrella */
//                     if ( (x[var_x(x_source, x_target)] > machine_eps) || (x[var_x(x_target, x_source)] > machine_eps) ||
//                          ( (source < n_customers) && (x[var_y(y_source, y_target)] > machine_eps) ) ||
//                          ( (target < n_customers) && (x[var_y(y_target, y_source)] > machine_eps) )
//                        )
//                     {
//                         component.push_back(target);
//                         visit_stack.push_back(target);
//                         if (target < n_customers)
//                         {
//                             not_visited_customers.erase(target);
//                             n_cust++;
//                         }
//                         visited[target] = true;
//                     }
//                     
//                 }
//             }
//         }
//         
//         if (component.size() > 1)
//         {
//             components.push_back(component);
//             n_component_customers.push_back(n_cust);
//         }
//     }
//     
//     /**************************************************************************
//      * Agregando restricciones a la formulación
//      **************************************************************************/
//     if (cuts_generated)
//     {
//         vector<int> in_component(n_customers + n_steiners);
//         for (int i = 0; i < n_customers + n_steiners; i++)
//             in_component[i] = -1;
//         
//         for (int k = 0; k < components.size(); k++)
//         {
//             for (int i = 0; i < components[k].size(); i++)
//             {
//                 in_component[components[k][i]] = k;
//             }
//         }
//         
//         int numnz = 2*_V*_V;
//         int rmatind[numnz];
//         double rmatval[numnz];
//         double rhs;
//         char sense;
//         int index = 0;
//         double lhs_value = 0.0;
//         int nz;
// 
//         for (int k = 0; k < components.size(); k++)
//         {
//             index = 0;
//             for (int i = 0; i < components[k].size(); i++)
//             {
//                 for (int j = 0; j < _V1; j++)
//                 {
//                     if (in_component[j] != k)
//                     {
//                         lhs_value += x[var_x(index_x_customer + components[k][i], index_x_customer + j)];
//                         rmatind[index] = var_x(index_x_customer + components[k][i], index_x_customer + j);
//                         rmatval[index] = 1.0;
//                         index++;
//                         
//                         if (components[k][i] < n_customers) // i es cliente
//                         {
//                             lhs_value += x[var_y(index_y_customer + components[k][i], index_y_customer + j)];
//                             rmatind[index] = var_y(index_y_customer + components[k][i], index_y_customer + j);
//                             rmatval[index] = 1.0;
//                             index++;
//                         }
//                     }
//                 }
//                 lhs_value += x[var_x(index_x_customer + components[k][i], index_x.D1)];
//                 rmatind[index] = var_x(index_x_customer + components[k][i], index_x.D1);
//                 rmatval[index] = 1.0;
//                 index++;
//                 
//             }
//             
//             if (lhs_value > ceil( ((double)n_component_customers[k])/ring_capacity )) // corta la solución actual
//             {
//                 add_cuts = true;
//                 sense = 'G';
//                 rhs = ceil( ((double)n_component_customers[k])/ring_capacity );
//                 status = CPXcutcallbackadd(env, cbdata, wherefrom, index, rhs, sense, rmatind, rmatval);
//                 if ( status ) {
//                     cout << "Error al intentar agregar cortes" << endl;
//                     goto TERMINATE;
//                 } 
// 
//             }
//             
//         }
//         
// /*        cout << "COMPONENTES: " << endl;
//         for (int k = 0; k < components.size(); k++)
//         {
//             vector<int> component = components[k];
//             cout << "COMPONENTE: " << endl;
//             for (int i = 0; i < component.size(); i++)
//             {
//                 if (component[i] < n_customers)
//                     cout << customers_labels[component[i]] << " ";
//                 else
//                     cout << steiners_labels[component[i]-n_customers] << " ";
//             }
//             cout << endl << "----------" << endl;
//         }*/
//         
//     }        
//     
//     if (add_cuts)
//     {
//         cout << "se agregaron cortes familia 1" << endl;
//     }
// 
// TERMINATE:
//     delete[] x;
//     return add_cuts;
// }
// 
// 
// 
// 
// 
// 
