/******************************************************************************
 * 
 * Capacitated m-Ring Star Problem Solver
 * 
 * Resuelve instancias del Problema del m-anillo estrella con capacidades
 * 
 ******************************************************************************/

#include <iostream>
#include <fstream>
#include <cmath>
#include <map>
#include <ctime>
#include <new>
#include <sys/times.h>

#include "util.h"
#include "CMRSPbranching.h"

#define __DEBUG

using namespace std;

CMRSRep* cmrs_rep;
HeuristicaInicial* heuristica_inicial;
vector<Separacion*> metodos_separacion;

CPXENVptr env;
CPXLPptr cmrsp_model;
double optim_time;
double opt_rel = -1.0;
double opt;
double gap = DBL_MAX;
int nodes_explored;
vector<int> cortes_aplicados;
ExecParam param;
//char** tmp_colname;
int iter = 0;
int last_node = 0;


// #define INDEX_CAPACITY 0
// #define INDEX_RING_XOR_STAR_1 1
// #define INDEX_RING_XOR_STAR_2 2
// const int n_cuts_families = 3;
// string cuts_families[] = { "capacity", "ring_xor_star_1", "ring_xor_star_2" };
// int max_cuts[] = {1000, 1000, 1000};
// int cuts_per_family[n_cuts_families];

#define INDEX_Z1_BOUNDS 0
const int n_bounds_families = 1;
string bounds_families[] = { "z1" };
int bounds_per_family[n_bounds_families];

typedef struct var_index
{
    int D0;
    int D1;
    int CUSTOMER;
    int STEINER;
} var_index;

var_index index_x, index_y, index_z, index_f, index_p, index_connection_cost, index_routing_cost;

int max_index;
void revisar_indice(int index, int constraint)
{
    if (index >= max_index)
        cout << "constraint " << constraint << ", index = " << index << endl;
}


CMRSP::CMRSP(CMRSRep* p_cmrs_rep, const ExecParam& parameters) throw(CMRSPException)
{
    /***************************************************************************
    * Inicialización del entorno CPLEX
    ***************************************************************************/
    env = NULL;
    int status = 0;
    
    env = CPXopenCPLEX (&status);
    
    if (env == NULL)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        throw CMRSPException("Error incializando el entorno de CPLEX"+string(errmsg));
    }
    
   /***************************************************************************
    * Setea los parametros de optimización de CPLEX
    ***************************************************************************/
    cmrs_rep = p_cmrs_rep;
    heuristica_inicial = NULL;
    metodos_separacion = vector<Separacion*>();
    
    param = parameters;
    
    CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
    
    CPXsetintparam (env, CPX_PARAM_CONFLICTDISPLAY, 2);

    CPXsetintparam(env, CPX_PARAM_MIPINTERVAL, param.logging_interval);
    
    if (param.initial_heuristic) CPXsetintparam (env, CPX_PARAM_ADVIND, 2); // Considera la solucion inicial del usuario
    
    if (!param.clique_cuts) CPXsetintparam (env, CPX_PARAM_CLIQUES, -1);
    if (!param.cover_cuts) CPXsetintparam (env, CPX_PARAM_COVERS, -1);
    if (!param.disjuntive_cuts) CPXsetintparam (env, CPX_PARAM_DISJCUTS, -1);
    if (!param.flowcover_cuts) CPXsetintparam (env, CPX_PARAM_FLOWCOVERS, -1);
    if (!param.flowpath_cuts) CPXsetintparam (env, CPX_PARAM_FLOWPATHS, -1);
    if (!param.fractional_cuts) CPXsetintparam (env, CPX_PARAM_FRACCUTS, -1);
    if (!param.gubcover_cuts) CPXsetintparam (env, CPX_PARAM_GUBCOVERS, -1);
    if (!param.mir_cuts) CPXsetintparam (env, CPX_PARAM_MIRCUTS, -1);
    
    // Time limit en segundos
    cout << "CPX_PARAM_TILIM = " << 1.0*param.time_limit << endl;
    CPXsetdblparam(env, CPX_PARAM_TILIM, 1.0*param.time_limit);
    
    // Priority order en la seleccion de variables para el branching
    if (param.var_priority > 0) CPXsetintparam(env, CPX_PARAM_MIPORDIND, CPX_ON);
    
    // 0 [CPX_NODESEL_DFS] Depth-first search 
    // 1 [CPX_NODESEL_BESTBOUND] Best-bound search 
    // 2 [CPX_NODESEL_BESTEST] Best-estimate search 
    // 3 [CPX_NODESEL_BESTEST_ALT] Alternative best-estimate search
    if (param.node_sel > -1) CPXsetintparam(env, CPX_PARAM_NODESEL, param.node_sel);

    // -1 [CPX_VARSEL_MININFEAS]  Branch on variable with minimum infeasibility 
    // 0 [CPX_VARSEL_DEFAULT]  Branch variable automatically selected 
    // 1 [CPX_VARSEL_MAXINFEAS]  Branch on variable with maximum infeasibility 
    // 2 [CPX_VARSEL_PSEUDO]  Branch based on pseudo costs 
    // 3 [CPX_VARSEL_STRONG]  Strong branching 
    // 4 [CPX_VARSEL_PSEUDOREDUCED]  Branch based on pseudo reduced costs 
    if (param.var_sel > -1) CPXsetintparam(env, CPX_PARAM_VARSEL, param.var_sel);
    
    // 0 [CPX_MIPEMPHASIS_BALANCED] Balance optimality and feasibility 
    // 1 [CPX_MIPEMPHASIS_FEASIBILITY] Emphasize feasibility over optimality 
    // 2 [CPX_MIPEMPHASIS_OPTIMALITY] Emphasize optimality over feasibility 
    // 3 [CPX_MIPEMPHASIS_BESTBOUND] Emphasize moving best bound 
    // 4 [CPX_MIPEMPHASIS_HIDDENFEAS] Emphasize finding hidden feasible solutions 
//     CPXsetintparam (env, CPX_PARAM_MIPEMPHASIS, CPX_MIPEMPHASIS_BESTBOUND);
    if (param.mip_emphasis > -1) CPXsetintparam(env, CPX_PARAM_MIPEMPHASIS, param.mip_emphasis);
    
    // Branching callback
//     if (param.branching_1)
//     {
//         CPXsetintparam(env, CPX_PARAM_PRELINEAR, CPX_OFF);
//         CPXsetintparam(env, CPX_PARAM_MIPCBREDLP, CPX_OFF);
//         
//         status = CPXsetbranchcallbackfunc (env, usersetbranch, NULL);
//     
//         if (status)
//         {
//             char errmsg[__MSG_ERR_LEN];
//             CPXgeterrorstring (env, status, errmsg);
//             throw CMRSPException("Error seteando la funcion callback para branching. "+string(errmsg));
//         }
//     }
    
    // Cut callback
    if (param.cutting_plane_1 > 0)
    {
        CPXsetintparam(env, CPX_PARAM_PRELINEAR, CPX_OFF);
        CPXsetintparam(env, CPX_PARAM_MIPCBREDLP, CPX_OFF);
        
        status = CPXsetcutcallbackfunc (env, mycutcallback, NULL);
        if (status)
        {
            char errmsg[__MSG_ERR_LEN];
            CPXgeterrorstring (env, status, errmsg);
            throw CMRSPException("Error seteando la funcion callback para la aplicación de cortes. "+string(errmsg));
        }
    }
    
   /***************************************************************************
    * Construcción de instancia del problema
    ***************************************************************************/
    cmrsp_model = NULL;
    cmrsp_model = CPXcreateprob (env, &status, "cmrsp_model");
    if (cmrsp_model == NULL)
    {
        throw CMRSPException("Error creando instancia del problema");
    }
    
}

CMRSP::~CMRSP()
{
    /***************************************************************************
    * Liberación de espacio de entorno CPLEX
    ***************************************************************************/
    int status;
    if (cmrsp_model != NULL)
        status = CPXfreeprob (env, &cmrsp_model);
    
    status = CPXcloseCPLEX (&env);

    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cerr << "Error cerrando entorno de CPLEX" << endl;
        cerr << errmsg << endl;
    }
    
}

void CMRSP::set_heuristica_inicial(HeuristicaInicial* h)
{
    heuristica_inicial = h;
}

void CMRSP::add_metodo_separacion(Separacion* m)
{
    metodos_separacion.push_back(m);
}


void CMRSP::optimize() throw(CMRSPException)
{
    /**************************************************************************
     * Construye la instancia CPLEX del problema
     **************************************************************************/
    #ifdef __DEBUG
    cout << "Construyendo instancia del problema " << param.filename << endl;
    #endif

    create_instance(param.filename);
    
    /**************************************************************************
     * Busca una solución con una heurística inicial y la agrega al problema
     **************************************************************************/
    CPXFILEptr logfile = create_logfile(param.filename);
    
    struct tms now;
    long init,end;
    
    times (&now);
    init = (long) (now.tms_utime*100)/sysconf(_SC_CLK_TCK);
    
    //time_t t_start = time(NULL);
    if (param.initial_heuristic)
    { 
        #ifdef __DEBUG
        cout << param.filename << ": heuristica inicial" << endl;
        #endif
        heuristica_inicial->aplicar(env, cmrsp_model);
    }
    
    /**************************************************************************
     * Optimización
     **************************************************************************/
    #ifdef __DEBUG
    cout << param.filename << ": optimizando" << endl;
    #endif
    mip_optimize();
    
    times (&now);
    end = (long) (now.tms_utime*100)/sysconf(_SC_CLK_TCK);


    //time_t t_end = time(NULL);
    //optim_time = difftime(t_end, t_start);
    optim_time = (end-init)/100;
    
    CPXfclose (logfile);
    
    /**************************************************************************
     * Obtiene el resultado de la optimización
     **************************************************************************/
    #ifdef __DEBUG
    cout << param.filename << ": obteniendo solucion" << endl;
    #endif
    mip_solution();
    
}


CMRS CMRSP::solution() throw(CMRSPException)
{
    /**************************************************************************
     * Obtiene la solucion encontrada por CPLEX
     **************************************************************************/
    int cur_numrows = CPXgetnumrows(env, cmrsp_model);
    int cur_numcols = CPXgetnumcols(env, cmrsp_model);
    
    double*  x = (double *) malloc (cur_numcols * sizeof(double));
    
    int status = CPXgetmipx(env, cmrsp_model, x, 0, cur_numcols-1);
    if (status)
    {
        cerr << "Error obteniendo la solucion" << endl;
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cerr << errmsg << endl;
        CPXcloseCPLEX(&env);
        free(x);
        throw CMRSPException(errmsg);
    }
    
    CMRS cmrs = solution(x);
    free(x);
    
    return cmrs;
}
   
CMRS CMRSP::solution(double* x)
{
    /**************************************************************************
     * Construye un CMRS optimo con la solución devuelta por CPLEX
     **************************************************************************/
    CMRS cmrs(cmrs_rep->n_rings(), cmrs_rep->ring_capacity());
    
    // agrega el depósito
    cmrs.depot(cmrs_rep->depot0_label()+"_"+cmrs_rep->depot1_label());

    // agrega clientes
    for (int i = 0; i < cmrs_rep->customers_labels().size(); i++)
    {
        cmrs.add_customer(cmrs_rep->customers_labels()[i]);
    }

    // agrega nodos de steiner
    for (int i = 0; i < cmrs_rep->steiners_labels().size(); i++)
    {
        cmrs.add_steiner(cmrs_rep->steiners_labels()[i]);
    }
    
    // agrega arcos de conexiones
    string source_label, target_label;
    for (int i = 0; i < cmrs_rep->V(); i++)
    {
        for (int j = 0; j < cmrs_rep->V(); j++)
        {
            if (x[cmrs_rep->var_x(i, j)] > m_eps)
            {
                source_label = cmrs_rep->label_x(0, i);
                target_label = cmrs_rep->label_x(1, j);
                if ((source_label.compare(cmrs_rep->depot0_label()) == 0) || (source_label.compare(cmrs_rep->depot1_label()) == 0))
                    source_label = cmrs.get_depot();
                if ((target_label.compare(cmrs_rep->depot0_label()) == 0) || (target_label.compare(cmrs_rep->depot1_label()) == 0))
                    target_label = cmrs.get_depot();

                cmrs.add_ring_connection(source_label, target_label, x[cmrs_rep->var_x(i, j)]);
            }
        }
    }
    
    // agrega arcos de ruteo
    for (int i = 0; i < cmrs_rep->n_customers(); i++)
    {
        for (int j = 0; j < cmrs_rep->n_customers() + cmrs_rep->n_steiners(); j++)
        {            
            if (x[cmrs_rep->var_y(i, j)] > m_eps)
            {
                source_label = cmrs_rep->label_y(0, i);
                target_label = cmrs_rep->label_y(1, j);
                cmrs.add_star_connection(source_label, target_label, x[cmrs_rep->var_y(i, j)]);
            }
        }
    }
    
    return cmrs;
}



void CMRSP::create_instance(const string& filename) throw(CMRSPException)
{
    /***************************************************************************
     * Lee los parametros de la instancia:
     * cantidad de anillos, capacidad de anillo, matrices de costos de conexión
     * y matrices de costo de ruteo
     ***************************************************************************/
    //load_parameters(filename);
    
    /***************************************************************************
     * Construye la estructura CPLEX de la instancia
     ***************************************************************************/
    build_instance();
}


// void CMRSP::load_parameters(const string& filename) throw(CMRSPException)
// {
//     /***************************************************************************
//      * 
//      * Lee los parametros de la instancia:
//      * cantidad de anillos, capacidad de anillo, matrices de costos de conexión
//      * y de costos de ruteo
//      * 
//      ***************************************************************************/
//     int i, j;
// 
//     ifstream ifs(filename.c_str());
//     if (!ifs)
//         throw CMRSPException("Error abriendo archivo: "+filename);
//         
//     ifs >> n_rings;         // cantidad de anillos
//     ifs >> ring_capacity;   // capacidad de cada anillo
//     ifs >> n_customers;     // cantidad de nodos cliente
//     ifs >> n_steiners;      // cantidad de nodos de transicion
//     ifs >> depot0_label;    // nombre del nodo de deposito (salida/inicial)
//     ifs >> depot1_label;    // nombre del nodo de deposito (llegada/final)
// 
//     _V = n_customers + n_steiners + 2;
//     _V1 = n_customers + n_steiners;
//     _V2 = _V*_V;
//     
//     customers_labels = vector<string>(n_customers);   // nombres de los nodos cliente
//     for (i = 0; i < n_customers; i++)
//         ifs >> customers_labels[i];
//     
//     steiners_labels = vector<string>(n_steiners);    // nombres de los nodos de transicion
//     for (i = 0; i < n_steiners; i++)
//         ifs >> steiners_labels[i];
//     
//     connection_cost = vector<vector<double> >(_V);    // matriz de costos de conexion
//     for (i = 0; i < _V; i++)
//     {
//         connection_cost[i] = vector<double>(_V);
//         for (j = 0; j < _V; j++)
//         {
//             ifs >> connection_cost[i][j];
//         }
//     }
// 
//     routing_cost = vector<vector<double> >(n_customers);        // matriz de costos de ruteo
//     for (i = 0; i < n_customers; i++)
//     {
//         routing_cost[i] = vector<double>(_V1);
//         for (j = 0; j < _V1; j++)
//         {
//             ifs >> routing_cost[i][j];
//         }
//     }
//         
//     ifs.close();
//     
//     cmrs_rep = new CMRSRep(filename);
// }


void CMRSP::build_instance() throw(CMRSPException)
{
    /***************************************************************************
     *
     *  Construye la estructura CPLEX de la formulación para la instancia
     *
     ***************************************************************************/
    int n_rings = cmrs_rep->n_rings();
    int ring_capacity = cmrs_rep->ring_capacity();
    int n_customers = cmrs_rep->n_customers();
    int n_steiners = cmrs_rep->n_steiners();
    v_index index_connection_cost = cmrs_rep->index_connection_cost();
    v_index index_routing_cost = cmrs_rep->index_routing_cost();
    v_index index_x = cmrs_rep->index_x();
    v_index index_y = cmrs_rep->index_y();
    v_index index_f = cmrs_rep->index_f();
    v_index index_z = cmrs_rep->index_z();
    vector<vector<double> > connection_cost = cmrs_rep->connection_cost();
    vector<vector<double> > routing_cost = cmrs_rep->routing_cost();
    int _V = cmrs_rep->V();
    int _V1 = cmrs_rep->V1();
    int _V2 = cmrs_rep->V2();
    vector<string> customers_labels = cmrs_rep->customers_labels();
    vector<string> steiners_labels = cmrs_rep->steiners_labels();

    vector<string> nodo = vector<string>(_V);
    int numvars_x = _V2;
    int numvars_y = n_customers*_V1;
    int numvars_z = _V1;
    int numvars_f = _V1;
//     int numvars_p = _V1;
//     int numvars_s = 1;
//     int numvars_t = 1;
//     int numvars_s1 = 1;
//     int numvars_t1 = 1;
    int numvars = numvars_x + numvars_y + numvars_z + numvars_f; // + numvars_p + numvars_s + numvars_t + numvars_s1 + numvars_t1;
    int index, cindex, row, row1, row2, col, col1, col2;
    int numrows_x = _V;
    int numcols_x = _V;
    int numrows_y = n_customers;
    int numcols_y = _V1;
    int numcnstrs_2 = 1;
    int numcnstrs_3 = 1;
    int numcnstrs_4 = _V;
    int numcnstrs_5 = n_customers;
    int numcnstrs_6 = n_customers;
    int numcnstrs_7 = n_steiners;
    int numcnstrs_8 = _V1;
    int numcnstrs_9 = n_customers*_V1;
    int numcnstrs_10 = _V1;
    int numcnstrs_11 = _V1;
    int numcnstrs_12 = _V1*_V1;
    int numcnstrs_13 = n_customers;
    int numcnstrs_14 = n_steiners;
    int numcnstrs_15 = _V1;
    int numcnstrs_16 = n_customers*_V1;
    int numcnstrs_17 = n_steiners*_V1;
    // constraints para rompimiento de simetria    
//     int numcnstrs_18 = _V1;      // p
//     int numcnstrs_19 = _V1*(_V1 - 1);
//     int numcnstrs_20 = _V1;
    // constraints para rompimiento de simetria (parcial)
//     int numcnstrs_21 = _V1;
//     int numcnstrs_22 = _V1;
//     int numcnstrs_23 = _V1;
//     int numcnstrs_24 = _V1;
//     int numcnstrs_25 = 1;    
//     int numcnstrs_26 = 1;    
    int numcnstrs = numcnstrs_2 + numcnstrs_3 + numcnstrs_4 +    // Cantidad total de restricciones
                    numcnstrs_5 + numcnstrs_6 + numcnstrs_7 + 
                    numcnstrs_8 + numcnstrs_9 + numcnstrs_10 + 
                    numcnstrs_11 + numcnstrs_12 + numcnstrs_13 + 
                    numcnstrs_14 + numcnstrs_15 + numcnstrs_16 + numcnstrs_17; //+
//                     numcnstrs_18 + numcnstrs_19 + numcnstrs_20 + 
//                     numcnstrs_21 + numcnstrs_22 + numcnstrs_23 + numcnstrs_24 + numcnstrs_25 + numcnstrs_26;
    int numnz_2 = _V1;                              // Cantidad de elementos no cero en restricciones de la familia 2
    int numnz_3 = _V1;                              // Cantidad de elementos no cero en restricciones de la familia 3
    int numnz_4 = _V;                               // Cantidad de elementos no cero en restricciones de la familia 4
    int numnz_5 = n_customers;                      // Cantidad de elementos no cero en restricciones de la familia 5
    int numnz_6 = (2*_V1 + 1)*n_customers;          // Cantidad de elementos no cero en restricciones de la familia 6
    int numnz_7 = (_V1 + 1)*n_steiners;             // Cantidad de elementos no cero en restricciones de la familia 7
    int numnz_8 = 2*(_V1 + 1)*_V1;                  // Cantidad de elementos no cero en restricciones de la familia 8
    int numnz_9 = n_customers*_V1*_V;               // Cantidad de elementos no cero en restricciones de la familia 9
    int numnz_10 = _V1;                             // Cantidad de elementos no cero en restricciones de la familia 10
    int numnz_11 = _V1;                             // Cantidad de elementos no cero en restricciones de la familia 11
    int numnz_12 = 3*_V1*_V1;                       // Cantidad de elementos no cero en restricciones de la familia 12
    int numnz_13 = n_customers*(n_customers + 1);   // Cantidad de elementos no cero en restricciones de la familia 13
    int numnz_14 = n_steiners*(n_customers + 1);    // Cantidad de elementos no cero en restricciones de la familia 14
    int numnz_15 = _V1;                             // Cantidad de elementos no cero en restricciones de la familia 15
    int numnz_16 = (3 + n_customers)*n_customers*_V1;   // Cantidad de elementos no cero en restricciones de la familia 16
    int numnz_17 = (3 + n_customers)*n_steiners*_V1;    // Cantidad de elementos no cero en restricciones de la familia 17
//     int numnz_18 = 2*_V1;
//     int numnz_19 = 3*_V1*(_V1 - 1);
//     int numnz_20 = 2*_V1;
//     int numnz_21 = 2*_V1;
//     int numnz_22 = 2*_V1;
//     int numnz_23 = 2*_V1;
//     int numnz_24 = 2*_V1;
//     int numnz_25 = 2;
//     int numnz_26 = 2;
    int numnz = numnz_2 + numnz_3  + numnz_4 + numnz_5 + // Cantidad total de elementos no cero
                numnz_6 + numnz_7 + numnz_8 + numnz_9 + 
                numnz_10 + numnz_11 + numnz_12 + numnz_13 + 
                numnz_14 + numnz_15 + numnz_16 + numnz_17; //+
//                 numnz_18 + numnz_19 + numnz_20 +
//                 numnz_21 + numnz_22 + numnz_23 + numnz_24 + numnz_25 + numnz_26;

max_index = numnz;
    
    double* obj;
    char* ctype;
    double* lb;
    double* ub;
    int* rmatbeg;
    int* rmatind;
    double* rmatval;
    double* rhs;
    char* sense;
    
    try
    {
        obj = new double[numvars];
        ctype = new char[numvars];
        lb = new double[numvars];
        ub = new double[numvars];
        rmatbeg = new int[numcnstrs];
        rmatind = new int[numnz];
        rmatval = new double[numnz];
        rhs = new double[numcnstrs];
        sense = new char[numcnstrs];
    }
    catch (bad_alloc& ba)
    {
        cout << "Error alocando memoria:" << endl << ba.what() << endl;
    }
    catch (exception& e)
    {
        cout << "Error:" << endl << e.what() << endl;
    }

    int u, u1, v, v1, v2, w;
    int name_size = 100;
    char* colname[numvars];
    //tmp_colname = new char*[numvars];
    int status;
    int i, j;

    cout << "numvars = " << numvars << ", numnz = " << numnz << ", numcnstrs = " << numcnstrs << endl;

    /* Inicializacion de indices. 
     * Usados para obtener los coeficientes de las variables en la matriz de restricciones
     */    
    index_x.D0 = 0;
    index_x.D1 = 1;
    index_x.CUSTOMER = 2;
    index_x.STEINER = 2 + n_customers;
    index_y.CUSTOMER = 0;
    index_y.STEINER = n_customers;
    index_z.CUSTOMER = 0;
    index_z.STEINER = n_customers;
    index_f.CUSTOMER = 0;
    index_f.STEINER = n_customers;
//     index_p.CUSTOMER = 0;
//     index_p.STEINER = n_customers;
    index_connection_cost.D0 = 0;
    index_connection_cost.D1 = 1;
    index_connection_cost.CUSTOMER = 2;
    index_connection_cost.STEINER = 2 + n_customers;
    index_routing_cost.CUSTOMER = 0;
    index_routing_cost.STEINER = n_customers;
    
    /* Etiqueta los nodos del grafo
     */
    nodo[0] = cmrs_rep->depot0_label();
    nodo[1] = cmrs_rep->depot1_label();
    for (int i = 0; i < n_customers; i++) nodo[2+i] = customers_labels[i];
    for (int i = 0; i < n_steiners; i++) nodo[2+n_customers+i] = steiners_labels[i];
    
    /* (1) Objetivo: 
        \min \sum_{v \in V} \sum_{v' \in V} c_{v, v'} x_{v, v'} + \sum_{u \in U} \sum_{v' \in V'} d_{u, v'} y_{u, v'} 
    */
    index = 0;
    for (i = 0; i < numrows_x; i++)     // x[v1,v2]
    {
        for (j = 0; j < numcols_x; j++)
        {
            obj[index] = connection_cost[i][j];
            lb[index] = 0.0;
            ub[index] = 1.0;
            
            string varname = "x_" + nodo[i] + "_" + nodo[j];
            colname[index] = new char[name_size];
            strlcpy(colname[index], varname.c_str(), name_size);
            //tmp_colname[index] = new char[name_size];
            //strlcpy(tmp_colname[index], varname.c_str(), name_size);
            
            index++;
        }
    }
    
    for (i = 0; i < numrows_y; i++)             // y[v1,v2]
    {
        for (j = 0; j < numcols_y; j++)
        {
            obj[index] = routing_cost[i][j];
            lb[index] = 0.0;
            ub[index] = 1.0;
            
            string varname = "y_" + nodo[i+2] + "_" + nodo[j+2];
            colname[index] = new char[name_size];
            strlcpy(colname[index], varname.c_str(), name_size);
            //tmp_colname[index] = new char[name_size];
            //strlcpy(tmp_colname[index], varname.c_str(), name_size);
            
            index++;
        }
    }
    
    for (i = 0; i < _V1; i++)         // z[v]
    {
        obj[index] = 0.0;
        lb[index] = 0.0;
        ub[index] = _V1 - 1;
        
        string varname = "z_" + nodo[i+2];
        colname[index] = new char[name_size];
        strlcpy(colname[index], varname.c_str(), name_size);
        //tmp_colname[index] = new char[name_size];
        //strlcpy(tmp_colname[index], varname.c_str(), name_size);
        
        index++;
    }

    for (i = 0; i < _V1; i++)         // f[v]
    {
        obj[index] = 0.0;
        lb[index] = 0.0;
        ub[index] = ring_capacity;
        
        string varname = "f_" + nodo[i+2];
        colname[index] = new char[name_size];
        strlcpy(colname[index], varname.c_str(), name_size);
        //tmp_colname[index] = new char[name_size];
        //strlcpy(tmp_colname[index], varname.c_str(), name_size);
        
        index++;        
    }

//     for (i = 0; i < _V1; i++)     // p[v]
//     {
//         obj[index] = 0.0;
//         lb[index] = 0.0;
//         ub[index] = _V1;
//         
//         string varname = "p_" + nodo[i+2];
//         colname[index] = new char[name_size];
//         strlcpy(colname[index], varname.c_str(), name_size);
//         tmp_colname[index] = new char[name_size];
//         strlcpy(tmp_colname[index], varname.c_str(), name_size);
//         
//         index++;
//     }
//     
//     // variables s, t, s1, t1
//     obj[index] = 0.0;
//     lb[index] = 0.0;
//     ub[index] = _V1;
//     colname[index] = new char[1];
//     strcpy(colname[index], "s");
//     tmp_colname[index] = new char[1];
//     strcpy(tmp_colname[index], "s");
//     index++;
//     
//     obj[index] = 0.0;
//     lb[index] = 0.0;
//     ub[index] = _V1;
//     colname[index] = new char[1];
//     strcpy(colname[index], "t");
//     tmp_colname[index] = new char[1];
//     strcpy(tmp_colname[index], "t");
//     index++;
//     
//     obj[index] = 0.0;
//     lb[index] = 0.0;
//     ub[index] = _V1;
//     colname[index] = new char[2];
//     strcpy(colname[index], "s1");
//     tmp_colname[index] = new char[2];
//     strcpy(tmp_colname[index], "s1");
//     index++;
//     
//     obj[index] = 0.0;
//     lb[index] = 0.0;
//     ub[index] = _V1;
//     colname[index] = new char[2];
//     strcpy(colname[index], "t1");
//     tmp_colname[index] = new char[2];
//     strcpy(tmp_colname[index], "t1");
//     index++;
    

    /**************************************************************************
     *
     * Variables del modelo
     *
     **************************************************************************/
    status = CPXnewcols(env, cmrsp_model, index, obj, lb, ub, NULL, colname);
    if (status) throw CMRSPException("Error construyendo la instancia del problema (obj)");       
    
    for (i = 0; i < numvars_x + numvars_y; i++)   // vector de variables binarias
        ctype[i] = CPX_BINARY;
    
    for (i = numvars_x + numvars_y; i < numvars_x + numvars_y + numvars_z + numvars_f; i++) // vector de variables enteras
        ctype[i] = CPX_INTEGER;
        //if (param.continuous_vars) ctype[i] = CPX_CONTINUOUS;
        //else ctype[i] = CPX_INTEGER;
    
    //for (i = numvars_x + numvars_y + numvars_z + numvars_f; i < numvars_x + numvars_y + numvars_z + numvars_f + numvars_p + numvars_s + numvars_t + numvars_s1 + numvars_t1; i++)
    //    ctype[i] = CPX_INTEGER;
    
    cout << "CPXcopyctype i = " << i << " (numvars = " << numvars << ")" << endl;
    CPXcopyctype (env, cmrsp_model, ctype);
    
    /**************************************************************************
     *
     * Prioridades en las variables para el branching
     *
     **************************************************************************/
    if (param.var_priority > 0)
    {
        int cnt = _V1 /* f */ + _V1 /* z */ + n_customers*_V1 + _V2;
        int indices[cnt], priority[cnt], direction[cnt];
        int f_priority, z_priority, y_priority, 
            x_with_source_depot_priority, x_with_target_depot_priority, 
            x_between_customers_priority, x_with_steiner_priority;
        
        switch (param.var_priority)
        {
            case 1:
                f_priority = 1; z_priority = 1; y_priority = 1; 
                x_with_source_depot_priority = 1; x_with_target_depot_priority = 1; x_between_customers_priority = 1; x_with_steiner_priority = 1;
                break;
            
            case 2:
                f_priority = 0; z_priority = 0; y_priority = 1; 
                x_with_source_depot_priority = 2; x_with_target_depot_priority = 2; x_between_customers_priority = 2; x_with_steiner_priority = 2;
                break;
        
            case 3:
                f_priority = 0; z_priority = 0; y_priority = 1; 
                x_with_source_depot_priority = 3; x_with_target_depot_priority = 2; x_between_customers_priority = 3; x_with_steiner_priority = 2;
                break;
        
            case 4:
                f_priority = 0; z_priority = 1; y_priority = 2; 
                x_with_source_depot_priority = 4; x_with_target_depot_priority = 3; x_between_customers_priority = 4; x_with_steiner_priority = 3;
                break;        
        }
        
        index = 0;        
        for (v = 0; v < _V1; v++)
        {
            indices[index] = cmrs_rep->var_f(index_f.CUSTOMER + v);
            priority[index] = f_priority;
            direction[index] = CPX_BRANCH_GLOBAL;
            index++;
        }
        
        for (v = 0; v < _V1; v++)
        {
            indices[index] = cmrs_rep->var_z(index_z.CUSTOMER + v);
            priority[index] = z_priority;
            direction[index] = CPX_BRANCH_GLOBAL;
            index++;
        }
        
        for (u = 0; u < n_customers; u++)
        {
            for (v = 0; v < _V1; v++)
            {
                indices[index] = cmrs_rep->var_y(index_y.CUSTOMER + u, index_y.CUSTOMER + v);
                priority[index] = y_priority;
                direction[index] = CPX_BRANCH_GLOBAL;
                index++;
            }
        }
        
        for (v = 0; v < n_customers + n_steiners; v++)
        {
            for (v1 = 0; v1 < n_customers + n_steiners; v1++)
            {
                indices[index] = cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.CUSTOMER + v1);
        
                if (v < n_customers && v1 < n_customers) priority[index] = x_between_customers_priority;
                else priority[index] = x_with_steiner_priority;
                
                direction[index] = CPX_BRANCH_GLOBAL;
                index++;
            }
            
            indices[index] = cmrs_rep->var_x(index_x.D0, index_x.CUSTOMER + v);
            priority[index] = x_with_source_depot_priority;
            direction[index] = CPX_BRANCH_GLOBAL;
            index++;
            
            indices[index] = cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.D1);
            priority[index] = x_with_target_depot_priority;
            direction[index] = CPX_BRANCH_GLOBAL;
            index++;
        
            indices[index] = cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.D0);
            priority[index] = 0;
            direction[index] = CPX_BRANCH_GLOBAL;
            index++;
        
            indices[index] = cmrs_rep->var_x(index_x.D1, index_x.CUSTOMER + v);
            priority[index] = 0;
            direction[index] = CPX_BRANCH_GLOBAL;
            index++;
        }
        
        status = CPXcopyorder(env, cmrsp_model, cnt, indices, priority, direction);
        if (status) throw CMRSPException("Error asignando prioridades en las variables para el branching.");
    }

    /**************************************************************************
     *
     * Restricciones
     *
     **************************************************************************/    
    /* Inicializacion de indices
     */
    index = 0;  // indice de columna
    cindex = 0; // indice de constraint
    rmatbeg[cindex] = 0;
    
    /* (2)
        \sum_{v \in V'} x_{d_0, v} = m
    */
    col = index_x.CUSTOMER;
    for (i = 0; i < n_customers + n_steiners; i++)
    {
        rmatind[index] = cmrs_rep->var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        col++;
        index++;
    }
    sense[cindex] = 'E';
    rhs[cindex] = n_rings;
    cindex++;
    
    /* (3)
        \sum_{v \in V'} x_{v,d_1} = m
    */
    rmatbeg[cindex] = index;
    col = index_x.CUSTOMER;
    for (i = 0; i < n_customers + n_steiners; i++)
    {
        rmatind[index] = cmrs_rep->var_x(col, index_x.D1);
        rmatval[index] = 1.0;
        col++;
        index++;
    }
    sense[cindex] = 'E';
    rhs[cindex] = n_rings;
    cindex++;
    
    /* (4)
        x_{v, v} = 0 \forall v \in V
    */
    col = index_x.D0;
    for (v = 0; v < _V; v++)
    {
        rmatbeg[cindex] = index;
        
        rmatind[index] = cmrs_rep->var_x(col, col);
        rmatval[index] = 1.0;
        index++;
        
        sense[cindex] = 'E';
        rhs[cindex] = 0.0;
        cindex++;
        
        col++;
    }
    
    /* (5)
        y_{u, u} = 0 \forall u \in U
    */
    col = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        rmatbeg[cindex] = index;
        
        rmatind[index] = cmrs_rep->var_y(col, col);;
        rmatval[index] = 1.0;
        index++;
        
        sense[cindex] = 'E';
        rhs[cindex] = 0.0;
        cindex++;
        
        col++;
    }
    
    /* (6)
        \sum_{v' \in V' \cup \{d_0\}} x_{v',u} + \sum_{v' \in V'} y_{u,v'} = 1 \forall u \in U
    */  
    col = index_x.CUSTOMER;
    col1 = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        rmatbeg[cindex] = index;
        
        rmatind[index] = cmrs_rep->var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        index++;
        row = index_x.CUSTOMER;
        row1 = index_y.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatind[index] = cmrs_rep->var_x(row, col);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = cmrs_rep->var_y(col1, row1);
            rmatval[index] = 1.0;
            index++;
            
            row++;
            row1++;
        }
        sense[cindex] = 'E';
        rhs[cindex] = 1.0;
        col++;
        col1++;
        cindex++;
    }

    /* (7)
        \sum_{v' \in V' \cup \{d_0\}} x_{v',w} \leq 1 \forall w \in W
    */
    col = index_x.STEINER;
    for (w = 0; w < n_steiners; w++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = cmrs_rep->var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        index++;
        
        row = index_x.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatind[index] = cmrs_rep->var_x(row, col);
            rmatval[index] = 1.0;
            index++;
            row++;
        }
        sense[cindex] = 'L';
        rhs[cindex] = 1.0;
        cindex++;
        col++;       
    }

    /* (8)
        \sum_{v'' \in V' \cup \{d_0\}} x_{v'',v'} =  \sum_{v'' \in V' \cup \{d_1\}} x_{v',v''} \forall v' \in V'
    */
    col = index_x.CUSTOMER;
    for (v1 = 0; v1 < _V1; v1++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = cmrs_rep->var_x(index_x.D0, col);
        rmatval[index] = 1.0;
        index++;
        row = index_x.CUSTOMER;
        for (v2 = 0; v2 < _V1; v2++)
        {
            rmatind[index] = cmrs_rep->var_x(row, col);
            rmatval[index] = 1.0;
            index++;
            row++;
        }
        
        row = col;
        col = index_x.CUSTOMER;
        rmatind[index] = cmrs_rep->var_x(row, index_x.D1);
        rmatval[index] = -1.0;
        index++;
        for (v2 = 0; v2 < _V1; v2++)
        {
            rmatind[index] = cmrs_rep->var_x(row, col);
            rmatval[index] = -1.0;
            index++;
            col++;         
        }
        col = row;
        col++;
        sense[cindex] = 'E';
        rhs[cindex] = 0.0;
        cindex++;
    }
    
    /* (9)
        y_{u, v'} \leq \sum_{v \in V' \cup \{d_0\}} x_{v, v'} \forall u \in U \forall v' \in V'
    */
    row = index_x.CUSTOMER;
    row1 = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        col = index_x.CUSTOMER;
        col1 = index_y.CUSTOMER;
        for (v1 = 0; v1 < _V1; v1++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = cmrs_rep->var_y(row1, col1);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = cmrs_rep->var_x(index_x.D0, col);
            rmatval[index] = -1.0;
            index++;
            
            row2 = index_x.CUSTOMER;
            for (v = 0; v < _V1; v++)
            {
                rmatind[index] = cmrs_rep->var_x(row2, col);
                rmatval[index] = -1.0;
                index++;
                row2++;
            }
            sense[cindex] = 'L';
            rhs[cindex] = 0.0;
            cindex++;
            col++;
            col1++;
        }
        row++;
        row1++;
    }
    
    /* (10)
        z_{v'} \geq 1 \forall v' \in V'
    */
    col = index_z.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = cmrs_rep->var_z(col);
        rmatval[index] = 1.0;
        index++;
        col++;
    
        sense[cindex] = 'G';
        rhs[cindex] = 1.0;
        cindex++;
    }
    
    /* (11)
        z_{v'} \leq \vert V' \vert - 1 \forall v' \in V'
    */
   col = index_z.CUSTOMER;
   for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = cmrs_rep->var_z(col);
        rmatval[index] = 1.0;
        index++;
        col++;
    
        sense[cindex] = 'L';
        //rhs[cindex] = _V1 - 1;
        rhs[cindex] = _V1;
        cindex++;
    }
    
    /* (12)
        z_{v'} \geq (z_{v} + 1) - (\vert V' \vert - 1)(1 - x_{v, v'}) \forall v, v' \in V'
    */
    row = index_x.CUSTOMER;
    row1 = index_z.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        col = index_x.CUSTOMER;
        col1 = index_z.CUSTOMER;
        for (v1 = 0; v1 < _V1; v1++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = cmrs_rep->var_z(col1);
            rmatval[index] = 1.0;
            index++;
        
            rmatind[index] = cmrs_rep->var_z(row1);
            rmatval[index] = -1.0;
            index++;
    
            rmatind[index] = cmrs_rep->var_x(row, col);
            rmatval[index] = 1 - _V1;
            index++;
    
            sense[cindex] = 'G';
            rhs[cindex] = 2 - _V1;
            cindex++;
            col++;
            col1++;
        }
        row++;
        row1++;
    }
    

    /* (13)
        f_{u'} \geq 1 + \sum_{u \in U} y_{u, u'} \forall u' \in U
    */
    col = index_y.CUSTOMER;
    col1 = index_f.CUSTOMER;
    for (u1 = 0; u1 < n_customers; u1++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = cmrs_rep->var_f(col1);
        rmatval[index] = 1.0;
        index++;
        
        row = index_y.CUSTOMER;
        for (u = 0; u < n_customers; u++)
        {
            rmatind[index] = cmrs_rep->var_y(row, col);
            rmatval[index] = -1.0;
            index++;
            row++;
        }
        sense[cindex] = 'G';
        rhs[cindex] = 1.0;
        cindex++;
        col++;
        col1++;
    }
    
    /* (14)
        f_{w'} \geq \sum_{u \in U} y_{u, w'} \forall w' \in W
    */
    col = index_y.STEINER;
    col1 = index_f.STEINER;
    for (w = 0; w < n_steiners; w++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = cmrs_rep->var_f(col1);
        rmatval[index] = 1.0;
        index++;
        row = index_y.CUSTOMER;
        for (u = 0; u < n_customers; u++)
        {
            rmatind[index] = cmrs_rep->var_y(row, col);
            rmatval[index] = -1.0;
            index++;
            row++;
        }
        sense[cindex] = 'G';
        rhs[cindex] = 0.0;
        cindex++; 
        col++;
        col1++;
    }
    
    /* (15)
        f_{v'} \leq Q \forall v' \in V'
    */
    col = index_f.CUSTOMER;
    for (v = 0; v < _V1; v++)
    {
        rmatbeg[cindex] = index;
        rmatind[index] = cmrs_rep->var_f(col);
        rmatval[index] = 1.0;
        index++;
        
        sense[cindex] = 'L';
        rhs[cindex] = ring_capacity;
        cindex++;
        col++;
    }

    /* (16)
        f_{u} \geq (f_{v'} + 1 + \sum_{u' \in U} y_{u', u}) - \vert U \vert(1 - x_{v', u}) \forall u \in U \forall v' \in V'
    */
    col = index_x.CUSTOMER;
    col1 = index_f.CUSTOMER;
    col2 = index_y.CUSTOMER;
    for (u = 0; u < n_customers; u++)
    {
        row = index_x.CUSTOMER;
        row1 = index_f.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = cmrs_rep->var_f(col1);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = cmrs_rep->var_f(row1);
            rmatval[index] = -1.0;
            index++;
            
            rmatind[index] = cmrs_rep->var_x(row, col);
            rmatval[index] = -n_customers;
            index++;
            
            row2 = index_y.CUSTOMER;
            for (u1 = 0; u1 < n_customers; u1++)
            {
                rmatind[index] = cmrs_rep->var_y(row2, col2);
                rmatval[index] = -1.0;
                index++;
                row2++;
            }
            
            sense[cindex] = 'G';
            rhs[cindex] = 1 - n_customers;
            cindex++;
            row++;
            row1++;
        }
        
        col++;
        col1++;
        col2++;
    }
    
    /* (17)
        f_{w} \geq (f_{v'} + \sum_{u' \in U} y_{u', w}) - \vert U \vert(1 - x_{v', w}) \forall w \in W \forall v' \in V'
    */
    col = index_x.STEINER;
    col1 = index_f.STEINER;
    col2 = index_y.STEINER;
    for (w = 0; w < n_steiners; w++)
    {
        row = index_x.CUSTOMER;
        row1 = index_f.CUSTOMER;
        for (v = 0; v < _V1; v++)
        {
            rmatbeg[cindex] = index;
            rmatind[index] = cmrs_rep->var_f(col1);
            rmatval[index] = 1.0;
            index++;
            
            rmatind[index] = cmrs_rep->var_f(row1);
            rmatval[index] = -1.0;
            index++;
            
            rmatind[index] = cmrs_rep->var_x(row, col);
            rmatval[index] = -n_customers;
            index++;
            
            row2 = index_y.CUSTOMER;
            for (u1 = 0; u1 < n_customers; u1++)
            {
                 rmatind[index] = cmrs_rep->var_y(row2, col2);
                 rmatval[index] = -1.0;
                 index++;
                 row2++;
            }

            sense[cindex] = 'G';
            rhs[cindex] = -n_customers;
            cindex++;
            row++;
            row1++;
//             row2++;
        }
        col++;
        col1++;
        col2++;
    }
    
    /**************************************************************************
     *
     * Restricciones para rompimiento (parcial) de simetria
     *
     **************************************************************************/
//     if (param.romp_simetria == 1)
//     {
//         cout << "Rompimiento de simetria (modo 1)" << endl;
//         /* (18)
//             p_{v_{j}} \geq j*x_{d0, v_{j}} \forall v_j \in V'
//         */
//         row = index_p.CUSTOMER;
//         col1 = index_x.CUSTOMER;
//         for (v = 0; v < _V1; v++)
//         {
//             rmatbeg[cindex] = index;
//             rmatind[index] = cmrs_rep->var_p(row);
//             rmatval[index] = 1.0;
//             index++;
//             
//             rmatind[index] = cmrs_rep->var_x(index_x.D0, col1);
//             rmatval[index] = -v - 1;
//             index++;
//             
//             sense[cindex] = 'G';
//             rhs[cindex] = 0.0;
//             cindex++;
//             row++;
//             col1++;
//         }
//         
//         /* (19) 
//             p_{j} \geq p_{i} + (-M)(1 - x_{v_{i}, v_{j}}) \forall v_{i}, v_{j} \in V' con M = |U|+|W|
//         */
//         for (v = 0; v < _V1; v++)
//         {
//             row = index_p.CUSTOMER+v;
//             row1 = index_x.CUSTOMER+v;
//             for (v1 = 0; v1 < _V1; v1++)
//             {            
//                 if (v1 == v) continue;
//                 
//                 col = index_p.CUSTOMER+v1;
//                 col1 = index_x.CUSTOMER+v1;
//                 
//                 rmatbeg[cindex] = index;
//                 rmatind[index] = cmrs_rep->var_p(row);
//                 rmatval[index] = 1.0;
//                 index++;
//                 
//                 rmatind[index] = cmrs_rep->var_p(col);
//                 rmatval[index] = -1.0;
//                 index++;
//     
//                 rmatind[index] = cmrs_rep->var_x(row1, col1);
//                 rmatval[index] = _V1;
//                 index++;
//                 
//                 sense[cindex] = 'G';
//                 rhs[cindex] = -_V1;
//                 cindex++;
//             }
//         }
//         
//         /* (20) 
//             p_{j} \leq j*x_{v_{j}, d1} + M(1 - x_{v_{j}, d1}) \forall v_{j} \in V' con M = |U|+|W|
//         */
//         row = index_p.CUSTOMER;
//         row1 = index_x.CUSTOMER;
//         for (v = 0; v < _V1; v++)
//         {
//             rmatbeg[cindex] = index;
//             rmatind[index] = cmrs_rep->var_p(row);
//             rmatval[index] = 1.0;
//             index++;
//             
//             rmatind[index] = cmrs_rep->var_x(row1, index_x.D1);
//             rmatval[index] = _V1;
//             index++;
//     
//             sense[cindex] = 'L';
//             rhs[cindex] = _V1 + v + 1;
//             cindex++;
//             row++;
//             row1++;
//         }
//         
//         /* (21)
//             s \leq j*x_{d0, v_{j}} \forall v_{j} in V'
//         */
//         for (v = 0; v < _V1; v++)
//         {
//             rmatbeg[cindex] = index;
//             rmatind[index] = cmrs_rep->var_s();
//             rmatval[index] = 1.0;
//             index++;
//             
//             rmatind[index] = cmrs_rep->var_x(index_x.D0, index_x.CUSTOMER+v);
//             rmatval[index] = -v - 1;
//             index++;
//             
//             sense[cindex] = 'L';
//             rhs[cindex] = 0.0;
//             cindex++;
//         }
//     
//         /* (22)
//             t \leq j*x_{v_{j}, d1} \forall v_{j} in V'
//         */
//         for (v = 0; v < _V1; v++)
//         {
//             rmatbeg[cindex] = index;
//             rmatind[index] = cmrs_rep->var_t();
//             rmatval[index] = 1.0;
//             index++;
//             
//             rmatind[index] = cmrs_rep->var_x(index_x.CUSTOMER+v, index_x.D1);
//             rmatval[index] = -v - 1;
//             index++;
//             
//             sense[cindex] = 'L';
//             rhs[cindex] = 0.0;
//             cindex++;
//         }
//         
//         /* (23)
//             s1 \geq j*x_{d0, v_{j}} \forall v_{j} in V'
//         */
//         for (v = 0; v < _V1; v++)
//         {
//             rmatbeg[cindex] = index;
//             rmatind[index] = cmrs_rep->var_s1();
//             rmatval[index] = 1.0;
//             index++;
//             
//             rmatind[index] = cmrs_rep->var_x(index_x.D0, index_x.CUSTOMER+v);
//             rmatval[index] = -v - 1;
//             index++;
//             
//             sense[cindex] = 'G';
//             rhs[cindex] = 0.0;
//             cindex++;
//         }
//         
//         /* (24)
//             t1 \geq j*x_{v_{j}, d1} \forall v_{j} in V'
//         */
//         for (v = 0; v < _V1; v++)
//         {
//             rmatbeg[cindex] = index;
//             rmatind[index] = cmrs_rep->var_t1();
//             rmatval[index] = 1.0;
//             index++;
//             
//             rmatind[index] = cmrs_rep->var_x(index_x.CUSTOMER+v, index_x.D1);
//             rmatval[index] = -v - 1;
//             index++;
//             
//             sense[cindex] = 'G';
//             rhs[cindex] = 0.0;
//             cindex++;
//         }
//         
//         /* (25)
//             s \leq t
//         */
//         rmatbeg[cindex] = index;
//         rmatind[index] = cmrs_rep->var_s();
//         rmatval[index] = 1.0;
//         index++;
//         
//         rmatind[index] = cmrs_rep->var_t();
//         rmatval[index] = -1.0;
//         index++;
//         
//         sense[cindex] = 'L';
//         rhs[cindex] = 0.0;
//         cindex++;    
//         
//         /* (26)
//             s1 \leq t1
//         */
//         rmatbeg[cindex] = index;
//         rmatind[index] = cmrs_rep->var_s1();
//         rmatval[index] = 1.0;
//         index++;
//         
//         rmatind[index] = cmrs_rep->var_t1();
//         rmatval[index] = -1.0;
//         index++;
//         
//         sense[cindex] = 'L';
//         rhs[cindex] = 0.0;
//         cindex++;
//     }
//     else if (param.romp_simetria == 2)
//     {
//         cout << "Rompimiento de simetria (modo 2)" << endl;
//         
//         /* 
//             \sum_{v \in V'} \sigma(v) x_{d_0 v} \leq \sum_{v \in V'} \sigma(v) x_{v d_1}
//         */
//         rmatbeg[cindex] = index;
//         for (v = 0; v < _V1; v++)
//         {
//             rmatind[index] = cmrs_rep->var_x(index_x.D0, index_x.CUSTOMER + v);
//             rmatval[index] = 1.0;
//             index++;
//             
//             rmatind[index] = cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.D1);
//             rmatval[index] = -1.0;
//             index++;            
//         }
//         
//         sense[cindex] = 'L';
//         rhs[cindex] = 0.0;
//         cindex++;        
//     }
    
    /***************************************************************************
     * Agrega las restricciones en el modelo
     ***************************************************************************/
agregar_restricciones:
    cout << "agregando restricciones..." << endl;
    cout << "cindex = " << cindex << ", index = " << index << endl;
    status = CPXaddrows(env, cmrsp_model, 0, cindex, index, rhs, sense, rmatbeg, rmatind, rmatval, NULL, NULL /*rowname*/);
    cout << "restricciones agregadas..." << endl;
    
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cout << errmsg << endl;
        throw CMRSPException("Error agregando restricciones: "+string(errmsg));
    }
    
    /***************************************************************************
     * Agrega los conjuntos SOS1
     ***************************************************************************/
//     if (param.sos)
//     {
//         int numsos = n_customers + n_steiners;
//         int numsosnz_1 = n_customers*(2*_V1 + 1);
//         int numsosnz_2 = n_steiners*_V1;
//         int numsosnz = numsosnz_1 + numsosnz_2;
//         char sostype[numsos];
//         int sosbeg[numsos];
//         int sosind[numsosnz];
//         double soswt[numsosnz];
//         
//         for (i = 0; i < numsos; i++)
//         {
//             sostype[numsos] = CPX_TYPE_SOS1;
//         }
//         
//         /*
//             SOS1 que se deduce de (6)
//             { x{d0, u}, x{v'[1],u}, ..., x{v'[n_customers+n_steiners],u}, y{u,v'[1]}, ..., y{u,v'[n_customers+n_steiners]} } para cada cliente u es un SOS1
//         */
//         index = 0;
//         cindex = 0;
//         for (u = 0; u < n_customers; u++)
//         {
//             sosbeg[cindex] = index;
//             cindex++;
//             col = index_x.CUSTOMER;
//             row = index_x.CUSTOMER;
//             row1 = index_y.CUSTOMER;
//             col1 = index_y.CUSTOMER;
//             for (v = 0; v < _V1; v++)
//             {
//                 sosind[index] = cmrs_rep->var_x(row, col);
//                 soswt[index] = 0.1*v;
//                 row++;
//                 index++;
//                 
//                 sosind[index] = cmrs_rep->var_y(row1, col1);
//                 soswt[index] = 0.1*_V1 + 0.1*v;
//                 col1++;
//                 index++;
//             }
//             
//             sosind[index] = cmrs_rep->var_x(index_x.D0, col);
//             soswt[index] = 0.1*_V1*2 + 0.1;
//             index++;
//         }
//         
//         /*
//             SOS1 que se deduce de (7)
//             { x{v'[1],w}, ..., x{v'[n_customers+n_steiners],w} } para cada nodo de steiner w
//         */
//         for (w = 0; w < n_steiners; w++)
//         {
//             sosbeg[cindex] = index;
//             cindex++;
//             row = index_x.CUSTOMER;
//             col = index_x.STEINER;
//             for (v = 0; v < n_customers + n_steiners; v++)
//             {
//                 sosind[index] = cmrs_rep->var_x(row, col);
//                 soswt[index] = 0.1*v;
//                 row++;
//                 index++;
//             }
//         }
//         
//         status = CPXcopysos(env, cmrsp_model, numsos, numsosnz, sostype, sosbeg, sosind, soswt, NULL);
//         
//         if (status) 
//         {
//             char errmsg[__MSG_ERR_LEN];
//             CPXgeterrorstring (env, status, errmsg);
//             throw CMRSPException("Error construyendo la instancia del problema (SOS1) "+string(errmsg));
//         }
//     }
    
    /***************************************************************************
     * Temporal: guarda el modelo en un archivo con formato lp para verificar 
     * si fue creado correctamente (solo para algunas pruebas)
     ***************************************************************************/    
    status = CPXwriteprob(env, cmrsp_model, "test.lp", NULL);
    
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cout << errmsg << endl;
        throw CMRSPException("Error grabando el lp del problema: "+string(errmsg));
    }

    status = CPXwriteprob(env, cmrsp_model, "test.sav", NULL);

    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cout << errmsg << endl;
        throw CMRSPException("Error grabando el sav del problema: "+string(errmsg));
    }

    status = CPXwriteparam(env, "test.prm");
    
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cout << errmsg << endl;
        throw CMRSPException("Error grabando parametros: "+string(errmsg));
    }
    
    delete[] obj;
    delete[] ctype;
    delete[] lb;
    delete[] ub;
    delete[] rmatbeg;
    delete[] rmatind;
    delete[] rmatval;
    delete[] rhs;
    delete[] sense;
    
}

CPXFILEptr CMRSP::create_logfile(const string& filename) throw(CMRSPException)
{
    CPXFILEptr logfile = CPXfopen ((filename+".cplex.optb.log").c_str(), "a");
    int status = CPXsetlogfile (env, logfile);
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cout << errmsg << endl;
        CPXcloseCPLEX (&env);
        throw CMRSPException("Error creando archivo de log");
    }
    return logfile;
}

void CMRSP::mip_optimize() throw(CMRSPException)
{
    cout << param.filename << ": CPXmipopt inicio" << endl;
    
    cortes_aplicados = vector<int>(metodos_separacion.size());
    for (int i = 0; i < metodos_separacion.size(); i++) cortes_aplicados[i] = 0;
    iter = 0;
    int status;
    
    try
    {
        status = CPXmipopt (env, cmrsp_model);
    }
    catch (bad_alloc& ba)
    {
        cout << "Error alocando memoria:" << endl << ba.what() << endl;
    }
    catch (exception& e)
    {
        cout << "Error optimizando:" << endl << e.what() << endl;
    }
    
    cout << param.filename << ": CPXmipopt fin" << endl;
    
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cout << param.filename << ": " << string(errmsg) << endl;
        CPXcloseCPLEX (&env);
        throw CMRSPException("Error optimizando "+string(errmsg));
    }
    
}

void CMRSP::mip_solution() throw(CMRSPException)
{
    double objval;
    int solstat = CPXgetstat(env, cmrsp_model);
    int status = CPXgetmipobjval(env, cmrsp_model, &objval);
    
    if (status)
    {
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        CPXcloseCPLEX (&env);
        throw CMRSPException("Error obteniendo el valor de la función objetivo"+string(errmsg));
    }
    
    nodes_explored = CPXgetnodecnt(env, cmrsp_model);
    cout << "Nodos explorados = " << nodes_explored << endl;
    for (int i = 0; i < cortes_aplicados.size(); i++) 
        cout << "Cortes aplicados de familia " << 
                metodos_separacion[i]->nombre_familia() << 
                ": " << cortes_aplicados[i] << endl;
    cout << endl << "Objetivo (min) = " << objval << " (status = " << solstat << ")" << endl;
    if (solstat == 101) gap = 0.0;
    cout << "Gap = " << gap << "%" << endl;
    CMRS cmrs = solution();
    cmrs.save_dot("cmrs_opt.dot");

}


void CMRSP::save_log() throw(CMRSPException)
{
    ofstream ofs;
    ofs.open((param.filename+".log").c_str(), ios_base::app);
    if (!ofs) throw CMRSPException("Error grabando log.");
    
    ofs << param.filename << " optb"
        << " " << cmrs_rep->n_rings()
        << " " << cmrs_rep->ring_capacity()
        << " " << cmrs_rep->n_customers()
        << " " << cmrs_rep->n_steiners()
        << " " << param.initial_heuristic
        << " " << param.var_priority
        << " " << param.var_sel
        << " " << param.node_sel
        << " " << param.mip_emphasis
        << " " << param.clique_cuts 
        << " " << param.cover_cuts 
        << " " << param.disjuntive_cuts 
        << " " << param.flowcover_cuts 
        << " " << param.flowpath_cuts 
        << " " << param.fractional_cuts 
        << " " << param.gubcover_cuts 
        << " " << param.mir_cuts 
        << " " << param.cutting_plane_1 
        << " " << param.cutting_plane_2 
        << " " << param.cutting_plane_3 
        << " " << param.cutting_plane_4 
        << " " << optim_time 
        << " " << nodes_explored
        << endl;    
    
    ofs.close();
}

// int CPXPUBLIC CMRSP::usersetbranch (CPXCENVptr env, void *cbdata, int wherefrom, void *cbhandle, int type, int sos, int nodecnt,
//     int bdcnt, const double *nodeest, const int *nodebeg, const int *indices, const char *lu, const int *bd, int *useraction_p)
//  {
//     int status = 0;
//     *useraction_p = CPX_CALLBACK_DEFAULT;
//     int cur_numrows = CPXgetnumrows(env, cmrsp_model);
//     int cur_numcols = CPXgetnumcols(env, cmrsp_model);
//     int feas[cur_numcols], node_indices[cur_numcols];
//     double node_objval;
//     double x[cur_numcols], node_lb[cur_numcols], node_ub[cur_numcols];
//     int node_bd[cur_numcols];
//     char node_lu[cur_numcols];
//     
// //     cout << "------- usersetbranch ----------" << endl;
// 
// //     if (sos >= 0) cout << "SOS: ";    
// //     cout << "branching=";
// //     if (type == CPX_TYPE_VAR) cout << "VAR";
// //     else if (type == CPX_TYPE_SOS1) cout << "SOS1";
// //     else if (type == CPX_TYPE_SOS2) cout << "SOS2";
// //     else if (type == CPX_TYPE_ANY) cout << "A";
// //     else cout << "?";
// //     cout << endl;
//     
// //     for (int branch = 0; branch < nodecnt; branch++)
// //     {
// //         cout << "nodebeg[" << branch << "] = " << nodebeg[branch] << endl;
// //     }
// // 
// //     for (int bound_index = 0; bound_index < bdcnt; bound_index++)
// //     {
// //         cout << "indices[" << bound_index << "] = " << indices[bound_index] << " (" << tmp_colname[indices[bound_index]] << "), "
// //              << "bd[" << bound_index << "] = " << bd[bound_index] << ", "
// //              << "lu[" << bound_index << "] = " << (unsigned int) lu[bound_index] << " (" << (char) lu[bound_index] << ")"
// //              <<  endl;
// //         
// //         cout << "   var = " << tmp_colname[indices[bound_index]] << ", value = " << bd[indices[bound_index]] << endl;
// //     }
// 
//     
// //     status = CPXgetcallbacknodex (env, cbdata, wherefrom, x, 0, cur_numcols-1);
// // 
// //     status = CPXgetcallbacknodeintfeas (env, cbdata, wherefrom, feas, 0, cur_numcols-1);
// 
// //     cout << "INTEGER FEASIBLE: ";
// //     for (int index = 0; index < cur_numcols; index++)
// //     {
// //         if (feas[index] == CPX_INTEGER_FEASIBLE) cout << tmp_colname[index] << "=" << x[index] << " ";
// //     }
// //     cout << endl;
// //     
// //     cout << "IMPLIED INTEGER FEASIBLE: "; 
// //     for (int index = 0; index < cur_numcols; index++)
// //     {
// //         if (feas[index] == CPX_IMPLIED_INTEGER_FEASIBLE) cout << tmp_colname[index] << "=" << x[index] << " ";
// //     }
// //     cout << endl;
// //     
// //     cout << "INTEGER INFEASIBLE: "; 
// //     for (int index = 0; index < cur_numcols; index++)
// //     {
// //         if (feas[index] == CPX_INTEGER_INFEASIBLE) cout << tmp_colname[index] << "=" << x[index] << " ";
// //     }
// //     cout << endl;
// 
// //     cout << "--------------------------------" << endl;
// 
//     status = CPXgetcallbacknodeintfeas (env, cbdata, wherefrom, feas, 0, cur_numcols-1) || 
//              CPXgetcallbacknodelb(env, cbdata, wherefrom, node_lb, 0, cur_numcols-1) || 
//              CPXgetcallbacknodeub(env, cbdata, wherefrom, node_ub, 0, cur_numcols-1) ||
//              CPXgetcallbacknodex (env, cbdata, wherefrom, x, 0, cur_numcols-1) ||
//              CPXgetcallbacknodeobjval(env, cbdata, wherefrom, &node_objval);
//     
//     return status;
//     
// 
//     int n_customers = cmrs_rep->n_customers();
//     int n_steiners = cmrs_rep->n_steiners();
//     bool new_bounds = false;
//     // x_{d_0, u} = 1 => z_{u} = 1
//     int cnt = 0;    
//     for (int v = 0; v < n_customers + n_steiners; v++)
//     {
// //         if ( feas[var_x(index_x.D0, index_x.CUSTOMER + u)] == CPX_INTEGER_FEASIBLE && 
// //              x[var_x(index_x.D0, index_x.CUSTOMER + u)] > 1.0 - machine_eps && 
// //              x[var_z(index_z.CUSTOMER + u)] > 1.0 )
// //         {
// //             cout << "fijando valor de " << tmp_colname[var_z(index_z.CUSTOMER + u)] << endl;
// //             node_indices[cnt] = var_z(index_z.CUSTOMER + u);
// //             node_lu[cnt] = 'B';
// //             node_bd[cnt] = 1;
// //             bounds_per_family[INDEX_Z1_BOUNDS]++;
// //             cnt++;
// //             new_bounds = true;
// //         }
//         
// //         for (int u1 = 0; u1 < n_customers; u1++)
// //         {
// //             if (u1 == u) continue;
// //             
// //             if ( feas[var_x(index_x.CUSTOMER + u, index_x.CUSTOMER + u1)] == CPX_INTEGER_FEASIBLE &&
// //                 feas[var_f(index_f.CUSTOMER + u)] == CPX_INTEGER_FEASIBLE && 
// //                 x[var_x(index_x.CUSTOMER + u, index_x.CUSTOMER + u1)] > 1.0 - machine_eps && 
// //                 x[var_f(index_f.CUSTOMER + u1)] + machine_eps < x[var_f(index_f.CUSTOMER + u)] + 1.0 )
// //             {
// //                 cout << "fijando valor de " << tmp_colname[var_f(index_f.CUSTOMER + u1)] << ", " << tmp_colname[var_x(index_x.CUSTOMER + u, index_x.CUSTOMER+u1)] << "=" << x[var_x(index_x.CUSTOMER + u, index_x.CUSTOMER+u1)] << endl;
// //                 node_indices[cnt] = var_f(index_f.CUSTOMER + u1);
// //                 node_lu[cnt] = 'L';
// //                 node_bd[cnt] = (int)x[var_f(index_f.CUSTOMER + u)] + 1;
// //                 bounds_per_family[INDEX_Z1_BOUNDS]++;
// //                 cnt++;
// //                 new_bounds = true;
// //             }
// //         }
//         for (int v1 = 0; v1 < n_customers+n_steiners; v1++)
//         {
//             if (v1 == v) continue;
//             
//             if ( feas[cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.CUSTOMER + v1)] == CPX_INTEGER_FEASIBLE &&
//                 feas[cmrs_rep->var_z(index_z.CUSTOMER + v)] == CPX_INTEGER_FEASIBLE && 
//                 x[cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.CUSTOMER + v1)] > 1.0 - m_eps && 
//                 x[cmrs_rep->var_z(index_z.CUSTOMER + v1)] > x[cmrs_rep->var_z(index_z.CUSTOMER + v)] + 1.0 + m_eps )
//             {
//                 cout << "fijando valor de " << tmp_colname[cmrs_rep->var_z(index_z.CUSTOMER + v1)] 
//                      << " (" << tmp_colname[cmrs_rep->var_z(index_z.CUSTOMER + v1)] << "=" << x[cmrs_rep->var_z(index_z.CUSTOMER + v1)] << ", "
//                      << tmp_colname[cmrs_rep->var_z(index_z.CUSTOMER + v)] << "=" << x[cmrs_rep->var_z(index_z.CUSTOMER + v)] << ", "
//                      << tmp_colname[cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.CUSTOMER+v1)] << "=" << x[cmrs_rep->var_x(index_x.CUSTOMER + v, index_x.CUSTOMER+v1)] 
//                      << ")" << endl;
//                 node_indices[cnt] = cmrs_rep->var_z(index_z.CUSTOMER + v1);
//                 node_lu[cnt] = 'B';
//                 node_bd[cnt] = (int)x[cmrs_rep->var_z(index_z.CUSTOMER + v)] + 1;
//                 bounds_per_family[INDEX_Z1_BOUNDS]++;
//                 cnt++;
// 
//                 new_bounds = true;
//             }
//             
// //             if ( v1 < n_customers && 
// //                  feas[var_x(index_x.CUSTOMER + v, index_x.CUSTOMER + v1)] == CPX_INTEGER_FEASIBLE &&
// //                  feas[var_f(index_f.CUSTOMER + v)] == CPX_INTEGER_FEASIBLE && 
// //                  feas[var_f(index_f.CUSTOMER + v1)] == CPX_INTEGER_INFEASIBLE &&
// //                  x[var_f(index_f.CUSTOMER + v1)] + machine_eps < (int)x[var_f(index_f.CUSTOMER + v)] + 1 &&
// //                  x[var_x(index_x.CUSTOMER + v, index_x.CUSTOMER + v1)] > 1.0 - machine_eps )
// //             {
// //                 cout << "fijando valor de " << tmp_colname[var_f(index_f.CUSTOMER + v1)] 
// //                      << " (" << tmp_colname[var_f(index_f.CUSTOMER + v1)] << "=" << x[var_f(index_f.CUSTOMER + v1)] << ", "
// //                      << tmp_colname[var_f(index_f.CUSTOMER + v)] << "=" << x[var_f(index_f.CUSTOMER + v)] << ", "
// //                      << tmp_colname[var_x(index_x.CUSTOMER + v, index_x.CUSTOMER+v1)] << "=" << x[var_x(index_x.CUSTOMER + v, index_x.CUSTOMER+v1)] 
// //                      << ")" << endl;
// //                 node_indices[cnt] = var_f(index_f.CUSTOMER + v1);
// //                 node_lu[cnt] = 'L';
// //                 node_bd[cnt] = (int)x[var_f(index_f.CUSTOMER + v)] + 1;                
// //                 bounds_per_family[INDEX_Z1_BOUNDS]++;
// //                 cnt++;
// // 
// //                 new_bounds = true;
// //             }
//             
//             if ( v1 < n_customers && 
//                  feas[cmrs_rep->var_y(index_y.CUSTOMER + v1, index_y.CUSTOMER + v)] == CPX_INTEGER_FEASIBLE &&
//                  feas[cmrs_rep->var_z(index_z.CUSTOMER + v1)] == CPX_INTEGER_INFEASIBLE &&
//                  x[cmrs_rep->var_z(index_z.CUSTOMER + v1)] > (int)x[cmrs_rep->var_z(index_z.CUSTOMER + v)] + m_eps &&
//                  x[cmrs_rep->var_y(index_y.CUSTOMER + v1, index_y.CUSTOMER + v)] > 1.0 - m_eps )
//             {
//                 cout << "fijando valor de " << tmp_colname[cmrs_rep->var_f(index_f.CUSTOMER + v1)] 
//                      << " (" << tmp_colname[cmrs_rep->var_f(index_z.CUSTOMER + v1)] << "=" << x[cmrs_rep->var_z(index_z.CUSTOMER + v1)] << ", "
//                      << tmp_colname[cmrs_rep->var_y(index_y.CUSTOMER + v, index_y.CUSTOMER+v1)] << "=" << x[cmrs_rep->var_y(index_y.CUSTOMER + v, index_y.CUSTOMER+v1)] 
//                      << ")" << endl;
//                 node_indices[cnt] = cmrs_rep->var_z(index_z.CUSTOMER + v1);
//                 node_lu[cnt] = 'B';
//                 node_bd[cnt] = 1;
//                 bounds_per_family[INDEX_Z1_BOUNDS]++;
//                 cnt++;
//                 
//                 new_bounds = true;
//             }
//             
//         }
//         
//     }
//     
//     if (new_bounds)
//     {
//         int seqnum_1;
//         status = CPXbranchcallbackbranchbds(env, cbdata, wherefrom, node_objval, cnt, node_indices, node_lu, node_bd, NULL, &seqnum_1);
//         if (status)
//         {
//             cout << "Error modificando las cotas de las variables" << endl;
//             return status;
//         }        
//         
//         *useraction_p = CPX_CALLBACK_SET;
//     }
//     
//     return status;
// }

int CPXPUBLIC CMRSP::mycutcallback (CPXCENVptr env, void *cbdata, int wherefrom, void *cbhandle, int *useraction_p)
{    
    // cout << "ejecutando mycutcallback ..." << endl;
    int cur_numrows = CPXgetnumrows(env, cmrsp_model);
    int cur_numcols = CPXgetnumcols(env, cmrsp_model);
    *useraction_p = CPX_CALLBACK_DEFAULT;
    int status = 0;
    int node_count, nodes_left, mip_iterations;
    double best_integer, best_remaining;    
    
    if (/*wherefrom != CPX_CALLBACK_MIP && wherefrom != CPX_CALLBACK_MIP_BRANCH && wherefrom != CPX_CALLBACK_MIP_INCUMBENT &&
        wherefrom != CPX_CALLBACK_MIP_NODE && wherefrom != CPX_CALLBACK_MIP_HEURISTIC && */wherefrom != CPX_CALLBACK_MIP_CUT)
    {
        cout << "no se aplican cortes (wherefrom invalido)" << endl; // no deberia suceder (?)
        return status;
    }
    
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_NODE_COUNT, &node_count);
    if (status)
    {
        cout << "Error recuperando la cantidad de nodos procesados" << endl;
        return status;
    }
    
    if (param.ipc > 0 && node_count > 0 && node_count == last_node && iter > param.ipc)
    {
        return status;
    }
    
    if (node_count != last_node)
    {
        last_node = node_count;
        iter = 0;
    }
    else
    {
        iter++;
    }
    
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_NODES_LEFT, &nodes_left);
    if (status)
    {
        cout << "Error recuperando la cantidad de nodos por procesar" << endl;
        return status;
    }
        
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_MIP_ITERATIONS, &mip_iterations);
    if (status)
    {
        cout << "Error recuperando la cantidad de iteraciones" << endl;
        return status;
    }
    
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &best_integer);
    if (status)
    {
        cout << "Error recuperando la mejor solucion entera encontrada hasta el momento" << endl;
        return status;
    }        
    
    status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING, &best_remaining);
    if (status)
    {
        cout << "Error recuperando la mejor cota por explorar" << endl;
        return status;
    }        
    
    gap = 100.0 * (best_integer - best_remaining) / (0.0000000001 + best_integer);
    
    time_t now = time(NULL);
    
    if (node_count % 100 == 0)
    {
        cout << "(cutcallback) procesados = " << node_count << ", activos = " << nodes_left << ", iter = " << mip_iterations << ", best int = " << best_integer << ", best remaining = " << best_remaining << ", gap = " << gap << "%" <<  endl;
    }
    
    if (param.skip_factor > 0 && node_count % param.skip_factor != 0)
    {
        return status;
    }   
    
    double* x = (double *) malloc (cur_numcols * sizeof(double));
    status = CPXgetcallbacknodex (env, cbdata, wherefrom, x, 0, cur_numcols-1);
    if (status)
    {
        cout << "Error recuperando la solucion del nodo actual" << endl;
        *useraction_p = CPX_CALLBACK_DEFAULT;
        return status;
    }
    
    // cout << "Buscando cortes... ";
//     bool cuts_1a = false, cuts_1b = false, cuts_4 = false;
    
//    if (param.cutting_plane_1) cuts_1a = apply_cuts1(x, cbdata, wherefrom);
//    if (param.cutting_plane_1) cuts_1b = separate1(x, cbdata, wherefrom);
//    if (param.cutting_plane_4) cuts_4 = apply_cuts4(x, cbdata, wherefrom);

/*  
    if (param.cutting_plane_1) cuts_1b = apply_cuts1(x, cbdata, wherefrom);
    if (param.cutting_plane_1) cuts_1a = ring_xor_star_cuts1(x, cbdata, wherefrom); */

    bool cortes = false;
    for (int i = 0; i < metodos_separacion.size(); i++)
    {
        //cout << "Buscando cortes flia " << metodos_separacion[i]->nombre_familia() << "... ";
        cortes = cortes || metodos_separacion[i]->separar(env, x, cbdata, wherefrom);
        //cout << "fin de la busqueda" << endl;
    }

//    if (param.cutting_plane_1) cuts_1a = separate_path(x, cbdata, wherefrom);
    
    if (cortes)
    {
        for (int i = 0; i < metodos_separacion.size(); i++)
        {
            cortes_aplicados[i] += metodos_separacion[i]->cortes_aplicados();
            //if (cortes_aplicados[i] > 0)
                //cout << "Nodo = " << node_count << ", iter = " << iter << ", cortes aplicados = " << metodos_separacion[i]->cortes_aplicados() << endl;
        }
        // cout << "Se aplicaron cortes" << endl;
        //solution(x).save_dot("cmrs1.dot");
        *useraction_p = CPX_CALLBACK_SET;
    }
    
    free(x);
    
    return status;
}

