#include "HeuristicaInicialClusterRoute1.h"
#include "CMRSRep.h"
#include "util.h"
#include <iostream>
#include <vector>
#include <set>
#include <map>
#include <cmath>
using namespace std;

HeuristicaInicialClusterRoute1::HeuristicaInicialClusterRoute1(CMRSRep* cmrsrep)
{
    this->cmrsrep = cmrsrep;
}

void HeuristicaInicialClusterRoute1::aplicar(CPXENVptr env, CPXLPptr cmrsp_model) throw(CMRSPException)
{
    if (cmrsrep->n_rings() == 1)
        heuristica_un_anillo(env, cmrsp_model);
    else
        heuristica_varios_anillos(env, cmrsp_model);    
}

void HeuristicaInicialClusterRoute1::heuristica_un_anillo(CPXENVptr env, CPXLPptr cmrsp_model) throw(CMRSPException)
{
    /**************************************************************************
     * Heurística inicial (versión 1):
     * 
     * Se construye anillo por anillo (hasta no terminar la construcción de
     * un anillo, no se comienza la construcción de otro).
     * La construcción de los anillos se raliza desde el depóstio.
     * 
     * La cantidad diferencia en la cantidad de clientes entre dos anillos
     * no es más que uno. 
     * Es decir, algunos anillos tienen floor(n_customers/n_rings) clientes
     * y n_customers%n_rings anillos tienen floor(n_customers/n_rings)+1.
     * En los primeros (n_customers%n_rings) anillos se asigna un cliente 
     * más que en los ultimos.
     * 
     * Se elige primero el de menor costo con deposito, se
     * lo conecta al deposito, luego se elige el de menor costo con el ultimo que
     * se agregó, registrando que clientes ya fueron visitados para no
     * volver a visitarlos. Cuando se alcanza la cantidad de clientes
     * asignados al anillo, el ultimo cliente agregado se conecta al deposito.
     * Se continua asi con el siguiente anillo. 
     * (Se asume n_customers >= n_rings)
     * 
     **************************************************************************/
    int n_rings = cmrsrep->n_rings();
    int ring_capacity = cmrsrep->ring_capacity();
    int n_customers = cmrsrep->n_customers();
    int n_steiners = cmrsrep->n_steiners();
    v_index index_connection_cost = cmrsrep->index_connection_cost();
    v_index index_routing_cost = cmrsrep->index_routing_cost();
    v_index index_f = cmrsrep->index_f();
    v_index index_z = cmrsrep->index_z();
    vector<vector<double> > connection_cost = cmrsrep->connection_cost();
    vector<vector<double> > routing_cost = cmrsrep->routing_cost();
    
    int n_ring_customers = (int) floor(n_customers/n_rings);
    int n_ring_customers_rem = n_customers % n_rings;
    int customers_left = n_customers;
    vector<bool> customer_processed = vector<bool>(n_customers);
    vector<vector<int> > rings;
    double solution_cost = 0.0;
    int numnz = n_customers + n_rings;
    
    for (int i = 0; i < n_customers; i++) customer_processed[i] = false;
    
    
    while (customers_left > 0)
    {
        int customers_ring = 0;
        vector<int> ring;
        
        int last_node = index_connection_cost.D0;
        int customers_in_ring = n_ring_customers;
        if (rings.size() < n_ring_customers_rem) customers_in_ring++; // agrego 1 del resto (porque la cantidad de clientes puede no ser múltiplo de m)
        while (customers_ring < customers_in_ring && customers_left > 0)
        {
            int best_node;
            double min_cost = DBL_MAX;
            
            for (int next_node = index_connection_cost.CUSTOMER; next_node < index_connection_cost.CUSTOMER + n_customers; next_node++)
            {
                if (!customer_processed[next_node - index_connection_cost.CUSTOMER])
                {
                    double cost = connection_cost[last_node][next_node];
                    if (cost < min_cost)
                    {
                        best_node = next_node;
                        min_cost = cost;
                    }
                }                
            }
            
            customer_processed[best_node - index_connection_cost.CUSTOMER] = true;
            ring.push_back(best_node);
            last_node = best_node;
            customers_ring++;
            customers_left--;
            solution_cost += min_cost;           
        }
        solution_cost += connection_cost[last_node][index_connection_cost.D1];
        rings.push_back(ring);
    }
    
/*    for (int r = 0; r < rings.size(); r++)
    {
        for (int cust = 0; cust < rings[r].size(); cust++)
        {
            cout << customers_labels[rings[r][cust]-index_connection_cost.CUSTOMER] << endl;
        }
    }*/
    
    
    /**************************************************************************
     * Aplicando la solución como solución inicial en la instancia CPLEX
     **************************************************************************/
    // parametro de CPLEX para que considere la solucion inicial
    CPXsetintparam (env, CPX_PARAM_ADVIND, 2);
    
    int indices[numnz];
    double value[numnz];
    int index = 0;
    
    for (int r = 0; r < rings.size(); r++)
    {
        // arcos anillo
        int last_node = index_connection_cost.D0;
        for (int cust = 0; cust < rings[r].size(); cust++)
        {
            int next_node = rings[r][cust];
            indices[index] = cmrsrep->var_x(last_node, next_node);
            value[index] = 1.0;
            last_node = next_node;
            index++;
        }
        indices[index] = cmrsrep->var_x(last_node, index_connection_cost.D1);
        value[index] = 1.0;
        index++;
    }
    cout << "index = " << index << ", numnz = " << numnz << endl;
    
    int status = CPXcopymipstart(env, cmrsp_model, numnz, indices, value);
    
    if (status)
    {
        cerr << "Error seteando la solución inicial" << endl;
        char errmsg[__MSG_ERR_LEN];
        CPXgeterrorstring (env, status, errmsg);
        cerr << errmsg << endl;
        CPXcloseCPLEX(&env);
        throw CMRSPException(errmsg);
    }

}

void HeuristicaInicialClusterRoute1::heuristica_varios_anillos(CPXENVptr env, CPXLPptr cmrsp_model) throw(CMRSPException)
{
    /**************************************************************************
     * Heurística inicial (version 3)
     * 
     * Busqueda de clientes iniciales
     * Construccion simultanea de anillos
     * 
     **************************************************************************/
    cout << "HeuristicaInicialClusterRoute1" << endl;

    int n_rings = cmrsrep->n_rings();
    int ring_capacity = cmrsrep->ring_capacity();
    int n_customers = cmrsrep->n_customers();
    int n_steiners = cmrsrep->n_steiners();
    v_index index_connection_cost = cmrsrep->index_connection_cost();
    v_index index_routing_cost = cmrsrep->index_routing_cost();
    v_index index_f = cmrsrep->index_f();
    v_index index_z = cmrsrep->index_z();
    vector<vector<double> > connection_cost = cmrsrep->connection_cost();
    vector<vector<double> > routing_cost = cmrsrep->routing_cost();

    if (n_rings > 1)
    {
        /**********************************************************************
         * Busqueda heurística de n_rings clientes iniciales alejados
         **********************************************************************/
        double max_cost = DBL_MIN;
        vector<int> initials = vector<int>(n_rings);
        set<int> not_visited;
        int cust1, cust2;
        
        for (int cust = index_connection_cost.CUSTOMER; cust < index_connection_cost.CUSTOMER + n_customers; cust++)
            not_visited.insert(cust);

        for (int i = index_connection_cost.CUSTOMER; i < index_connection_cost.CUSTOMER + n_customers; i++)
        {
            for (int j = index_connection_cost.CUSTOMER; j < index_connection_cost.CUSTOMER + n_customers; j++)
            {
                if (i != j)
                {
                    double cost = connection_cost[i][j];
                    if (cost > max_cost)
                    {
                        cust1 = i;
                        cust2 = j;
                        max_cost = cost;
                    }
                }
            }
        }
        
        initials[0] = cust1;
        initials[1] = cust2;
        int n_initials = 2;
        not_visited.erase(cust1);
        not_visited.erase(cust2);
        while (n_initials < n_rings)
        {
            vector<pair<double, int> > costs_sum = vector<pair<double, int> >(n_customers - n_initials);
            vector<pair<double, int> > costs_var = vector<pair<double, int> >(n_customers - n_initials);
            vector<pair<double, int> > costs_var_orig = vector<pair<double, int> >(n_customers - n_initials);
            map<int, int> customer_index = map<int, int>();
            
            int k = 0;
            for (int candidate_customer = index_connection_cost.CUSTOMER; candidate_customer < index_connection_cost.CUSTOMER + n_customers; candidate_customer++)
            {
                if (find(initials.begin(), initials.end(), candidate_customer) == initials.end())
                {
                    /* Costos entre los clientes iniciales que ya agregué y el candidato */
                    double cost_sum = 0.0, cost_var = 0.0, cost_avg;
                    
                    for (int j = 0; j < n_initials; j++)
                    {
                        cost_sum += connection_cost[initials[j]][candidate_customer];
                    }
                    cost_avg = cost_sum / n_initials;
                    
                    for (int j = 0; j < n_initials; j++)
                    {
                        cost_var += (connection_cost[initials[j]][candidate_customer] - cost_avg)*(connection_cost[initials[j]][candidate_customer] - cost_avg);
                    }
                    
                    customer_index[candidate_customer] = k;
                    costs_sum[k] = pair<double, int> (cost_sum, candidate_customer);
                    costs_var[k] = pair<double, int> (cost_var, candidate_customer);
                    costs_var_orig[k] = pair<double, int> (cost_var, candidate_customer);
                    k++;
                }
            }
            
            sort(costs_sum.begin(), costs_sum.end());
            sort(costs_var.begin(), costs_var.end());
            
            // busco que en promedio/suma sea grande y que en varianza o desvios sea suficientemente pequeño 
            // (< cost_var[n_initials/2] menor o igual que la mediana de los desvios)
            double max_var = costs_var[(n_customers - n_initials)/2].first;
        
            int cust;
            for (int i = 0; i < n_customers - n_initials; i++)
            {
                cust = costs_sum[n_customers - n_initials - 1 - i].second;
                
                if (costs_var_orig[cust].first < max_var) break;
            }
            
            initials[n_initials] = cust;
            not_visited.erase(cust);
            n_initials++;
        }
        
//         cout << "Iniciales:" << endl;
//         for (int i = 0; i < initials.size(); i++)
//         {
//             cout << customers_labels[initials[i] - index_connection_cost.CUSTOMER] << endl;
//         }
//         
//         cout << "---------" << endl;
        
        /**********************************************************************
         * Construcción de anillos a partir de los clientes iniciales
         **********************************************************************/
        vector<vector<int> > rings = vector<vector<int> >(n_rings);
        vector<int> heads = vector<int>(n_rings);
        vector<int> tails = vector<int>(n_rings);
        vector<vector<pair<int, int> > > attachs = vector<vector<pair<int, int> > >(n_rings);
        vector<int> attachs_exists = vector<int>(n_customers);
        
        // inicialmente los anillos contienen los clientes iniciales
        for (int i = 0; i < n_rings; i++)
        {
            rings[i].push_back(initials[i]);
            heads[i] = initials[i];
            tails[i] = initials[i];
            attachs[i] = vector<pair<int, int> >();
        }
        
        // inicialmente ningun customer tiene otros adjuntos
        for (int i = 0; i < n_customers; i++) attachs_exists[i] = false;
        
        while (!not_visited.empty())
        {
            // se busca el cliente no visitado y anillo sin completar su capacidad lo mas cercano posible
            set<int>::const_iterator new_cust_it;
            int best_cust, best_ring;
            double cost, min_cost = DBL_MAX;
            bool head_connection;
            
            for (new_cust_it = not_visited.begin(); new_cust_it != not_visited.end(); new_cust_it++)
            {
                int new_cust = *new_cust_it;
                
                for (int ring = 0; ring < n_rings; ring++)
                {
                    if (rings[ring].size() + attachs[ring].size() < ring_capacity)
                    {
                        cost = connection_cost[new_cust][tails[ring]];
                        if (cost < min_cost)
                        {
                            best_cust = new_cust;
                            best_ring = ring;
                            head_connection = false;
                            min_cost = cost;
                        }
                        
                        cost = connection_cost[heads[ring]][new_cust];
                        if (cost < min_cost)
                        {
                            best_cust = new_cust;
                            best_ring = ring;
                            head_connection = true;
                            min_cost = cost;
                        }
                    }
                }
            }
            
            if (head_connection)
            {
                if (rings[best_ring].size() > 1)
                {
                    // conexion actual: ... -> cust_last2 -> cust_last1
                    // best_cust es el cliente con menor costo de conexion a cust_last1
                    int cust_last1 = rings[best_ring][rings[best_ring].size() - 1];
                    int cust_last2 = rings[best_ring][rings[best_ring].size() - 2];
                    
                    // se analizan las siguientes alternativas:
                    // 1) ... -> cust_last2 -> cust_last1 -> best_cust (anillo)
                    // 2) ... -> cust_last2 -> best_cust (anillo)  +   cust_last1 -> cust_last2 (arco estrella)
                    // 3) ... -> cust_last2 -> best_cust (anillo)  +   cust_last1 -> best_cust (arco estrella)
                    // para las alternativas 2) y 3) no pueden haber clientes adjuntos a cust_last1 (arcos estrella conectados a cust_last1)
                    int best_cust_route = best_cust - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_last1_route = cust_last1 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_last2_route = cust_last2 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    
                    double cost1 = connection_cost[cust_last2][cust_last1] + connection_cost[cust_last1][best_cust];
                    double cost2 = connection_cost[cust_last2][best_cust] + routing_cost[cust_last1_route][cust_last2_route];
                    double cost3 = connection_cost[cust_last2][best_cust] + routing_cost[cust_last1_route][best_cust_route];
                    
                    // elige una mejor alternativa de las 3
                    if ( attachs_exists[cust_last1_route - index_routing_cost.CUSTOMER] || ( (cost1 <= cost2) && (cost1 <= cost3) ) )
                    {
                        rings[best_ring].push_back(best_cust);
                    }
                    else if ((cost2 <= cost1) && (cost2 <= cost3))
                    {
                        rings[best_ring].pop_back();
                        rings[best_ring].push_back(best_cust);
                        attachs[best_ring].push_back(pair<int, int>(cust_last1_route, cust_last2_route));
                        attachs_exists[cust_last2_route - index_routing_cost.CUSTOMER] = true;
                    }
                    else // (cost3 <= cost1) && (cost3 <= cost2)
                    {
                        rings[best_ring].pop_back();
                        rings[best_ring].push_back(best_cust);
                        attachs[best_ring].push_back(pair<int, int>(cust_last1_route, best_cust_route));
                        attachs_exists[best_cust_route - index_routing_cost.CUSTOMER] = true;
                    }
                }
                else
                {
                    rings[best_ring].push_back(best_cust);
                }
                heads[best_ring] = best_cust;
            }
            else // !head_connection
            {
                if (rings[best_ring].size() > 1)
                {
                    // conexion actual: cust_fist1 -> cust_first2 -> ...
                    // best_cust es el cliente con menor costo de conexion a cust_first1
                    int cust_first1 = rings[best_ring][0];
                    int cust_first2 = rings[best_ring][1];
                    
                    // se analizan las siguientes alternativas:
                    // 1) best_cust -> cust_first1 -> cust_first2 -> ... (anillo)
                    // 2) best_cust -> cust_first2 -> ... (anillo)  +   cust_first1 -> cust_first2 (arco estrella)
                    // 3) best_cust -> cust_first2 -> ... (anillo)  +   cust_first1 -> best_cust (arco estrella)
                    // para las alternativas 2) y 3) no pueden haber clientes adjuntos a cust_first1 (arcos estrella conectados a cust_first1)
                    int best_cust_route = best_cust - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_first1_route = cust_first1 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;
                    int cust_first2_route = cust_first2 - index_connection_cost.CUSTOMER + index_routing_cost.CUSTOMER;

                    double cost1 = connection_cost[best_cust][cust_first1] + connection_cost[cust_first1][cust_first2];
                    double cost2 = connection_cost[best_cust][cust_first2] + routing_cost[cust_first1_route][cust_first2_route];
                    double cost3 = connection_cost[best_cust][cust_first2] + routing_cost[cust_first1_route][best_cust_route];

                    // elige una mejor alternativa de las 3
                    if (attachs_exists[cust_first1_route - index_routing_cost.CUSTOMER] || ((cost1 <= cost2) && (cost1 <= cost3)))
                    {
                        rings[best_ring].insert(rings[best_ring].begin(), best_cust);
                    }
                    else if ((cost2 <= cost1) && (cost2 <= cost3))
                    {
                        rings[best_ring][0] = best_cust;
                        attachs[best_ring].push_back(pair<int, int>(cust_first1_route, cust_first2_route));
                        attachs_exists[cust_first1_route - index_routing_cost.CUSTOMER] = true;
                    }
                    else // (cost3 <= cost1) && (cost3 <= cost2)
                    {
                        rings[best_ring][0] = best_cust;
                        attachs[best_ring].push_back(pair<int, int>(cust_first1_route, best_cust_route));
                        attachs_exists[best_cust_route - index_routing_cost.CUSTOMER] = true;
                    }
                }
                else
                {
                    rings[best_ring].insert(rings[best_ring].begin(), best_cust);
                }
                tails[best_ring] = best_cust;
            }
            
            not_visited.erase(best_cust);
        }
        
        
        /**********************************************************************
         * Solucion generada por la heurística
         **********************************************************************/
//         for (int ring = 0; ring < n_rings; ring++)
//         {
//             cout << "anillo " << ring << endl;
//             for (int cust = 0; cust < rings[ring].size(); cust++)
//             {
//                 cout << "\t" << customers_labels[rings[ring][cust] - index_connection_cost.CUSTOMER] << endl;
//             }
//             
//             cout << "adjuntos: " << attachs[ring].size() << endl;
//             for (int cust = 0; cust < attachs[ring].size(); cust++)
//             {
// /*                cout << index_routing_cost.CUSTOMER << " " << attachs[ring][cust].first << " -> " 
//                         << attachs[ring][cust].second << endl;*/
//                 cout << attachs[ring][cust].first - index_routing_cost.CUSTOMER << " " << customers_labels[attachs[ring][cust].first - index_routing_cost.CUSTOMER] << " -> " 
//                         << attachs[ring][cust].second - index_routing_cost.CUSTOMER << " " << customers_labels[attachs[ring][cust].second - index_routing_cost.CUSTOMER] << endl;
//             }
//         }        
//         cout << "fin" << endl;
        /**********************************************************************
         * Aplicando la solución como solución inicial en la instancia CPLEX
         **********************************************************************/    
        double obj = 0.0;
        int numnz = n_customers + n_rings + 2*(n_customers + n_steiners);
        int indices[numnz];
        double value[numnz];
        int index = 0;
        int last_node, next_node, node_z, node_f;
        int z, f;
        
        z = 0;
        for (int r = 0; r < n_rings; r++)
        {
            // arcos anillo
//             cout << "------ ANILLO " << r << endl;
            last_node = index_connection_cost.D0;
            f = 0;
            for (int cust = 0; cust < rings[r].size(); cust++)
            {
                next_node = rings[r][cust];
                obj += connection_cost[last_node][next_node];
                indices[index] = cmrsrep->var_x(last_node, next_node);
                value[index] = 1.0;
                last_node = next_node;
                index++;
                z++;
                f++;
                
                if (next_node >= index_connection_cost.CUSTOMER && next_node < index_connection_cost.CUSTOMER + n_customers)
                {
                    for (int cust1 = 0; cust1 < attachs[r].size(); cust1++)
                    {
                        int target = attachs[r][cust1].second - index_routing_cost.CUSTOMER + index_connection_cost.CUSTOMER;
//                         cout << "next_node = " << tmp_colname[next_node] << ", target = " << tmp_colname[target] << endl;
                        if (target == next_node) f++;
                    }
                }
                
                node_z = rings[r][cust] - index_connection_cost.CUSTOMER + index_z.CUSTOMER;
                indices[index] = cmrsrep->var_z(node_z);
                value[index] = z;
                index++;
                
                node_f = rings[r][cust] - index_connection_cost.CUSTOMER + index_f.CUSTOMER;
                indices[index] = cmrsrep->var_f(node_f);
                value[index] = f;
                index++;
            }
            obj += connection_cost[last_node][index_connection_cost.D1];
            indices[index] = cmrsrep->var_x(last_node, index_connection_cost.D1);
            value[index] = 1.0;
            index++;
            
            // arcos estrella
            f = 0;
            for (int cust = 0; cust < attachs[r].size(); cust++)
            {
                int source = attachs[r][cust].first;
                int target = attachs[r][cust].second;
                obj += routing_cost[source][target];
                indices[index] = cmrsrep->var_y(source, target);
                value[index] = 1.0;
                index++;
            }
        }
        
        cout << "Heuristica incial: obj = " << obj << endl;
//         cout << "index = " << index << ", numnz = " << numnz << endl;
//         cout << "--- Solucion inicial --- " << endl;
//         for (int i = 0; i < index; i++)
//         {
//             cout << tmp_colname[indices[i]] << " = " << value[i] << endl;
//         }
//         cout << "--- Fin solucion inicial --- " << endl;
        
        int status = CPXcopymipstart(env, cmrsp_model, index /*numnz*/, indices, value);
        
//int status = 0;
        if (status)
        {
            cerr << "Error seteando la solución inicial" << endl;
            char errmsg[__MSG_ERR_LEN];
            CPXgeterrorstring (env, status, errmsg);
            cerr << errmsg << endl;
            CPXcloseCPLEX(&env);
            throw CMRSPException(errmsg);
        }
    
    }    
    
}


// void CMRSP::initial_heuristic2() throw(CMRSPException)
// {
//     /**************************************************************************
//      * Heurística inicial (versión 2):
//      * 
//      * Se construyen los anillos de manera simultanea. 
//      * La construcción de los anillos se raliza desde el depóstio.
//      * 
//      * Se buscan los n_rings clientes mas cercanos al deposito. 
//      * Se realizan n_rings conexiones, desde el deposito con cada uno
//      * de estos clientes. Es decir, cada uno de estos clientes se asgina
//      * a un anillo distinto. Estos clientes se marcan como ultimos visitados
//      * de cada anillo. 
//      * 
//      * Mientras haya clientes sin visitar. Se busca el cliente sin visitar
//      * que tiene menor costo con los ultimos visitados (de los distintos anillos 
//      * que no estan cerrados o terminados). Se conecta el cliente a ese 
//      * ultimo cliente visitado. Una vez que un anillo alcanza los ring_capacity
//      * clientes, se toma el ultimo cliente y se lo conecta al deposito. 
//      * 
//      **************************************************************************/
//     set<int> not_visited = set<int>();
//     vector<vector<int> > rings = vector<vector<int> >(n_rings);
//     vector<double> mins = vector<double>(n_rings);
//     vector<int> last_nodes = vector<int>(n_rings);
//     int customers_left = n_customers;
//     int last_node, next_node;
//     
//     for (int cust = index_connection_cost.CUSTOMER; cust < index_connection_cost.CUSTOMER + n_customers; cust++)
//         not_visited.insert(cust);
//     
//     for (int i = 0; i < n_rings; i++) { mins[i] = DBL_MAX; last_nodes[i] = -1; }    
//     
//     last_node = index_connection_cost.D0;
//     for (int next_node = index_connection_cost.CUSTOMER; next_node < index_connection_cost.CUSTOMER + n_customers; next_node++)
//     {
//         for (int j = 0; j < n_rings; j++)
//         {
//             double cost = connection_cost[last_node][next_node];
//             if (cost < mins[j])
//             {
//                 for (int k = n_rings-1; k > j; k--)
//                 {
//                     mins[k] = mins[k-1];
//                     last_nodes[k] = last_nodes[k-1];
//                 }
//                 mins[j] = cost;
//                 last_nodes[j] = next_node;
//                 break;
//             }
//         }
//     }
//     
//     for (int i = 0; i < n_rings; i++)
//     {
//         rings[i].push_back(last_nodes[i]);
//         not_visited.erase(last_nodes[i]);
//     }
//     
//     int best_ring, best_node;
//     double min_cost;
//     while (!not_visited.empty())
//     {
//         min_cost = DBL_MAX;
//         set<int>::const_iterator it;
//         for (it = not_visited.begin(); it != not_visited.end(); it++)
//         {
//             next_node = *it;
//             for (int i = 0; i < n_rings; i++)
//             {
//                 if (rings[i].size() < ring_capacity)
//                 {
//                     double cost = connection_cost[last_nodes[i]][next_node];
//                     if (cost < min_cost)
//                     {
//                         best_ring = i;
//                         best_node = next_node;
//                         min_cost = cost;
//                     }
//                 }
//             }
//         }
//         rings[best_ring].push_back(best_node);
//         not_visited.erase(best_node);
//         last_nodes[best_ring] = best_node;
//     }
//     
// //     cout << "Resultado de heuristíca inicial" << endl;
// //     double solution_cost = 0.0;
// //     for (int i = 0; i < n_rings; i++)
// //     {
// //         last_node = index_connection_cost.D0;
// //         cout << "anillo " << i << endl;
// //         for (int cust = 0; cust < rings[i].size(); cust++)
// //         {
// //             solution_cost += connection_cost[last_node][rings[i][cust]];
// //             last_node = rings[i][cust];
// //             cout << customers_labels[rings[i][cust]-index_connection_cost.CUSTOMER] << endl;
// //         }
// //         solution_cost += connection_cost[last_node][index_connection_cost.D1];
// //     }
// //     cout << "costo = " << solution_cost << endl;
//     
//     /**************************************************************************
//      * Aplicando la solución como solución inicial en la instancia CPLEX
//      **************************************************************************/    
//     int numnz = n_customers + n_rings;
//     int indices[numnz];
//     double value[numnz];
//     int index = 0;
//     
//     for (int r = 0; r < rings.size(); r++)
//     {
//         last_node = index_connection_cost.D0;
//         for (int cust = 0; cust < rings[r].size(); cust++)
//         {
//             next_node = rings[r][cust];
//             indices[index] = var_x(last_node, next_node);
//             value[index] = 1.0;
//             last_node = next_node;
//             index++;
//         }
//         indices[index] = var_x(last_node, index_connection_cost.D1);
//         value[index] = 1.0;
//         index++;
//     }
//     cout << "index = " << index << ", numnz = " << numnz << endl;
//     
//     int status = CPXcopymipstart(env, cmrsp_model, numnz, indices, value);
//     
//     if (status)
//     {
//         cerr << "Error seteando la solución inicial" << endl;
//         char errmsg[__MSG_ERR_LEN];
//         CPXgeterrorstring (env, status, errmsg);
//         cerr << errmsg << endl;
//         CPXcloseCPLEX(&env);
//         throw CMRSPException(errmsg);
//     }
// 
// }
// 
